header "Acquisition Structures"
theory Acqh
imports Main Semantics WellNested SpecialLemmas
begin
text_raw {*\label{thy:Acqh}*}

subsection "Utilities"
  
subsubsection {* Combinators for @{text option}-datatype *}

text {* Extending a function to option datatype, where @{term None} indicates 
  failure *}
fun opt_ext1 :: "('a \<Rightarrow> 'b option) \<Rightarrow> 'a option \<Rightarrow> 'b option" where
  "opt_ext1 f None = None" |
  "opt_ext1 f (Some x) = f x"

fun opt_ext2 :: "('a \<Rightarrow> 'b \<Rightarrow> 'c option) \<Rightarrow> 'a option \<Rightarrow> 'b option \<Rightarrow> 'c option" 
  where
  "opt_ext2 f None _ = None" |
  "opt_ext2 f _ None = None" |
  "opt_ext2 f (Some x) (Some y) = f x y"

lemma opt_ext2_simps[simp]:
  "opt_ext2 f x None = None" by (cases x) auto

lemma opt_ext2_alt: 
    "opt_ext2 f x y = (
      case x of
        None \<Rightarrow> None |
        Some xx \<Rightarrow> (case y of
          None \<Rightarrow> None |
          Some yy \<Rightarrow> f xx yy
        )
    )"
  by (cases "(f,x,y)" rule: opt_ext2.cases) auto


subsection "Acquisition Structures"
text {*
  Acquisition structures are an abstraction of scheduling trees, that are sufficient
  to decide whether a tree is schedulable. The basic concept of acquisition 
  structures was invented by Kahlon et al. \cite{KIG05, KG06} as abstraction of a 
  linear execution of a single pushdown system. We extend this concept here to 
  scheduling trees of DPNs.
  *}

text {*
  An acquisition or release history is a partial map from locks to set of locks. 
  This is the same representation as in \cite{KG06}. Another, equivalent 
  representation is as a set of locks and a graph on locks.
  
  An acquisition structure is a triple of a release history, a set of locks and
  an acquisition history.
  *}

types
  'X ah = "'X \<Rightarrow> 'X set option"
  'X as = "'X ah \<times> 'X set \<times> 'X ah"

text {*
  This is a collection of the common split-lemmas required when reasoning about 
  acquisition histories
  *}
lemmas eahl_splits = option.split_asm list.split_asm prod.split_asm split_if_asm

subsubsection "Parallel Composition"

fun as_comp :: "'X as \<Rightarrow> 'X as \<Rightarrow> 'X as option" where
    "as_comp (l,u,e) (l',u',e') = (
       if dom l \<inter> dom l' = {} \<and> dom e \<inter> dom e' = {} then
         Some (l++l',u\<union>u',e++e')
       else
         None
    )"
  
definition as_comp_op 
  :: "'X as option \<Rightarrow> 'X as option \<Rightarrow> 'X as option" (infixr "\<parallel>" 56) where
  "op \<parallel> == opt_ext2 as_comp"

lemma as_comp_op_simps[simp]:
  "None \<parallel> x = None"
  "x \<parallel> None = None"
  "Some a \<parallel> Some b = as_comp a b"
  by (unfold as_comp_op_def) auto

lemma as_comp_assoc_helper: 
  "(Some x \<parallel> Some y) \<parallel> Some z = Some x \<parallel> Some y \<parallel> Some z"
  by (cases x, cases y, cases z) auto

lemma as_comp_assoc: "(x\<parallel>y)\<parallel>z = x\<parallel>y\<parallel>z"
  apply (cases x, simp) 
  apply (cases y, simp) 
  apply (cases z, simp)
  apply (simp only: as_comp_assoc_helper)
  done

interpretation as_comp_acz: ACIZ["op \<parallel>" "Some (empty,{},empty)" "None"]
  apply (unfold_locales)
  apply (auto simp add: as_comp_assoc)
  apply (case_tac "(as_comp,x,y)" rule: opt_ext2.cases)
  apply (auto simp add: map_add_comm)
  apply auto
  apply (case_tac x)
  apply simp_all
  apply (case_tac a, case_tac b)
  apply simp
  done
    (* declare as_comp_acz.simps [simp del]
  
    lemmas [simp] = as_comp_acz.left_zero as_comp_acz.right_zero as_comp_acz.left_ident as_comp_acz.right_ident
    *)

lemma as_comp_SomeE: 
  "\<lbrakk>h1 \<parallel> h2 = Some (l,u,e); 
    !!l1 u1 e1 l2 u2 e2. \<lbrakk> h1=Some (l1,u1,e1); h2=Some (l2,u2,e2); 
                           dom l1 \<inter> dom l2 = {}; dom e1 \<inter> dom e2 = {}; 
                           l=l1++l2; u=u1\<union>u2; e=e1++e2 
                         \<rbrakk> \<Longrightarrow> P
    \<rbrakk> \<Longrightarrow> P"
  apply (unfold as_comp_op_def)
  apply (cases h1, cases h2, simp_all)
  apply (cases h2, simp_all)
  apply (case_tac "(a,aa)" rule: as_comp.cases)
  apply (simp split: split_if_asm)
  apply blast
  done

subsubsection "Acquisition Structures of Scheduling Trees and Hedges"

text {* 
  This function adds a set of locks to every entry in a release history.
  On graph interpretation, this corresponds to adding edges from any initially 
  released lock to any lock in @{text X}. 
  *}
definition l_add_use :: "'X ah \<Rightarrow> 'X set \<Rightarrow> 'X ah" where 
  "l_add_use l X == \<lambda>x. case l x of None \<Rightarrow> None | Some Y \<Rightarrow> Some (Y\<union>X)"

text {*
  This function removes an initially released lock @{text x} from the release history.
  On graph interpretation, this corresponds to removing the node @{text x} from the 
  graph.
  *}
definition l_remove :: "'X ah \<Rightarrow> 'X \<Rightarrow> 'X ah" where 
  "l_remove l x == \<lambda>y. if y=x then None else l y"

text {*
  The acquisition history of a tree is defined inductively over the tree structure. 
  Note that we assume that spawn steps have no lock operation. For spawn steps with
  an operation on locks, the acquisition structure is defined to be @{const None}. 
  We further assume that a tree contains no two initial releases of the same lock. 
  In this case, its acquisition structure has no meaning any more. However, if an 
  execution tree contains two final acquisitions of the same lock, its acquisition 
  structure is defined to be @{const None}.
  
  Intuitively, the release history maps all locks that are initially released to the 
  set of locks that have to be used before the initial release. The set of used locks
  contains the locks that are used by the execution tree (But not the locks that are
  only initially released or finally acquired). The acquisition history maps all locks
  that are finally acquired to the set of locks that have to be used after the final
  acquisition.
  *}
fun as :: "('P,'\<Gamma>,'L,'X) lex_tree \<Rightarrow> 'X as option" where
  "as (NLEAF \<pi>) = Some (empty,{},empty)" |
  "as (NNOSPAWN (LNone l) t) = as t" |
  "as (NSPAWN (LNone l) ts t) = as ts \<parallel> as t" |
  "as (NNOSPAWN (LAcq x) t) = (
    case as t of
      None \<Rightarrow> None |
      Some (l,u,e) \<Rightarrow>
        if x\<in>dom l then
          Some (l_add_use (l_remove l x) {x},insert x u,e)
        else if x\<notin>dom e then
          Some (l,u, e(x\<mapsto>u))
        else
          None
  )" |
  "as (NNOSPAWN (LRel x) t) = (
    case as t of
      None \<Rightarrow> None |
      Some (l,u,e) \<Rightarrow> Some (l(x\<mapsto>{}),u,e)
  )" |
  "as _ = None"


text {* The aquisition structure of a hedge is the parallel composition of the 
  acquisition structures of its trees. The acquisition structure of the empty hedge 
  is the identity acquisition structure @{term "Some (empty,{},empty)"}. *}
fun ash :: "('P,'\<Gamma>,'L,'X) lex_hedge \<Rightarrow> 'X as option" where
  "ash [] = Some (empty,{},empty)" |
  "ash (t#h) = as t \<parallel> ash h"

lemma l_add_use_dom[simp]: "dom (l_add_use l X) = dom l" 
  by (unfold l_add_use_def) (auto split: option.split_asm)

lemma l_add_use_empty[simp]: "l_add_use empty X = empty" 
  by (rule ext) (auto simp add: l_add_use_def split: option.split) 

lemma l_add_use_eq_empty[simp]: "l_add_use f X = empty \<longleftrightarrow> f=empty"
  apply (auto)
  apply (rule ext)
  apply (drule_tac x=x in fun_cong)
  apply (simp add: l_add_use_def split: option.split_asm)
  done

lemma l_add_use_add[simp]: 
  "l_add_use (l++l') X = l_add_use l X ++ l_add_use l' X"
  apply (unfold l_add_use_def)
  apply (rule ext)
  by (auto split: option.split simp add: map_add_def)

lemma l_add_use_le: "l \<le> l_add_use l X"
  apply (auto simp add: l_add_use_def intro!: le_funI)
  apply (case_tac "l x")
  apply auto
  done

lemma l_remove_add[simp]: "l_remove (l1++l2) m = l_remove l1 m ++ l_remove l2 m"
  by (unfold l_remove_def map_add_def) (auto intro: ext)

lemma l_remove_no_eff[simp]: "x\<notin>dom l \<Longrightarrow> l_remove l x = l"
  by (unfold l_remove_def) (auto intro: ext)

lemma l_remove_dom[simp]: "dom (l_remove l x) = dom l - {x}"
  by (unfold l_remove_def) (auto split: split_if_asm)

lemma l_remove_app[simp]: 
  "l_remove l x x = None"
  "x\<noteq>x' \<Longrightarrow> l_remove l x x' = l x'"
  by (unfold l_remove_def) auto

lemma l_remove_eq_empty: "l_remove l x = empty \<Longrightarrow> dom l \<subseteq> {x}"
  by (fastsimp simp add: l_remove_def dest: fun_cong split: split_if_asm)

lemma l_remove_le_l [simp]: "l_remove l x \<le> l"
  by (auto simp add: l_remove_def intro: le_funI)

lemma as_ran_e_le_u: "as t = Some (l,u,e) \<Longrightarrow> \<Union>ran e \<subseteq> u"
  apply (induct t arbitrary: l u e)
  apply fastsimp
  apply (case_tac L)
  apply (simp_all split: eahl_splits)
  apply fastsimp
  apply fastsimp
  apply (case_tac L)
  apply (simp_all)
  apply (fastsimp elim: as_comp_SomeE)
  done

lemma ash_le_u: "ash h = Some (l,u,e) \<Longrightarrow> \<Union>ran e \<subseteq> u" 
proof (induct h arbitrary: l u e rule: ash.induct)
  case 1 thus ?case by auto
next
  case 2 thus ?case
    apply simp
    apply (erule as_comp_SomeE)
    apply (fastsimp dest!: as_ran_e_le_u)
    done
qed

lemma ash_final[simp]: "final h \<Longrightarrow> ash h=Some (empty,{},empty)"
  apply (induct h)
  apply auto
  apply (case_tac a)
  apply simp_all
  done

lemma ash_append[simp]: "ash (h1@h2) = ash h1 \<parallel> ash h2"
  by (induct h1 arbitrary: h2) (auto simp add: as_comp_acz.simps)

lemma ash_LNone_simps[simp]: 
  "ash (h1@NSPAWN (LNone l) ts t#h2) = ash (h1@ts#t#h2)"
  "ash (h1@NNOSPAWN (LNone l) t#h2) = ash (h1@t#h2)"
  by (simp_all add: as_comp_acz.simps)

subsection "Consistency of Acquisition Structures"
  text {*
    The consistency criterium of an acquisition structure decides whether the 
    corresponding hedge can be scheduled. Note that we currently do not check this 
    criterium during construction of the acquisition structure, but only at the end, for 
    the completely constructed acquisition structure.

    The consistency criterium has two parts. The first part is a generalization of 
    the @{text "\<not>\<exists>m\<^sub>1,m\<^sub>2. m\<^sub>1\<in>h\<^sub>1(m\<^sub>2) \<and> m\<^sub>2\<in>h\<^sub>2(m\<^sub>1)"}-condition of \cite{KIG05}.
    There, the condition was checked for two separate acquisition histories 
    @{text "h\<^sub>1"} and @{text "h\<^sub>2"} that resulted from executions of two independent 
    pushdown systems.
    Here, we have one execution described as a tree. This criterium can
    be interpreted as checking acyclicity of a graph defined by the acquisition 
    histories.
    In \cite{KIG05}, every possible cycle has length two, hence their condition is 
    sufficient. In our setting, a cycle may have arbitrary length (bounded only by 
    the number of locks), hence we use a general cyclicity check.
    *}

  text {*
    The acquisition and release histories encode a graph between locks. For an 
    acquisition history @{term e}, the graph contains an edge @{term "(x,x')"} 
    if @{term x} has to be finally acquired before @{term "x'"} is used, that is if 
    @{term "x\<in>dom e \<and> x'\<in>the (e x)"}

    For a release history @{term l}, the graph contains an edge @{term "(x,x')"} if 
    @{term x} has to be used before @{term "x'"} is initially released, that is if 
    @{term "x'\<in>dom l \<and> x\<in>the (l x')"}
    *}
  definition agraph :: "'X ah \<Rightarrow> ('X\<times>'X) set" where 
    "agraph e == { (x,x') . x\<in>dom e \<and> x'\<in>the (e x) }"
  definition rgraph :: "'X ah \<Rightarrow> ('X\<times>'X) set" where 
    "rgraph l == { (x,x') . x'\<in>dom l \<and> x\<in>the (l x') }"
  
  lemma agraph_alt: "agraph e = { (x,x') . \<exists>X'. e x = Some X' \<and> x'\<in>X'}" 
    by (unfold agraph_def) auto
  lemma rgraph_alt: "rgraph l = { (x,x') . \<exists>X. l x' = Some X \<and> x\<in>X }" 
    by (unfold rgraph_def) auto

  text {*
    For the same map, the acquisition graph is the converse of the release graph. This
    lemma makes reasoning simpler at some points, as acquisition and release 
    histories have the same type, and cyclicity is equivalent for
    a graph and its converse.
    *}
  lemma agraph_rgraph_converse: "agraph h = (rgraph h)\<inverse>"
    by (unfold agraph_def rgraph_def) auto

  lemma agraph_add_union: 
    "\<lbrakk>dom e \<inter> dom e' = {}\<rbrakk> \<Longrightarrow> agraph (e++e') = agraph e \<union> agraph e'"
    by (unfold agraph_def) (auto simp add: map_add_def split: option.split_asm)

  lemma rgraph_add_union: 
    "\<lbrakk>dom l \<inter> dom l' = {}\<rbrakk> \<Longrightarrow> rgraph (l++l') = rgraph l \<union> rgraph l'"
    by (unfold rgraph_def) (auto simp add: map_add_def split: option.split_asm)

  lemma agraph_domain_simp[simp]: 
    "Domain (agraph h) = dom h - { x . h x = Some {} }"
    by (unfold agraph_def) auto

  lemma agraph_range_simp[simp]: "Range (agraph h) = \<Union>ran h"
    by (unfold agraph_def) (auto simp add: ran_def)

  lemma rgraph_domain_simp[simp]: "Domain (rgraph h) = \<Union>ran h"
    by (unfold rgraph_def) (auto simp add: ran_def)

  lemma rgraph_range_simp[simp]: 
    "Range (rgraph h) = dom h - { x . h x = Some {} }"
    by (unfold rgraph_def) auto

  lemma graph_empty[simp]: 
    "agraph empty = {}" 
    "rgraph empty = {}" 
    by (auto simp add: agraph_def rgraph_def)

  lemma rgraph_add_use: "rgraph (l_add_use l X) = rgraph l \<union> X\<times>dom l"
    by (unfold rgraph_def l_add_use_def) (auto split: option.split_asm)
    
  lemma rgraph_remove: "rgraph (l_remove l x) = rgraph l - UNIV\<times>{x}"
    by (unfold rgraph_def l_remove_def) (auto split: option.split_asm)

  lemma rgraph_upd: "x\<notin>dom l \<Longrightarrow> rgraph (l(x\<mapsto>X)) = rgraph l \<union> X\<times>{x}"
    by (unfold rgraph_def) auto

  lemmas rgraph_ops = rgraph_add_use rgraph_remove rgraph_upd

  lemma agraph_upd: "x\<notin>dom e \<Longrightarrow> agraph (e(x\<mapsto>X)) = agraph e \<union> {x}\<times>X"
    by (unfold agraph_def) (auto split: split_if_asm)

  lemmas agraph_ops = agraph_upd

  lemma rgraph_mono: "l\<le>l' \<Longrightarrow> rgraph l \<subseteq> rgraph l'"
    apply (unfold rgraph_alt)
    apply auto
    apply (drule_tac x=b in le_funD)
    apply (auto elim: le_optE)
    done

  lemma agraph_mono: "e\<le>e' \<Longrightarrow> agraph e \<subseteq> agraph e'"
    by (simp add: agraph_rgraph_converse rgraph_mono)

  text {*
    An acquisition or release history is consistent, iff its graph is acyclic.
    *}
  abbreviation cons_rh :: "'X ah \<Rightarrow> bool" where "cons_rh h == acyclic (rgraph h)"
  abbreviation cons_ah :: "'X ah \<Rightarrow> bool" where "cons_ah h == acyclic (agraph h)"
  abbreviation "cons_h == cons_rh"

  text {*
    As noted above, the cyclicity criterion is equivalent for a graph and its converse,
    such that we can use @{term cons_h} for both, acquisition and release histories.
    *}
  lemma cons_ah_rh_eq: 
    "cons_ah e = cons_h e" 
    "cons_rh r = cons_h r" 
    by (simp_all add: agraph_rgraph_converse)

  lemma cons_h_empty[simp]: "cons_h empty"
    apply (unfold rgraph_def)
    apply auto
    apply (metis Collect_def wfP_acyclicP wfP_empty)
    done

  lemma cons_h_add: 
    "\<lbrakk>dom h \<inter> dom h' = {}; cons_h (h++h')\<rbrakk> \<Longrightarrow> cons_h h"
    "\<lbrakk>dom h \<inter> dom h' = {}; cons_h (h++h')\<rbrakk> \<Longrightarrow> cons_h h'"
    by (auto dest: acyclic_union simp add: rgraph_add_union)

  lemma cons_h_antimono: "\<lbrakk>l\<le>l'; cons_h l'\<rbrakk> \<Longrightarrow> cons_h l"
    using acyclic_subset[OF _ rgraph_mono] .


  lemma cons_h_update: 
    assumes A: "cons_h h" "X\<inter>insert x (dom h) = {}" 
    shows "cons_h (h(x\<mapsto>X))"
  proof -
    have "l_remove h x \<le> h" (is "?h \<le> _") by auto
    with cons_h_antimono A(1) have CONS: "cons_h ?h" by blast
    have MND[simp]: "x\<notin>dom ?h" by auto
    have [simp]: "h(x\<mapsto>X) = ?h(x\<mapsto>X)" by (auto simp add: l_remove_def intro: ext)
    have "cons_h (?h(x\<mapsto>X))" proof (rule ccontr, erule cyclicE)
      fix y assume "(y,y)\<in> (rgraph (l_remove h x(x \<mapsto> X)))\<^sup>+"
      hence "(y, y) \<in> (rgraph (l_remove h x) \<union> X\<times>{x})\<^sup>+" by (simp add: rgraph_ops)
      thus False proof (cases rule: trancl_multi_insert)
        case orig with CONS show False by (auto simp add: acyclic_def)
      next
        case (via x') hence C: "(x,x')\<in>(rgraph ?h)\<^sup>*" by auto
        show False using C proof (cases rule: rtrancl.cases)
          case rtrancl_refl with A(2) via(1) show False by auto
        next
          case (rtrancl_into_rtrancl _ b) hence "(b,x')\<in>rgraph ?h" by auto
          hence "x'\<in>dom ?h" by (auto simp add: rgraph_def l_remove_def)
          hence "x'\<in>dom h" by (auto simp add: l_remove_def split: split_if_asm)
          with A(2) via(1) show False by auto
        qed
      qed
    qed
    thus ?thesis by simp
  qed

  lemma cons_h_update2: 
    assumes A: "cons_h h" "x\<notin>dom h" "x\<notin>X" "x\<notin>\<Union>ran h" 
    shows "cons_h (h(x\<mapsto>X))"
  proof -
    from A(1) have A': "acyclic (agraph h)" by (simp add: agraph_rgraph_converse)
    from A(4) have XNIR: "x\<notin>Range (agraph h)" by simp
    hence [simp]: "!!y. \<not> (y,x)\<in>(agraph h)" by blast
    have "agraph (h(x\<mapsto>X)) = agraph h \<union> {x}\<times>X" 
      by (simp add: agraph_ops[OF A(2)])
    moreover have "acyclic (agraph h \<union> {x}\<times>X)" 
      apply (rule ccontr)
      apply (erule cyclicE)
    proof -
      fix xa assume "(xa, xa) \<in> (agraph h \<union> {x} \<times> X)\<^sup>+"
      thus False proof (cases rule: trancl_multi_insert2)
        case orig thus False using A' by (unfold acyclic_def) auto
      next
        case (via xb) hence "(xb,x)\<in>(agraph h)\<^sup>*" by auto
        thus False proof (cases rule: rtrancl.cases)
          case rtrancl_refl 
          with via(1) A(3) show False by auto
        next
          case (rtrancl_into_rtrancl a b c) 
          hence "(b,x)\<in>agraph h" by simp 
          thus False by simp
        qed
      qed
    qed
    ultimately have "acyclic (agraph (h(x\<mapsto>X)))" by simp
    thus ?thesis by (simp add: agraph_rgraph_converse)
  qed

  lemma cons_h_remove: "cons_h l \<Longrightarrow> cons_h (l_remove l m)"
    by (auto simp add: rgraph_ops intro: acyclic_subset)

  lemma cons_h_add_use: "\<lbrakk>m\<notin>dom l; cons_h l\<rbrakk> \<Longrightarrow> cons_h (l_add_use l {m})" 
    apply (rule ccontr)
    apply (erule cyclicE)
  proof -
    fix x
    assume A: "m \<notin> dom l" "cons_h l" "(x, x) \<in> (rgraph (l_add_use l {m}))\<^sup>+"
    from A(3) have "(x,x)\<in>(rgraph l \<union> {m}\<times>dom l)\<^sup>+" by (simp add: rgraph_ops)
    thus False 
    proof (cases rule: trancl_multi_insert2)
      case orig 
      with A(2) show False by (auto simp add: acyclic_def)
    next
      case (via xh) from via(2) show False 
      proof (cases rule: rtrancl.cases)
        case rtrancl_refl 
        hence [simp]: "x=m" by blast
        from via(3)[simplified] show False 
        proof (cases rule: rtrancl.cases)
          case rtrancl_refl 
          hence "xh=m" by blast 
          with A(1) via(1) show False by simp
        next
          case rtrancl_into_rtrancl 
          hence "m\<in>dom l" by (auto simp add: rgraph_def)
          with A(1) via(1) show False by simp
        qed
      next
        case rtrancl_into_rtrancl 
        hence "m\<in>dom l" by (auto simp add: rgraph_def)
        with A(1) via(1) show False by simp
      qed
    qed
  qed

  lemma cons_h_add_remove: "cons_h l \<Longrightarrow> cons_h (l_add_use (l_remove l m) {m})"
    by (auto intro: cons_h_add_use cons_h_remove)

  lemma cons_h_add_remove_partial: 
    "\<lbrakk> m\<notin>dom l1; cons_h (l1++l2) \<rbrakk> \<Longrightarrow> 
       cons_h (l1 ++ l_add_use (l_remove l2 m) {m})"
  proof -
    assume A: "m\<notin>dom l1"
    hence 
      LE: "l1 ++ l_add_use (l_remove l2 m) {m} \<le> 
             l_add_use (l_remove (l1++l2) m) {m}"
      apply simp
      apply (rule map_add_first_le)
      apply (simp add: l_add_use_le)
      done
    assume "cons_h (l1++l2)"
    hence "cons_h (l_add_use (l_remove (l1++l2) m) {m})" 
      by (blast intro: cons_h_add_remove)
    with cons_h_antimono[OF LE] show ?thesis by blast
  qed


  text {*
    The consistency condition for acquisition structures checks available locks in 
    addition to consistency of the acquisition and release histories.
    *}
  fun cons_as :: "'X as \<Rightarrow> 'X set \<Rightarrow> bool" where
    "cons_as (l,u,e) \<xi> \<longleftrightarrow> 
       u\<inter>(\<xi>-dom l) = {} \<and> dom e \<inter> (\<xi>-dom l) = {} \<and> cons_h l \<and> cons_h e"
 
  lemma cons_as_antimono: "\<lbrakk>cons_as h \<xi>; \<xi>'\<subseteq>\<xi>\<rbrakk> \<Longrightarrow> cons_as h \<xi>'"
    by (cases h) auto

  fun cons where
    "cons None X = False" |
    "cons (Some (l,u,e)) X = cons_as (l,u,e) X"

  subsubsection "Minimal Elements"

  lemma finite_acyclic_wf: "\<lbrakk>finite r; acyclic r\<rbrakk> \<Longrightarrow> wf r"
    apply (simp only: finite_wf_eq_wf_converse[symmetric])
    apply (blast intro: finite_acyclic_wf_converse)
    done


  text {*
    The minimal elements of acquisition and release histories corresponds to those 
    final acquisitions or initial releases that can safely be scheduled as next step 
    ---  for an acquisition history without blocking any further locks usage and 
    for a release history without requiring usage of already acquired locks.
    *}
  
  abbreviation "rh_min l m == m\<in>dom l \<and> dom l \<inter> the (l m) = {}"
  abbreviation "ah_min e m == m\<in>dom e \<and> m\<notin>\<Union>ran e"

  lemma rh_min_alt: 
    "rh_min l m = (case l m of None \<Rightarrow> False | Some M \<Rightarrow> dom l \<inter> M = {})"
    by (fastsimp split: option.split_asm)

  text {*
    There exists a minimal element in a consistent release history. Note that this 
    lemma depends on the set of locks being finite, as assumed by the 
    @{const LDPN} locale.
    *}
    (* TODO: This seems to be some interesting property of acyclic graphs, something 
       like:
          In each acyclic, nonempty graph there exists a vertex whose predecessors 
          have no predecessors themselves.
    *)
    (*
      TODO: Hmm, have we understood this completely? We seem to have lost symmetry to 
        acquisition histories. In some sense, if we find a minimal element, this 
        corresponds to a schedulable use-step. IR-steps are always schedulable.
    *)
  theorem (in LDPN) cons_h_ex_rh_min: 
    fixes l :: "'X ah" 
    assumes A: "l\<noteq>empty" "cons_h l" 
    shows "\<exists>m. rh_min l m" 
  proof -
    {
      fix M and mx::'X and k
      assume "\<forall>m. \<not>rh_min l m"
      hence B: "!!m lm. l m = Some lm \<Longrightarrow> dom l \<inter> lm \<noteq> {}" 
        by (unfold rh_min_alt) (auto split: option.split_asm)
      have "\<lbrakk> card (UNIV::'X set) - card M = k; mx\<notin>M; mx\<in>dom l; 
              !!m. m\<in>M \<Longrightarrow> (mx,m)\<in>(rgraph l)\<^sup>+
            \<rbrakk> \<Longrightarrow> False"
      proof (induct k arbitrary: M mx)
        case 0 hence "M=UNIV" by auto
        with 0 have False by simp
        thus ?case ..
      next
        case (Suc n) 
        then obtain lmx where LMX: "l mx = Some lmx" by auto
        with B obtain m' where M': "m'\<in>dom l" "m'\<in>lmx" by blast
        with LMX have G: "(m',mx)\<in>rgraph l" by (unfold rgraph_def) auto
        {
          assume "m'\<in>M" 
          with Suc.prems have "(mx,m')\<in>(rgraph l)\<^sup>+" by auto
          also note r_into_trancl[OF G] 
          finally have False using A(2) by (unfold acyclic_def) auto
        } moreover {
          assume C: "m'\<notin>M" "m'\<noteq>mx" hence C': "m'\<notin>M\<union>{mx}" by auto
          with Suc.prems(4) G have 1: "!!m. m\<in>M\<union>{mx} \<Longrightarrow> (m',m)\<in>(rgraph l)\<^sup>+" 
            by (auto intro: r_into_trancl trancl_trans)
          from Suc.prems(1,2) have 
            2: "card (UNIV::'X set) - card (M\<union>{mx}) = n" 
            by (simp)
          from Suc.hyps[OF 2 C' M'(1) 1] have False .
        } moreover {
          assume "m'=mx" 
          with r_into_trancl[OF G] have False using A(2) 
            by (unfold acyclic_def) auto
        } ultimately show False by blast
      qed
    } note X=this
    from A obtain m where "m\<in>dom l" by (subgoal_tac "dom l \<noteq> {}") (blast, auto)
    with X[of "{}" _ m] A show ?thesis by - (rule ccontr, auto)
  qed

  
  text {*
    There exists a minimal element in a consistent acquisition history.

    Note that this lemma depends on the set of locks being finite, as constrained by 
    the @{const LDPN} locale.
    *}
  theorem (in LDPN) cons_h_ex_ah_min: 
    fixes e :: "'X ah" 
    assumes A: "e\<noteq>empty" "cons_h e" 
    shows "\<exists>m. ah_min e m"
  proof (cases "agraph e = {}")
    case True from A(1) obtain m where "m\<in>dom e" by (blast elim: nempty_dom)
    moreover with True have "m\<notin>\<Union>ran e" by (auto simp add: agraph_def ran_def)
    ultimately show ?thesis by blast
  next
    case False
    from A(2) cons_ah_rh_eq(1)[symmetric, of e] have "cons_ah e" by simp
    hence WF: "wf (agraph e)" by (auto intro: finite_acyclic_wf)
    from wf_min[of "agraph e", OF WF False] obtain m where 
      "m \<in> Domain (agraph e) - Range (agraph e)" .
    hence "m\<in>dom e" "m\<notin>\<Union>ran e" by (auto simp add: agraph_def ran_def)
    thus ?thesis by blast
  qed
    

subsubsection "Well-Nestedness and Acquisition Structures"

text {*
  Only locks that are on the lock-stack can be initially released:
*}
lemma wn_t_dom_l_lower_\<mu>: 
  "\<lbrakk>wn_t' t \<mu>; as t = Some (l, u, e)\<rbrakk> \<Longrightarrow> dom l \<subseteq> set \<mu>"
  apply (induct t arbitrary: \<mu> l u e)
  apply fastsimp
  apply (case_tac L)
  apply fastsimp
  apply (auto split: option.split_asm list.split_asm split_if_asm 
              simp add: l_remove_def l_add_use_def)
  apply (fastsimp)
  apply (fastsimp)
  apply (fastsimp)
  apply (case_tac L)
  apply (fastsimp elim: as_comp_SomeE)+
  done

lemmas wn_dom_l_empty = wn_t_dom_l_lower_\<mu>[of _ "[]", simplified]

lemma wn_h_dom_l_lower_\<mu>: 
  "\<lbrakk>wn_h h \<mu>; ash h = Some (l,u,e)\<rbrakk> \<Longrightarrow> dom l \<subseteq> locks_\<mu> \<mu>"
  apply (induct h \<mu> arbitrary: l u e rule: wn_h.induct)
  apply auto
  apply (force dest: wn_t_dom_l_lower_\<mu> elim!: as_comp_SomeE)
  done


text {*
  Due to well-nestedness, if a lock @{text x} is left, all locks that are above this 
  lock on the stack are left, too. This lemma expresses leaving a lock by means of 
  the domain of the release-history. Moreover, the release histories of the locks 
  released before are smaller or equal than the release history of @{text x}, and do 
  not contain @{text x}.
  *}

lemma wn_t_dom_l_stack: "\<lbrakk>wn_t' t \<mu>; as t = Some (l,u,e); x\<in>dom l\<rbrakk> \<Longrightarrow> 
  \<exists>\<mu>1 \<mu>2. \<mu>=\<mu>1@x#\<mu>2 \<and> set \<mu>1 \<subseteq> dom l \<and> 
          (\<forall>x'\<in>set \<mu>1. l x' \<le> l x \<and> 
             (case l x' of None \<Rightarrow> True | Some lx' \<Rightarrow> x\<notin>lx' \<and> x'\<notin>lx')
          )"
proof (induct t arbitrary: \<mu> l u e x)
  case NLEAF thus ?case by fastsimp
next
  case (NSPAWN lab ts t) 
  from NSPAWN.prems(1) obtain nlab where [simp]: "lab=LNone nlab" 
    by (cases lab, simp_all)
  from NSPAWN.prems(1) have WN: "wn_t' ts []" "wn_t' t \<mu>" by auto
  from NSPAWN.prems(2) have "as ts \<parallel> as t = Some (l,u,e)" by simp
  then obtain l1 u1 e1 l2 u2 e2 where 
    [simp]: "l=l1++l2" "u=u1\<union>u2" "e=e1++e2" and 
      SPLIT: "as ts = Some (l1,u1,e1)" "as t = Some (l2,u2,e2)" 
             "dom l1 \<inter> dom l2 = {}" "dom e1 \<inter> dom e2 = {}" 
    by (blast elim!: as_comp_SomeE)
  have [simp]: "l1 = empty" proof -
    {
      fix x assume A: "x\<in>dom l1"
      from NSPAWN.hyps(1)[OF WN(1) SPLIT(1) A] have False by blast
    }
    thus ?thesis by force
  qed
  from `x\<in>dom l` have A: "x\<in>dom l2" by auto
  from NSPAWN.hyps(2)[OF WN(2) SPLIT(2) A] obtain \<mu>1 \<mu>2 where 
    "\<mu>=\<mu>1@x#\<mu>2" "set \<mu>1 \<subseteq> dom l" 
    "\<forall>x'\<in>set \<mu>1. l x' \<le> l x \<and> 
       (case l x' of None \<Rightarrow> True | Some lx' \<Rightarrow> x \<notin> lx' \<and> x'\<notin>lx')"
    by auto
  thus ?case by blast
next
  case (NNOSPAWN lab t)
  show ?case proof (cases lab)
    case (LNone nlab) with NNOSPAWN show ?thesis by simp blast
  next
    case (LAcq x')[simp]
    from NNOSPAWN.prems(2) obtain l' u' e' where 
      HTFMT: "as t = Some (l',u',e')" 
      by (auto split: option.split_asm list.split_asm split_if_asm prod.split_asm)
    with NNOSPAWN.prems(2,3) have MNE: "x\<noteq>x'" 
      by (auto split: split_if_asm simp add: l_remove_def l_add_use_def)
    from NNOSPAWN.prems(1) have WN: "wn_t' t (x'#\<mu>)" by simp
    {
      assume "x'\<in>dom l'"
      with NNOSPAWN.prems(2) HTFMT have 
        [simp]: "l=l_add_use (l_remove l' x') {x'}" "u = insert x' u'" "e'=e" 
        by (auto split: option.split_asm list.split_asm split_if_asm prod.split_asm)
      with MNE NNOSPAWN.prems(3) have MID: "x\<in>dom l'" by auto
      from NNOSPAWN.hyps[OF WN HTFMT MID] obtain \<mu>1 \<mu>2 where 
        IHAPP: "x'#\<mu> = \<mu>1@x#\<mu>2" "set \<mu>1 \<subseteq> dom l'" 
               "\<forall>x'\<in>set \<mu>1. l' x' \<le> l' x \<and> 
                  (case l' x' of None \<Rightarrow> True | Some lx' \<Rightarrow> x \<notin> lx' \<and> x'\<notin>lx')" 
        by blast
      from IHAPP(3) MNE have 
        IHAPP3': "\<forall>x'\<in>set \<mu>1. l x' \<le> l x \<and> 
                    (case l x' of None \<Rightarrow> True | Some lx' \<Rightarrow> x \<notin> lx' \<and> x'\<notin>lx')" 
        (* TODO: Cleanup !*)
        apply safe
        apply (case_tac "x'=x'a")
        apply (simp add: l_add_use_def)
        apply (subgoal_tac "l' x'a \<le> l' x")
        apply (erule le_optE)
        apply (simp add: l_add_use_def split: option.split)
        apply (auto simp add: l_add_use_def split: option.split) [1]
        apply simp
        apply (simp add: l_add_use_def l_remove_def)
        apply (split option.split_asm option.split)+
        apply meson
        apply fast+
        done
      from IHAPP(2) MNE have IHAPP2': "l' x \<le> l x" 
        by (auto simp add: l_add_use_def split: option.split)
      from wnt_eq_wnt' WN wnt_distinct have "distinct (x'#\<mu>)" by blast
      with MNE IHAPP IHAPP3' obtain \<mu>1' where 
        "\<mu>=\<mu>1'@x#\<mu>2" "set \<mu>1' \<subseteq> dom l" 
        "\<forall>x'\<in>set \<mu>1'. l x' \<le> l x \<and> 
          (case l x' of None \<Rightarrow> True | Some lx' \<Rightarrow> x \<notin> lx' \<and> x'\<notin>lx')"
        by (cases \<mu>1) auto
      hence ?thesis by blast
    } moreover {
      assume A: "x'\<notin>dom l'"
      with NNOSPAWN.prems(2) HTFMT have [simp]: "l=l'" 
        by (auto split: split_if_asm)
      from NNOSPAWN.hyps[OF WN HTFMT NNOSPAWN.prems(3)[simplified]] 
      obtain \<mu>1 \<mu>2 where IHAPP: "x'#\<mu> = \<mu>1@x#\<mu>2" "set \<mu>1 \<subseteq> dom l'" 
        by blast
      with MNE have "x'\<in>dom l'" by (cases \<mu>1) auto
      with A have False ..
    } ultimately show ?thesis by blast
  next
    case (LRel x') [simp]
    from NNOSPAWN.prems(1) obtain \<mu>' where WN: "\<mu>=x'#\<mu>'" "wn_t' t \<mu>'" 
      by auto
    from NNOSPAWN.prems(2) obtain l' u' where 
      HTFMT: "as t = Some (l',u',e)" and 
      [simp]: "l=l'(x'\<mapsto>{})" "u=u'"
      by (auto split: option.split_asm prod.split_asm list.split_asm)
    {
      assume "x=x'"
      with WN(1) have "\<mu>=[]@x#\<mu>'" "set [] \<subseteq> dom l" 
                      "(\<forall>x'\<in>set []. l x' \<le> l x \<and> 
                         (case l x' of None \<Rightarrow> True | Some lx' \<Rightarrow> x \<notin> lx' \<and> x'\<notin>lx'))"
        by auto
      hence ?thesis by blast
    } moreover {
      assume MNE: "x\<noteq>x'"
      with NNOSPAWN.prems(3) have MIDL': "x\<in>dom l'" 
        by (auto simp add: l_add_use_def split: option.split_asm)
      with NNOSPAWN.hyps[OF WN(2) HTFMT] obtain \<mu>1 \<mu>2 where 
        IHAPP: "\<mu>'=\<mu>1@x#\<mu>2" "set \<mu>1 \<subseteq> dom l'" 
               "(\<forall>x'\<in>set \<mu>1. l' x' \<le> l' x \<and> 
                  (case l' x' of None \<Rightarrow> True | Some lx' \<Rightarrow> x \<notin> lx' \<and> x'\<notin>lx'))" 
        by blast
      with WN(1) have "\<mu>=(x'#\<mu>1)@x#\<mu>2" by simp
      moreover from IHAPP(2) NNOSPAWN.prems(3) have 
        "set (x'#\<mu>1) \<subseteq> dom l" 
        by auto
      moreover from IHAPP(3) MNE MIDL' have 
        "(\<forall>x'\<in>set (x'#\<mu>1). l x' \<le> l x \<and> 
          (case l x' of None \<Rightarrow> True | Some lx' \<Rightarrow> x \<notin> lx' \<and> x'\<notin>lx'))" 
        by (fastsimp simp add: l_add_use_def split: option.split)
      ultimately have ?thesis by blast
    } ultimately show ?thesis by blast
  qed
qed


lemma wn_t_dom_l_stack': "\<lbrakk>wn_t' t \<mu>; as t = Some (l,u,e); x\<in>dom l\<rbrakk> \<Longrightarrow> 
  \<exists>\<mu>1 \<mu>2. \<mu>=\<mu>1@x#\<mu>2 \<and> set \<mu>1 \<subseteq> dom l \<and> 
          (\<forall>x'\<in>set \<mu>1. l x' \<le> l x \<and> x\<notin>the (l x') \<and> x'\<notin>the (l x'))"
  apply (drule (2) wn_t_dom_l_stack)
  apply (elim exE)
  apply (rule_tac x=\<mu>1 in exI)
  apply (rule_tac x=\<mu>2 in exI)
  apply (force)
  done

subsection "Soundness of the Consistency Condition"

context LDPN
begin

  text {*
    The consistency condition for acquisition structures is sound, i.e. if a hedge 
    @{text h} is schedulable with initial locks @{text X}, and is well-nested 
    w.r.t. a lock stack list @{text \<mu>} containing the locks from @{text X}, then 
    the acquisition structure of @{text h} is consistent w.r.t. @{text X}.
    *}
  theorem acqh_sound: 
    "\<lbrakk> lsched h X w; wn_h h \<mu>; X=locks_\<mu> \<mu> \<rbrakk> \<Longrightarrow> 
       \<exists>l u e. ash h = Some (l,u,e) \<and> cons_as (l,u,e) (locks_\<mu> \<mu>)"
    -- {*
      The proof works by induction over the schedule, in each induction step
      prepending a step to teh schedele. 

      For steps that have perform operation on locks, the proof is straightforward.

      If the first step of the execution is a release of a lock,
      the acquisition history of the new hedge (with prepended release step at one tree)
      remains consistent. Acyclicity is preserved, as the release-step is 
      the first step of the execution. Consistency w.r.t. used locks is also preserved.

      If the first step of the execution is an acquisition step, we further have to 
      distinguish whether it is a usage or a final acquisition.
    *}
  proof (induct arbitrary: \<mu> rule: lsched.induct)
    case lsched_final thus ?case by (auto simp add: ash_final)
  next
    case (lsched_spawn h1 ts t h2 Xh w X lab \<mu>)
    note [simp] = lsched_spawn.prems(2)
    from lsched_spawn.prems obtain nlab where [simp]: "lab=LNone nlab" 
      by (auto elim: wn_h_spawn_imp_LNoneE)
    from lsched_spawn.hyps(3) have [simp]: "Xh=X" by auto
    from wn_h_preserve_spawn[OF _ lsched_spawn.prems(1), of X, simplified] 
    obtain \<mu>' where [simp]: "locks_\<mu> \<mu> = locks_\<mu> \<mu>'" "wn_h (h1@ts#t#h2) \<mu>'" 
      by blast
    from lsched_spawn.hyps(2)[of \<mu>', simplified] obtain l u e where 
      "ash (h1@ts#t#h2) = Some (l,u,e)" "cons_as (l,u,e) (locks_\<mu> \<mu>)" 
      by auto
    moreover hence "ash (h1@NSPAWN lab ts t#h2) = Some (l,u,e)" by simp
    ultimately show ?case by auto
  next
    case (lsched_nospawn h1 t h2 Xh w X lab \<mu>) note lsched_nospawn.prems(2)[simp]
    from wn_h_split_nospawn[OF lsched_nospawn.hyps(3)[simplified] 
         lsched_nospawn.prems(1)] obtain \<mu>1 xs \<mu>2 xsh where
      [simp]: "\<mu> = \<mu>1 @ xs # \<mu>2" "Xh = locks_\<mu> \<mu>1 \<union> set xsh \<union> locks_\<mu> \<mu>2" and 
        LVX: "lock_valid_xs lab xs xsh" and 
        WNSPLIT: "wn_t' (NNOSPAWN lab t) xs" "wn_t' t xsh" 
                 "wn_h h1 \<mu>1" "wn_h h2 \<mu>2" and
        LDIST:  "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> set xsh = {}" 
                "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" "locks_\<mu> \<mu>2 \<inter> set xs = {}" 
                "locks_\<mu> \<mu>2 \<inter> set xsh = {}" and
        WNH: "wn_h (h1 @ t # h2) (\<mu>1 @ xsh # \<mu>2)" 
      .
    have WNHR: "wn_h (h1@h2) (\<mu>1@\<mu>2)" using WNSPLIT LDIST 
      by (auto intro: wn_h_appendI)

    from lsched_nospawn.hyps(2)[OF WNH] obtain l u e where 
      IHAPP: "ash h1 \<parallel> as t \<parallel> ash h2 = Some (l,u,e)" 
             "cons_as (l,u,e) (locks_\<mu> \<mu>1 \<union> set xsh \<union> locks_\<mu> \<mu>2)" and 
      IHAPP': "ash (h1 @ t # h2) = Some (l, u, e)"
      by (auto simp add: Un_ac)
    then obtain lt ut et l2 u2 e2 where 
      [simp]: "as t = Some (lt,ut,et)" "(ash h1 \<parallel> ash h2) = Some (l2,u2,e2)" 
              "l=lt++l2" "u=ut\<union>u2" "e=et++e2" and 
        ASS: "dom lt \<inter> dom l2 = {}" "dom et \<inter> dom e2 = {}"
    proof -
      from IHAPP have "as t \<parallel> ash h1 \<parallel> ash h2 = Some (l,u,e)" by simp
      thus ?thesis by (erule_tac as_comp_SomeE) (rule that)
    qed
    from wn_h_dom_l_lower_\<mu>[OF WNHR] have 
      DOML2: "dom l2 \<subseteq> locks_\<mu> \<mu>1 \<union> locks_\<mu> \<mu>2"
      by fastsimp
    from wn_t_dom_l_lower_\<mu>[OF WNSPLIT(2)] have 
      DOMLT: "dom lt \<subseteq> set xsh" 
      by fastsimp
    have DOMDISJ: "dom lt \<inter> dom l2 = {}"
    proof -
      from LDIST have "set xsh \<inter> (locks_\<mu> \<mu>1 \<union> locks_\<mu> \<mu>2) = {}" by blast
      with DOMLT DOML2 show ?thesis by blast
    qed
    show ?case proof (cases lab)
      case (LNone nlab)[simp] from LVX have [simp]: "set xsh = set xs" 
        by (auto elim: lock_valid_xs.cases)
      from IHAPP show ?thesis by auto
    next
      case (LRel x)[simp]
      from LVX have [simp]: "xs=x#xsh" by (auto elim: lock_valid_xs.cases)
      have "ash (h1@(NNOSPAWN lab t)#h2) = 
              as (NNOSPAWN lab t) \<parallel> Some (l2,u2,e2)"
        apply (simp del: LRel)
        apply (subst as_comp_acz.assoc[symmetric])
        by (simp)
      also from IHAPP have "as (NNOSPAWN lab t) = Some (lt(x\<mapsto>{}),ut,et)"
        by simp
      hence "as (NNOSPAWN lab t) \<parallel> Some (l2,u2,e2) = Some (l(x\<mapsto>{}),u,e)" 
        using ASS DOML2 LDIST by (auto simp add: map_add_comm)
      finally have 
        G1: "ash (h1@(NNOSPAWN lab t)#h2) = Some (l(x\<mapsto>{}),u,e)" .
      moreover from IHAPP(2) have G2: "cons_as (l(x\<mapsto>{}),u,e) (locks_\<mu> \<mu>)" 
        by simp (blast intro: cons_h_update[where X="{}", simplified])
      ultimately show ?thesis by blast
    next
      case (LAcq x)[simp]
      from LVX have 
        [simp]: "xsh=x#xs" and 
          XNIXS: "x\<notin>set xs" 
        by (auto elim: lock_valid_xs.cases)
      from DOML2 have XNIDL2: "x\<notin>dom l2" using LDIST by auto
      show ?thesis proof (cases "x\<in>dom lt")
        case True -- "The first step enters a lock that is left again, 
          thus converting an initial release to a use step"
        -- "The consistency of the acquisition structure is preserved, as a use-step 
          of a lock is added that is not initially released (any more)"
        have "ash (h1@(NNOSPAWN lab t)#h2) = 
                as (NNOSPAWN lab t) \<parallel> Some (l2,u2,e2)" 
          apply (simp del: LAcq)
          apply (subst as_comp_acz.assoc[symmetric])
          by (simp)
        also from True have 
          "as (NNOSPAWN lab t) = 
             Some (l_add_use (l_remove lt x) {x},insert x ut,et)"
          by simp
        hence "as (NNOSPAWN lab t) \<parallel> Some (l2,u2,e2) = 
                Some (l2 ++ l_add_use (l_remove lt x) {x},insert x u,e)" 
          using ASS DOML2 LDIST 
          by (auto simp add: map_add_comm)
        finally have G1: "ash (h1@(NNOSPAWN lab t)#h2) = 
                            Some (l2 ++ l_add_use (l_remove lt x) {x},insert x u,e)" .
        moreover 
        have G2: "cons_as (l2 ++ l_add_use (l_remove lt x) {x},insert x u,e) 
                          (locks_\<mu> \<mu>)"
        proof -
          from IHAPP(2) have "cons_h (l2 ++ l_add_use (l_remove lt x) {x})" 
            using cons_h_add_remove_partial[OF XNIDL2, of lt] 
            by (simp add: map_add_comm[OF DOMDISJ])
          moreover have 
            "insert x u \<inter> 
              (locks_\<mu> \<mu> - dom (l2 ++ l_add_use (l_remove lt x) {x})) = {}" 
            using XNIXS LDIST[simplified] IHAPP(2) by simp blast
          moreover have 
            "dom e \<inter> (locks_\<mu> \<mu> - dom (l2 ++ l_add_use (l_remove lt x) {x})) = {}" 
            using XNIXS LDIST[simplified] IHAPP(2) by simp blast
          moreover from IHAPP(2) have "cons_h e" by simp
          ultimately show ?thesis by simp
        qed
        ultimately show ?thesis by blast
      next
        case False -- "The first step finally enters a lock"
        from False XNIDL2 IHAPP(2) have XNIUE: "x\<notin>u" "x\<notin>dom e" by auto
        -- "The consistency of the acquisition structure is preserved, as no cycles
            are added by insertion of the final acquisition."
        have "ash (h1@(NNOSPAWN lab t)#h2) = 
                as (NNOSPAWN lab t) \<parallel> Some (l2,u2,e2)"
          apply (simp del: LAcq)
          apply (subst as_comp_acz.assoc[symmetric])
          by (simp)
        also from False have "as (NNOSPAWN lab t) = Some (lt,ut,et(x\<mapsto>ut))" 
          using XNIUE by simp
        hence "as (NNOSPAWN lab t) \<parallel> Some (l2,u2,e2) = Some (l,u,e(x\<mapsto>ut))" 
          using ASS XNIUE 
          by (auto simp add: map_add_comm)
        finally have 
          G1: "ash (h1@(NNOSPAWN lab t)#h2) = Some (l,u,e(x\<mapsto>ut))" .
        moreover
        from cons_h_update2[of e x ut] IHAPP(2) ash_le_u[OF IHAPP'] XNIUE 
        have "cons_h (e(x\<mapsto>ut))" by auto
        with IHAPP(2) have "cons_as (l,u,e(x\<mapsto>ut)) (locks_\<mu> \<mu>)" 
          using LDIST XNIXS by simp blast
        ultimately show ?thesis by blast
      qed
    qed
  qed
end


subsection "Precision of the Consistency Condition"

subsubsection "Custom Size Function"
text {*
  In the following we construct a custom size function for hedges that is suited to 
  do induction over hedges. This size function decreases on any step done on the hedge.
*}
fun list_size' where
  "list_size' f [] = (0::nat)" |
  "list_size' f (a#l) = f a + list_size' f l"

fun size_t where
  "size_t (NLEAF \<pi>) = Suc 0" |
  "size_t (NNOSPAWN lab t) = Suc (size_t t)" |
  "size_t (NSPAWN lab ts t) = Suc (size_t ts + size_t t)"

lemma list_size'_conc[simp]: "list_size' f (a@b) = list_size' f a + list_size' f b"
  by (induct a) auto

abbreviation hedge_size :: "('P,'\<Gamma>,'L,'X) lex_hedge \<Rightarrow> nat" where 
  "hedge_size h == list_size' size_t h"

lemma hedge_size_zero[simp]: "hedge_size h = 0 \<longleftrightarrow> h=[]"
  apply (cases h)
  apply auto
  apply (case_tac a)
  apply simp_all
done



text {*
  This function checks whether a lock is released in the current execution tree, and 
  returns the set of locks that are acquired before this lock is released. Note that 
  this function ignores the lock-effect of labels of spawn-nodes, as we assume that
  spawn-nodes have no lock-operation.
  *}
fun closing :: "'X \<Rightarrow> ('P,'\<Gamma>,'L,'X) lex_tree \<Rightarrow> 'X set option" where
  "closing x (NLEAF \<pi>) = None" |
  "closing x (NSPAWN lab ts t) = closing x t" |
  "closing x (NNOSPAWN (LNone nlab) t) = closing x t" |
  "closing x (NNOSPAWN (LAcq x') t) = (
    case closing x t of None   \<Rightarrow> None | 
                        Some X \<Rightarrow> Some (insert x' X)
  )" |
  "closing x (NNOSPAWN (LRel x') t) = (if x=x' then Some {} else closing x t)"

text {*
  Function that checks whether a tree starts with the acquisition of a lock that is
  used (i.e. not finally acquired) and returns all the locks that are used from
  the acquisition to to the release of that lock:
  *}
fun closing' where
  "closing' (NNOSPAWN (LAcq x) t) = closing x t" |
  "closing' _ = None"

text {*
  The following functions define the set of locks that are acquired at the roots of a 
  tree/hedge. This function is used in the case of the precision proof, where all the 
  roots of the hedge are either leafs or final acquisitions.
*}
fun rootlocks_t where
  "rootlocks_t (NNOSPAWN (LAcq x) t) = {x}" |
  "rootlocks_t _ = {}"
fun rootlocks where
  "rootlocks [] = {}" |
  "rootlocks (t # h) = rootlocks_t t \<union> rootlocks h"

lemma rootlocks_conc[simp]: "rootlocks (h1@h2) = rootlocks h1 \<union> rootlocks h2"
  by (induct h1) auto

lemma rootlocks_split: 
  "\<lbrakk> x\<in>rootlocks h; !!h1 t h2. h=h1@NNOSPAWN (LAcq x) t#h2 \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
proof (induct h arbitrary: P)
  case Nil thus ?case by simp
next
  case (Cons tp h) from Cons.prems(1)[simplified] show ?case proof
    assume "x\<in>rootlocks_t tp" 
    with Cons.prems(2)[of "[]", simplified] show ?thesis 
      by (cases tp rule: rootlocks_t.cases) auto
  next
    assume A: "x\<in>rootlocks h" from Cons.hyps[OF A] obtain h1 t h2 where 
      "h = h1 @ NNOSPAWN (LAcq x) t # h2" .
    hence "tp#h = (tp#h1)@NNOSPAWN (LAcq x) t # h2" by simp
    thus ?thesis by (blast intro!: Cons.prems(2))
  qed
qed

text {*
  If a lock @{text x} is closed (before it is acquired), the value of the release 
  history for @{text x} is precisely the set of used locks before @{text x} is closed.
  Closing @{text x} before it is acquired is expressed by well-nestedness w.r.t. a 
  lock-stack that contains @{text x}.
*}
lemma closing_dom_l: 
  "\<lbrakk> wn_t' t (xs1@x#xs2); closing x t = Some Xu; as t = Some (l,u,e) \<rbrakk> \<Longrightarrow> 
     l x = Some Xu"
proof (induct t arbitrary: xs1 l u e Xu)
  case NLEAF thus ?case by auto
next
  case (NSPAWN lab ts t) 
  then obtain nlab where [simp]: "lab=LNone nlab" by (cases lab) auto
  from NSPAWN show ?case by (fastsimp elim: as_comp_SomeE dest: wn_dom_l_empty)
next
  case (NNOSPAWN lab t) show ?case proof (cases lab)
    case (LNone nlab) with NNOSPAWN show ?thesis by auto
  next
    case (LAcq x')[simp]
    from NNOSPAWN.prems obtain Xu' where 
      HP1: "wn_t' t ((x'#xs1)@x#xs2)" "closing x t = Some Xu'" and 
      [simp]: "Xu=insert x' Xu'" 
      by (auto split: option.split_asm)
    from NNOSPAWN.prems obtain l' u' e' where 
      HP2: "as t = Some (l',u',e')" 
      by (auto split: eahl_splits)
    from NNOSPAWN.hyps[OF HP1 HP2] have IHAPP: "l' x = Some Xu'" .
    (* TODO: Make this a corollary of wn_t_dom_l_stack and 
      give it an own name ! *)
    from wn_t_dom_l_stack[OF HP1(1) HP2, of x] 
         IHAPP distinct_match[OF wnt_distinct'[OF HP1(1)]] have 
      "set (x'#xs1) \<subseteq> dom l'" 
      by fastsimp
    hence X'IDL': "x'\<in>dom l'" by simp
    with NNOSPAWN.prems(3) HP2 IHAPP 
    have "l = l_add_use (l_remove l' x') {x'}" by (simp split: eahl_splits)
    moreover from wnt_distinct'[OF HP1(1)] have MNE: "x'\<noteq>x" by (auto)
    ultimately show "l x = Some Xu" using IHAPP by (auto simp add: l_add_use_def)
  next
    case (LRel x')[simp]
    show ?thesis proof (cases "x=x'")
      case True with NNOSPAWN.prems have "l x = Some {}" "Xu={}" 
        by (auto split: eahl_splits)
      thus ?thesis by blast
    next
      case False with NNOSPAWN.prems obtain xs1' where 
        [simp]: "xs1=x'#xs1'" and 
          HP1: "wn_t' t (xs1'@x#xs2)" "closing x t = Some Xu" 
        by (cases xs1) auto
      from NNOSPAWN.prems obtain l' u' e' where 
        HP2: "as t = Some (l',u',e')" and 
        [simp]: "l=l'(x'\<mapsto>{})" 
        by (auto split: eahl_splits)
      from NNOSPAWN.hyps[OF HP1 HP2(1)] have "l' x = Some Xu" .
      with False show "l x = Some Xu" by auto
    qed
  qed
qed

text {*
  A lock must not be used before it is closed.
*}
lemma wn_closing_ni: "\<lbrakk>wn_t' t (\<mu>1@x#\<mu>2); closing x t = Some Xu\<rbrakk> \<Longrightarrow> x\<notin>Xu"
proof (induct t arbitrary: \<mu>1 Xu)
  case NLEAF thus ?case by auto
next
  case (NSPAWN lab ts t) 
  then obtain nlab where [simp]: "lab=LNone nlab" by (cases lab) auto
  from NSPAWN show ?case by auto
next
  case (NNOSPAWN lab t)
  show ?case proof (cases lab)
    case (LNone nlab) thus ?thesis using NNOSPAWN by auto
  next
    case (LAcq x')[simp]
    from NNOSPAWN.prems(1) have WN: "wn_t' t ((x'#\<mu>1)@x#\<mu>2)" by auto
    from NNOSPAWN.prems(2) obtain Xu' where 
      CL: "closing x t = Some Xu'" "Xu = insert x' Xu'" 
      by (auto split: option.split_asm)
    from NNOSPAWN.hyps[OF WN CL(1)] have "x\<notin>Xu'" .
    moreover from wnt_distinct'[OF WN] have "x'\<noteq>x" by auto
    ultimately show ?thesis by (auto simp add: CL(2))
  next
    case (LRel x') 
    thus ?thesis 
      using NNOSPAWN by (cases \<mu>1) (auto split: split_if_asm)
  qed
qed
   
text {*
  This lemma gives porperties of the acquisition structure after an acquisition step 
  of a lock usage. It is used in the case when there is a tree starting with a usage,
  to reason about the acquisition structure after the root node of this tree has been
  scheduled.
*} 
lemma wn_closing_as_fmt:
  assumes A: "wn_t' (NNOSPAWN (LAcq x) t) \<mu>" 
             "as (NNOSPAWN (LAcq x) t) = Some (l,u,e)" 
             "closing x t = Some Xu" 
  assumes C: "!!l' u'. \<lbrakk> as t = Some (l',u',e); l' \<le> l(x\<mapsto>Xu); 
                         u=insert x u'; dom l' = insert x (dom l)
                       \<rbrakk> \<Longrightarrow> P"
  shows P
proof -
  from A(1) have WN: "wn_t' t ([]@x#\<mu>)" by auto
  from A(2) obtain l' u' e' where AS': "as t = Some (l',u',e')" 
    by (auto split: eahl_splits)
  from closing_dom_l[OF WN A(3) AS'] have L'X: "l' x = Some Xu" .
  with A(2) AS' have 
    LFMT: "l = l_add_use (l_remove l' x) {x}" and 
    [simp]: "u=insert x u'" "e'=e" 
    by (auto split: eahl_splits)
  from LFMT L'X have G2: "l' \<le> l(x\<mapsto>Xu)" 
    by (rule_tac le_funI) (auto simp add: l_add_use_def split: option.split)
  from LFMT L'X have G3: "dom l' = insert x (dom l)" by auto
  from C[OF _ G2 _ G3] show P by (simp add: AS')
qed


text {*
  A lock that occurs in the release history is closed in the execution tree, using 
  the locks as described in the RH.
  *}
lemma dom_l_closing: 
  "\<lbrakk> as t = Some (l,u,e); wn_t' t \<mu>; l x = Some Xu \<rbrakk> \<Longrightarrow> closing x t = Some Xu"
proof (induct t \<mu> arbitrary: l u e Xu rule: wn_t'.induct)
  case (1 ms) thus ?case by auto
next
  case 2 thus ?case by force
next
  case 3 thus ?case by (fastsimp elim!: as_comp_SomeE dest!: wn_dom_l_empty)
next
  case (4 xa t \<mu>) note C=this
  (* TODO: This proof is a bit complicated. Can we find some nice lemma that 
    simplifies the usage of wn_t_dom_l_stack' here, i.e. somehow combines the 
    assumption of values l',u',e' for the AS of t and later derivation of properties 
    of these values ?*)
  from C(3) have WN: "wn_t' t (xa#\<mu>)" by auto
  from C(2) obtain l' u' e' where AS: "as t = Some (l',u',e')" 
    by (auto split: eahl_splits)
  from C(2,4) have XNE: "xa\<noteq>x" by (auto split: eahl_splits simp add: l_add_use_def)
  with AS C(2,4) obtain Xu' where P: "l' x = Some Xu'" 
    by (auto split: eahl_splits simp add: l_add_use_def)
  from C(1)[OF AS WN, OF P] have IHAPP: "closing x t = Some Xu'" .
  from wn_t_dom_l_stack'[OF WN AS, of x] P obtain \<mu>1 \<mu>2 where 
    "xa#\<mu> = \<mu>1@x#\<mu>2" "set \<mu>1 \<subseteq> dom l'" 
    by blast
  with XNE have "xa\<in>dom l'" by (cases \<mu>1) auto
  with AS C(2,4) have "l = l_add_use (l_remove l' xa) {xa}" 
    by (auto split: eahl_splits)
  with XNE P C(4) have "Xu = (insert xa Xu')" by (auto simp add: l_add_use_def)
  moreover from IHAPP 
  have "closing x (NNOSPAWN (LAcq xa) t) = Some (insert xa Xu')"
    by auto
  ultimately show ?case by blast
next  
  case 5 thus ?case by (fastsimp split: eahl_splits)
qed auto


text {*
  If a tree starts with a final acquisition of @{term x}, its release history is empty
  and the acquisition history of @{term x} contains all the used locks.

  With Lemma~@{thm [source] as_ran_e_le_u} we then also have that the ranges of the 
  acquisition histories contain precisely the used locks.
*}
lemma ncl_as_fmt_single: 
  assumes A: "wn_t' (NNOSPAWN (LAcq x) t) \<mu>" 
             "closing' (NNOSPAWN (LAcq x) t) = None" 
             "as (NNOSPAWN (LAcq x) t) = Some (l,u,e)" 
  shows "u=\<Union>ran e" "l=empty" "e x = Some u"
proof -
  from A(1) have WN: "wn_t' t (x#\<mu>)" by auto
  from A(2) have NC: "closing x t = None" by auto
  from A(3) obtain l' u' e' where AS: "as t = Some (l',u',e')" 
    by (auto split: eahl_splits)
  from dom_l_closing[OF AS WN] NC have XNIDL': "\<not>x\<in>dom l'" by auto
  with AS A(3) have 
    EFMT: "e=e'(x\<mapsto>u)" "x\<notin>dom e'" and 
    [simp]: "l=l'" 
    by (auto split: eahl_splits)
  from EFMT(1) show "e x = Some u" by auto
  with EFMT have "u \<subseteq> \<Union>ran e" by auto
  with as_ran_e_le_u[OF A(3)] show "u=\<Union>ran e" by simp
  {
    fix x'
    assume CONTR: "x'\<in>dom l'"
    with XNIDL' have XNE: "x'\<noteq>x" by auto
    from wn_t_dom_l_stack'[OF WN AS CONTR] obtain \<mu>1 \<mu>2 where 
      DS: "x#\<mu> = \<mu>1@x'#\<mu>2" "set \<mu>1 \<subseteq> dom l'" 
      by blast
    with XNE have "x\<in>dom l'" by (cases \<mu>1) auto
    with XNIDL' have False ..
  } thus "l=empty" 
    by (auto simp add: dom_empty_simp[symmetric] simp del: dom_empty_simp)
qed
  
text {*
  This lemma describes properties of the acquisition structure of a tree after a final
  acquisition has been scheduled.
*}
lemma ncl_as_fmt_single': 
  assumes A: "wn_t' (NNOSPAWN (LAcq x) t) \<mu>" 
             "closing' (NNOSPAWN (LAcq x) t) = None" 
             "as (NNOSPAWN (LAcq x) t) = Some (l,u,e)" 
  assumes C: "!!e'. \<lbrakk> as t = Some (empty, u, e'); 
                      u=\<Union>ran e; l=empty; 
                      e = e'(x\<mapsto>u); x\<notin>dom e' 
                    \<rbrakk> \<Longrightarrow> P"
  shows P
proof -
  from A(1) have WN: "wn_t' t (x#\<mu>)" by auto
  from A(2) have NC: "closing x t = None" by auto
  from A(3) obtain l' u' e' where AS: "as t = Some (l',u',e')" 
    by (auto split: eahl_splits)
  from dom_l_closing[OF AS WN] NC have XNIDL': "\<not>x\<in>dom l'" by auto
  with AS A(3) have 
    EFMT: "e=e'(x\<mapsto>u)" "x\<notin>dom e'" and 
    [simp]: "l'=l" "u'=u" 
    by (auto split: eahl_splits)
  with EFMT have "u \<subseteq> \<Union>ran e" by auto
  with as_ran_e_le_u[OF A(3)] have UFMT: "u=\<Union>ran e" by simp
  {
    fix x'
    assume CONTR: "x'\<in>dom l'"
    with XNIDL' have XNE: "x'\<noteq>x" by auto
    from wn_t_dom_l_stack'[OF WN AS CONTR] obtain \<mu>1 \<mu>2 where 
      DS: "x#\<mu> = \<mu>1@x'#\<mu>2" "set \<mu>1 \<subseteq> dom l'" 
      by blast
    with XNE have "x\<in>dom l'" by (cases \<mu>1) auto
    with XNIDL' have False ..
  } hence LFMT[simp]: "l=empty" 
    by (auto simp add: dom_empty_simp[symmetric] simp del: dom_empty_simp)
  from C[OF _ UFMT LFMT EFMT] AS show P by simp
qed


text {*
  The acquisition structure of a hedge whose trees start with final acquisitions or 
  are leafs has a special structure:
  \begin{itemize}
    \item The release history is empty.
    \item The ranges of the acquisition histories contain precisely the used locks.
    \item The acquisition histories for the locks at the roots of the hedge contain 
          precisely the used locks.
    \item The acquisistion histories are defined for the locks at the roots of the 
          hedge.
  \end{itemize}

  The first proposition follows because an initial release cannot come after a final 
  acquisition due to well-nestedness. The second and third propositions follow as the 
  roots of the hedge preceed every other node in the hedge. The forth proposition 
  follows directly from the assumption that every root node that acquired a lock is a 
  final acquisistion.
*}
lemma ncl_as_fmt: 
  "\<lbrakk> 
    wn_h h \<mu>; ash h = Some (l,u,e); 
    !!Q t. \<lbrakk> t\<in>set h; !!x t'. t=NNOSPAWN (LAcq x) t' \<Longrightarrow> Q; 
             !!p w. t=NLEAF (p,w) \<Longrightarrow> Q 
           \<rbrakk> \<Longrightarrow> Q; 
    \<forall>t\<in>set h. closing' t = None
  \<rbrakk> \<Longrightarrow> l=empty \<and> u=\<Union>ran e \<and> 
        \<Union>ran (e |` rootlocks h) = \<Union>ran e \<and> 
        rootlocks h \<subseteq> dom e"
proof (induct h arbitrary: \<mu> l u e)
  case Nil thus ?case by auto
next
  case (Cons t h) 
  from Cons.prems(1) obtain xs \<mu>' where 
    [simp]: "\<mu>=xs#\<mu>'" and 
      WN_SPLIT: "wn_t' t xs" "wn_h h \<mu>'" and 
      WN_DISJ: "set xs \<inter> locks_\<mu> \<mu>' = {}" 
    by (auto elim!: wn_h_prepend_h)
  from Cons.prems(2) obtain l1 u1 e1 l2 u2 e2 where 
    [simp]: "l=l1++l2" "u=u1\<union>u2" "e=e1++e2" and 
      AS_SPLIT: "as t = Some (l1,u1,e1)" "ash h = Some (l2,u2,e2)" and
      AS_DISJ: "dom l1 \<inter> dom l2 = {}" "dom e1 \<inter> dom e2 = {}"
    by (fastsimp elim!: as_comp_SomeE)
  have "l2=empty \<and> u2=\<Union>ran e2 \<and> 
        \<Union>ran (e2 |` rootlocks h) = \<Union>ran e2 \<and> rootlocks h \<subseteq> dom e2" 
    apply (rule_tac Cons.hyps[OF WN_SPLIT(2) AS_SPLIT(2)])
    apply (rule_tac t=t in Cons.prems(3))
    apply auto
    apply (rule_tac Cons.prems(4)[rule_format])
    apply simp
    done
  hence IHAPP: "l2=empty"
               "u2=\<Union>ran e2"
               "\<Union>ran (e2 |` rootlocks h) = \<Union>ran e2"
               "rootlocks h \<subseteq> dom e2"
    by auto (* TODO: Isn't there an automatic method to split conjunctions ? *)
  have "t\<in>set (t#h)" by simp
  thus ?case proof (cases rule: Cons.prems(3)[cases set, case_names acquire leaf])
    case leaf[simp] with AS_SPLIT(1) have [simp]: "l1=empty" "u1={}" "e1=empty" by auto
    from IHAPP show ?thesis by simp
  next
    case (acquire x t')[simp]
    from ncl_as_fmt_single[of x t' xs l1 u1 e1] WN_SPLIT(1) AS_SPLIT(1) 
         Cons.prems(4)[rule_format, of t] have 
      P: "l1=empty" "u1=\<Union>ran e1" "e1 x = Some u1" 
      by auto
    from P IHAPP AS_DISJ have G1: "l=empty \<and> u=\<Union>ran e" by auto
    from P(3) have G2_1: "rootlocks_t t \<subseteq> dom e1" by auto
    from P(2,3) have G3_1: "\<Union>ran (e1 |` rootlocks_t t) = \<Union>ran e1" 
      by (auto simp add: restrict_map_def ran_def)
    from G2_1 IHAPP(4) AS_DISJ have 
      "\<Union>ran ((e1 ++ e2) |` (rootlocks_t t \<union> rootlocks h)) = \<Union>ran e1 \<union> \<Union>ran e2"
      by (rule_tac union_ran_add_aux[OF G3_1 IHAPP(3)]) auto
    hence G3: "\<Union>ran (e|` rootlocks (t#h)) = \<Union>ran e" using AS_DISJ by auto
    show ?thesis using G1 G2_1 IHAPP(4) G3 by auto
  qed
qed



text {*
  This lemma makes explicit the case-distinction along which the precision proof is 
  done. 
  The cases are:
  \begin{description}
    \item[final] All trees are leaf nodes.
    \item[spawn] There is a tree starting with a @{term "NSPAWN x"} - node.
    \item[none] There is a tree starting with a @{term "NNOSPAWN LNone"} - node.
    \item[release] There is a tree starting with a @{term "NNOSPAWN (LRel x)"}-node.
    \item[acquire] All trees start with a @{term "NNOSPAWN (LAcq x)"}-node or are 
                   leafs. At least one tree is no leaf. 
  \end{description}
  *}
lemma h_cases[case_names final spawn none release acquire]: 
  assumes C: 
  "final h \<Longrightarrow> P"
  "!!h1 lab ts t h2. h=h1@NSPAWN lab ts t#h2 \<Longrightarrow> P"
  "!!h1 t nlab h2. h=h1@NNOSPAWN (LNone nlab) t#h2 \<Longrightarrow> P"
  "!!h1 x t h2. h=h1@NNOSPAWN (LRel x) t#h2 \<Longrightarrow> P"
  "\<lbrakk> !!Q t. \<lbrakk> t\<in>set h; !!x t'. t=NNOSPAWN (LAcq x) t' \<Longrightarrow> Q; 
              !!p w. t=NLEAF (p,w) \<Longrightarrow> Q 
            \<rbrakk> \<Longrightarrow> Q; 
     !!Q. \<lbrakk> !!t' x. NNOSPAWN (LAcq x) t' \<in> set h \<Longrightarrow> Q  \<rbrakk> \<Longrightarrow> Q 
   \<rbrakk> \<Longrightarrow> P"
  shows P
proof (cases "h=[]")
  case True with C(1) show P by simp
next
  case False hence "set h \<noteq> {}" by simp
  {
    assume "\<exists>t nlab. NNOSPAWN (LNone nlab) t\<in>set h" 
    with C(3) have P by (blast elim: in_set_list_format)
  } moreover {
    assume "\<exists>t x. NNOSPAWN (LRel x) t\<in>set h" 
    with C(4) have P by (blast elim: in_set_list_format)
  } moreover {
    assume "\<exists>lab ts t. NSPAWN lab ts t\<in>set h" 
    with C(2) have P by (blast elim: in_set_list_format)
  } moreover {
    assume "\<forall>t\<in>set h. \<not>(\<exists>lab t. NNOSPAWN lab t\<in>set h) \<and> 
            \<not> (\<exists>lab ts t. NSPAWN lab ts t\<in>set h)"
    hence "\<forall>t\<in>set h. final_t t" 
      apply safe 
      apply (case_tac t) 
      apply auto 
      done
    with C(1) have P by (auto simp add: list_all_iff)
  } moreover {
    assume A: "\<not>(\<exists>t nlab. NNOSPAWN (LNone nlab) t\<in>set h)" 
              "\<not>(\<exists>t x. NNOSPAWN (LRel x) t\<in>set h)" 
              "\<not>(\<exists>lab ts t. NSPAWN lab ts t\<in>set h)"
              "(\<exists>lab t. NNOSPAWN lab t\<in>set h)"
    hence "(\<exists>t x. NNOSPAWN (LAcq x) t\<in>set h)" 
      apply auto 
      apply (case_tac lab) 
      by auto
    with A(1,2,3) have P apply auto
      apply (rule_tac C(5))
      apply auto
      apply (case_tac ta)
      apply auto
      apply fast
      apply (case_tac L)
      apply auto
      apply fast
      done
  } ultimately show ?thesis by blast
qed

text {*
  This lemma determines the tree within a hedge whose release history contains a 
  specific lock.
  *}
lemma ash_find_l_t[consumes 2]: 
  "\<lbrakk> ash h = Some (l,u,e); x\<in>dom l; 
     !!h1 t h2 l1 u1 e1 l2 u2 e2. \<lbrakk> 
       h=h1@t#h2; l=l1++l2; u=u1\<union>u2; e=e1++e2; 
       as t = Some (l1,u1,e1); ash h1 \<parallel> ash h2 = Some (l2,u2,e2); 
       x\<in>dom l1; dom l1\<inter>dom l2 = {}; dom e1\<inter>dom e2 = {} 
      \<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
proof (induct h arbitrary: l u e P rule: ash.induct)
  case 1 thus ?case by fastsimp
next
  case (2 t h) note C=this
  from as_comp_SomeE[OF C(2)[simplified]] obtain l1 u1 e1 l2 u2 e2 where
    SPLIT_simps[simp]: "l = l1 ++ l2" "u = u1 \<union> u2" "e = e1 ++ e2" and
    SPLIT: "as t = Some (l1, u1, e1)" "ash h = Some (l2, u2, e2)" 
           "dom l1 \<inter> dom l2 = {}" "dom e1 \<inter> dom e2 = {}" 
    .
  from C(3) have "x\<in>dom l1 \<or> x\<in>dom l2" by auto
  moreover {
    assume A: "x\<in>dom l1"
    moreover have "t#h = []@t#h" by simp
    ultimately have ?case 
      by (rule_tac C(4)) (assumption, (simp add: SPLIT)+) (* Strange: simp_all does 
        not give the expected result here \<dots> *)
  } moreover {
    assume A: "x\<in>dom l2"
    from C(1)[OF SPLIT(2) A] obtain h1 tt h2 l21 u21 e21 l22 u22 e22 where 
      IHAPP_simp[simp]: "h = h1 @ tt # h2"   "l2=l21++l22" 
                        "u2=u21\<union>u22"         "e2=e21++e22" and 
      IHAPP: "as tt = Some (l21, u21, e21)" 
             "ash h1 \<parallel> ash h2 = Some (l22,u22,e22)" 
             "x\<in>dom l21"
             "dom l21\<inter>dom l22={}" 
             "dom e21\<inter>dom e22={}" 
      .
    from SPLIT IHAPP have 
      DS: "dom l1 \<inter> dom l21 = {}" "dom e1 \<inter> dom e21 = {}" 
      by auto
    have "t#h=(t#h1)@tt#h2"   "l = l21 ++ (l1++l22)" 
         "u=u21\<union>(u1\<union>u22)"    "e=e21 ++ (e1++e22)"
      by (auto simp add: map_add_comm[OF DS(1)] map_add_comm[OF DS(2)])
    moreover have "ash (t#h1) \<parallel> ash h2 = Some (l1++l22,u1\<union>u22,e1++e22)" 
    proof -
      have "ash (t#h1) \<parallel> ash h2 = as t \<parallel> (ash h1 \<parallel> ash h2)" by (simp)
      also have "\<dots> = as_comp (l1,u1,e1) (l22,u22,e22)" 
        by (simp add: IHAPP(2) SPLIT(1))
      also have "\<dots> = Some (l1++l22,u1\<union>u22,e1++e22)" 
        using SPLIT IHAPP by auto
      finally show ?thesis .
    qed
    ultimately have ?case using SPLIT(3,4) IHAPP(1,3,4,5) 
      by (rule_tac C(4)) (assumption+, auto)
  } ultimately show ?case by blast
qed


text {*
  This lemma determines the tree within a hedge whose acquisition history contains a 
  specific lock.
  *}
lemma ash_find_e_t[consumes 2]: 
  "\<lbrakk> ash h = Some (l,u,e); x\<in>dom e; 
     !!h1 t h2 l1 u1 e1 l2 u2 e2. \<lbrakk> 
       h=h1@t#h2; l=l1++l2; u=u1\<union>u2; e=e1++e2; 
       as t = Some (l1,u1,e1); ash h1 \<parallel> ash h2 = Some (l2,u2,e2); 
       x\<in>dom e1; dom l1\<inter>dom l2 = {}; dom e1\<inter>dom e2 = {} 
     \<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
proof (induct h arbitrary: l u e P rule: ash.induct)
  case 1 thus ?case by fastsimp
next
  case (2 t h) note C=this
  from as_comp_SomeE[OF C(2)[simplified]] obtain l1 u1 e1 l2 u2 e2 where
    SPLIT_simps[simp]: "l = l1 ++ l2" "u = u1 \<union> u2" "e = e1 ++ e2" and
    SPLIT: "as t = Some (l1, u1, e1)" "ash h = Some (l2, u2, e2)" 
           "dom l1 \<inter> dom l2 = {}" "dom e1 \<inter> dom e2 = {}" 
    .
  from C(3) have "x\<in>dom e1 \<or> x\<in>dom e2" by auto
  moreover {
    assume A: "x\<in>dom e1"
    moreover have "t#h = []@t#h" by simp
    ultimately have ?case by (rule_tac C(4)) (assumption, (simp add: SPLIT)+) (* 
      Strange: simp_all gives not the expected result here \<dots> *)
  } moreover {
    assume A: "x\<in>dom e2"
    from C(1)[OF SPLIT(2) A] obtain h1 tt h2 l21 u21 e21 l22 u22 e22 where 
      IHAPP_simp[simp]: "h = h1 @ tt # h2" "l2=l21++l22" 
                        "u2=u21\<union>u22" "e2=e21++e22" and 
      IHAPP: "as tt = Some (l21, u21, e21)" 
             "ash h1 \<parallel> ash h2 = Some (l22,u22,e22)" 
             "x\<in>dom e21"
             "dom l21\<inter>dom l22={}" 
             "dom e21\<inter>dom e22={}" 
      .
    from SPLIT IHAPP have 
      DS: "dom l1 \<inter> dom l21 = {}" "dom e1 \<inter> dom e21 = {}" 
      by auto
    have "t#h=(t#h1)@tt#h2" "l = l21 ++ (l1++l22)" 
         "u=u21\<union>(u1\<union>u22)" "e=e21 ++ (e1++e22)"
      by (auto simp add: map_add_comm[OF DS(1)] map_add_comm[OF DS(2)])
    moreover have "ash (t#h1) \<parallel> ash h2 = Some (l1++l22,u1\<union>u22,e1++e22)" proof -
      have "ash (t#h1) \<parallel> ash h2 = as t \<parallel> (ash h1 \<parallel> ash h2)" by (simp)
      also have "\<dots> = as_comp (l1,u1,e1) (l22,u22,e22)" 
        by (simp add: IHAPP(2) SPLIT(1))
      also have "\<dots> = Some (l1++l22,u1\<union>u22,e1++e22)" 
        using SPLIT IHAPP by auto
      finally show ?thesis .
    qed
    ultimately have ?case using SPLIT(3,4) IHAPP(1,3,4,5) 
      by (rule_tac C(4)) (assumption+, auto)
  } ultimately show ?case by blast
qed

text {*
  Auxilliary lemma to split the acquisistion history of a hedge by some tree in that 
  hedge.
*}
lemma ash_split_aux: 
  assumes AS: "ash h = Some (l,u,e)" and 
          HFMT[simp]: "h=h1@t#h2" and
          C: "!!l1 u1 e1 l2 u2 e2. \<lbrakk> 
                l=l1++l2; u=u1\<union>u2; e=e1++e2; as t = Some (l1,u1,e1); 
                ash h1 \<parallel> ash h2 = Some (l2,u2,e2); 
                dom l1 \<inter> dom l2 = {}; dom e1 \<inter> dom e2 = {} 
              \<rbrakk> \<Longrightarrow> P"
  shows "P"
proof -
  have "as t \<parallel> (ash h1 \<parallel> ash h2) = ash h" by simp
  also note AS
  finally have 1: "as t \<parallel> ash h1 \<parallel> ash h2 = Some (l, u, e)" .
  show P by (rule as_comp_SomeE[OF 1], rule C) assumption+
qed

text {*
  Auxilliary lemma that combines @{thm [source] ash_split_aux} and 
  @{thm [source] wn_h_split_aux}.
*}
lemma wn_ash_split_aux: 
  assumes 
    WN: "wn_h h \<mu>" and 
    AS: "ash h = Some (l,u,e)" and 
    HFMT[simp]: "h=h1@t#h2" and
    C: "!!\<mu>1 xs \<mu>2 l1 u1 e1 l2 u2 e2. \<lbrakk> 
          \<mu>=\<mu>1@xs#\<mu>2; l=l1++l2; u=u1\<union>u2; e=e1++e2; 
          wn_t' t xs; wn_h h1 \<mu>1; wn_h h2 \<mu>2; 
          as t = Some (l1,u1,e1); ash h1 \<parallel> ash h2 = Some (l2,u2,e2);
          locks_\<mu> \<mu>1 \<inter> set xs = {}; locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}; 
          set xs \<inter> locks_\<mu> \<mu>2 = {}; dom l1 \<inter> dom l2 = {}; dom e1 \<inter> dom e2 = {} 
        \<rbrakk> \<Longrightarrow> P"
  shows P
  apply (rule wn_h_split_aux[OF WN HFMT])
  apply (rule ash_split_aux[OF AS HFMT])
  apply (rule C)
  apply assumption+
  done


context LDPN
begin
  text {*
    Precision of the acqusisition structure construction, i.e. for a well-nested 
    hedge, a consistent acquisistion history implies a schedule.
    *}
  theorem acqh_precise: 
    fixes h::"('P,'\<Gamma>,'L,'X) lex_hedge" 
    assumes A: "ash h=Some (l,u,e)" "cons_as (l,u,e) (locks_\<mu> \<mu>)" "wn_h h \<mu>" 
    shows "\<exists>w. lsched h (locks_\<mu> \<mu>) w"
    -- {*
    The proof is done by induction on the size of the hedge.

    Given a non-empty hedge, it constructs the first step of the schedule and shows 
    that the acquisistion structure remains consistent.

    It considers the following cases:
    \begin{itemize}
      \item If the hedge contains a root that has no effect on locks, this root is 
        scheduled. Those steps can always be scheduled, as the acquisition structure 
        and the set of acquired locks do not change.
      \item If the hedge contains a root that initially releases a lock @{term x}, it 
        is scheduled. A release can always be scheduled, as it cannot block. 
        The new acquisition structure remains consistent: The acqusisition history 
        is unchanged, the release history decreases (the lock @{term x} is removed). 
        Consistency is preserved, as the lock @{term x} does not occur in the set of 
        acquired locks any more.
      \item If the hedge contains only roots that are lock acquisitions or leafs, we 
        further distinguish whether some of the roots are usages, or there are only 
        final acquisitions.
      \begin{itemize}
        \item If some of the roots are usages, we can find a usage where the used 
          locks are disjoint from the domain of the release history (Due to 
          acyclicity of the RH). Intuitively, this is a usage where the required locks
          are already released. This usage could be scheduled as a whole, without 
          changing the RH, AH or set of acquired locks, and only decreasing the set of
          used locks. However, we chose another way here and show that scheduling only
          the first acquisition step of the usage also preserves consistency of the 
          AS. We chose this approach in order to not having to formalize the 
          scheduling of a usage. We assume that this simplifies formalization overhead
          (Perhaps at the cost of increased proof complexity).
        \item If all of the roots are leafs or final acquisitions, due to acyclicity 
          of the AH, we can select a final acquisition that acquires a lock that is 
          not used in the rest of the hedge. Scheduling this acquisition preserves 
          consistency of the AS.
      \end{itemize}
    \end{itemize}
      *}
  proof -
    {
      fix h::"('P,'\<Gamma>,'L,'X) lex_hedge" and l u e \<mu> s
      assume A: "ash h=Some (l,u,e)" "cons_as (l,u,e) (locks_\<mu> \<mu>)" "wn_h h \<mu>" 
                "hedge_size h = s"
      from A have "\<exists>w. lsched h (locks_\<mu> \<mu>) w" 
      proof (induct s arbitrary: h l u e \<mu> rule: nat_compl_induct')
        case 0 -- "Empty hedge, the proposition is trivial"
        thus ?case by (rule_tac x="[]" in exI) (auto intro: lsched.intros)
      next
        case (Suc s) 
          -- "Non-empty hedge. Make the case-distinction depicted above"
        show ?case 
        proof (cases rule: h_cases[of h])
          case final -- "The hedge only contains leafs. The proposition is also 
            trivial then, as the empty path is a valid schedule."
          thus ?thesis by (rule_tac x="[]" in exI) (auto intro: lsched.intros)
        next
          case (spawn h1 lab ts t h2)[simp] -- "The hedge contains a spawn step. 
            By assumption, spawn steps have no effect on locks. hence, scheduling the 
            spawn step does not affect the consistency criteria."
          from Suc.prems(3)[simplified] obtain nlab where 
            [simp]: "lab=LNone nlab" 
            by (auto elim: wn_h_spawn_imp_LNoneE)
          have SIZE: "hedge_size (h1@ts#t#h2) \<le> s" using Suc.prems(4) by simp
          from wn_h_preserve_spawn[of \<mu> "LNone nlab" "locks_\<mu> \<mu>", 
                                   OF _ Suc.prems(3)[simplified]] 
          obtain \<mu>' where 
            [simp]: "locks_\<mu> \<mu>'=locks_\<mu> \<mu>" and 
              WNH: "wn_h (h1@ts#t#h2) \<mu>'" 
            by auto
          from Suc.hyps[OF SIZE _ _ WNH] Suc.prems(1,2) obtain w where 
            LS: "lsched (h1@ts#t#h2) (locks_\<mu> \<mu>') w" 
            by fastsimp
          from lsched_spawn[OF LS, of "locks_\<mu> \<mu>" "LNone nlab"] show ?thesis 
            by auto
        next
          case (none h1 t nlab h2)[simp] -- "The hedge contains a non-spawning step 
            with no effects on locks. Scheduling this step does not affect the 
            consistency criteria."
          have SIZE: "hedge_size (h1@t#h2) \<le> s" using Suc.prems(4) by simp
          from wn_h_preserve_nospawn[of \<mu> "LNone nlab" "locks_\<mu> \<mu>", 
                                     OF _ Suc.prems(3)[simplified]] 
          obtain \<mu>' where 
            [simp]: "locks_\<mu> \<mu>'=locks_\<mu> \<mu>" and 
              WNH: "wn_h (h1@t#h2) \<mu>'" 
            by auto
          from Suc.hyps[OF SIZE _ _ WNH] Suc.prems(1,2) obtain w where 
            LS: "lsched (h1@t#h2) (locks_\<mu> \<mu>') w" 
            by fastsimp
          from lsched_nospawn[OF LS, of "locks_\<mu> \<mu>" "LNone nlab"] show ?thesis 
            by auto
        next
          case (release h1 x t h2)[simp] -- "We have at least one release step. 
            Scheduling a release step is always possible and will not make the 
            release history inconsistent, as its effect is to remove an entry from 
            the release history"
          have SIZE: "hedge_size (h1@t#h2) \<le> s" using Suc.prems(4) by simp
          from Suc.prems(3)[simplified] obtain \<mu>1 xs \<mu>2 where 
            [simp]: "\<mu>=\<mu>1@xs#\<mu>2" and 
              WN_SPLIT: "wn_h h1 \<mu>1" "wn_t' (NNOSPAWN (LRel x) t) xs" 
                        "wn_h h2 \<mu>2" and
              WN_DISJ: "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" 
                       "set xs \<inter> locks_\<mu> \<mu>2 = {}"
            by (fastsimp elim: wn_h_prepend_h wn_h_append_h)
          from WN_SPLIT(2) obtain xsh where 
            [simp]: "xs=x#xsh" and 
              XS_SPLIT: "x\<notin>set xsh" "wn_t' t xsh" 
            by auto
          from WN_SPLIT WN_DISJ XS_SPLIT have 
            WNH: "wn_h (h1@t#h2) (\<mu>1@xsh#\<mu>2)" and 
            WNH': "wn_h (h1@h2) (\<mu>1@\<mu>2)" 
            by (auto intro!: wn_h_appendI wn_h_prependI)
          have "ash (h1@(NNOSPAWN (LRel x) t)#h2) = 
                 as (NNOSPAWN (LRel x) t) \<parallel> ash (h1@h2)" 
            by auto
          with Suc.prems(1) have 
            "as (NNOSPAWN (LRel x) t) \<parallel> ash (h1@h2) = Some (l,u,e)" 
            by simp
          then obtain lt ut et l2 u2 e2 where 
            ASS_simps: "as (NNOSPAWN (LRel x) t) = Some (lt,ut,et)" 
                       "ash (h1@h2) = Some (l2,u2,e2)" 
                       "l=lt++l2" "u=ut\<union>u2" "e=et++e2" and
            ASS: "dom lt \<inter> dom l2 = {}" "dom et \<inter> dom e2 = {}"
            by (erule_tac as_comp_SomeE) blast
          from ASS_simps(1) have XIDLT: "x\<in>dom lt" by (auto split: eahl_splits)
          from wn_h_dom_l_lower_\<mu>[OF WNH', simplified] WN_DISJ[simplified] 
          have XNIDL2: "x\<notin>dom l2" by (simp add: ASS_simps[simplified]) blast
          from ASS_simps(1) have AS_T: "as t = Some (l_remove lt x, ut, et)" 
            apply (auto split: option.split_asm prod.split_asm)
            apply (drule_tac wn_t_dom_l_lower_\<mu>[OF XS_SPLIT(2)])
            apply (force simp add: l_remove_def intro!: ext iff add: XS_SPLIT(1)) 
            done
          have "ash (h1@t#h2) = as t \<parallel> ash (h1@h2)" by simp
          also from XNIDL2 ASS 
          have "as t \<parallel> ash (h1@h2) = Some (l_remove l x, u, e)"
            apply (simp only: AS_T ASS_simps(2))
            apply (simp add: ASS_simps)
            apply (auto simp add: l_remove_def map_add_comm)
            apply (force intro!: ext simp add: map_add_def split: option.split)
            done
          finally have G1: "ash (h1@t#h2) = Some (l_remove l x, u, e)" .
          from Suc.prems(2) have 
            G2: "cons_as (l_remove l x, u, e) (locks_\<mu> (\<mu>1@xsh#\<mu>2))" 
            using XIDLT WN_DISJ[simplified] XS_SPLIT(1)
            by simp (blast 5 intro!: cons_h_remove)
          from Suc.hyps[OF SIZE G1 G2 WNH] obtain w where 
            IHAPP: "lsched (h1 @ t # h2) (locks_\<mu> (\<mu>1 @ xsh # \<mu>2)) w" 
            by blast
          moreover have "lock_valid (locks_\<mu> \<mu>) (LRel x) (locks_\<mu> (\<mu>1@xsh#\<mu>2))" 
            using WN_DISJ XS_SPLIT(1) by simp
          ultimately have "lsched (h) (locks_\<mu> \<mu>) ((LRel x)#w)" 
            by (auto intro: lsched.intros)
          thus ?thesis by blast
        next
          case acquire -- {* All the trees start either with acquisitions or are leafs.
            This case is the complex part of the proof. 
            
            We first distinguish whether there is a usage or all acquisitions are 
            final acquisitions.
            *}
          {
            assume C: "\<exists>Xu. \<exists>t\<in>set h. closing' t = Some Xu" -- "There is a usage"
            -- "Find a tree that starts with a usage, where the used locks are 
              disjoint from the release history."
            obtain x Xu t where 
              USE: "NNOSPAWN (LAcq x) t \<in> set h" "closing x t = Some Xu" 
                   "insert x Xu \<inter> dom l = {}" 
            proof (cases "dom l = {}")
              case True[simp] -- "Simple case: Domain of RH is empty, hence we can 
                take any tree in h"
              from C obtain Xu t where 1: "t\<in>set h" "closing' t = Some Xu" 
                by blast
              then obtain x t' where 
                [simp]: "t=NNOSPAWN (LAcq x) t'" and 
                  CL: "closing x t' = Some Xu" 
                by (cases t rule: closing'.cases) auto
              with 1 show ?thesis by (rule_tac that) simp_all
            next
              case False -- "Complex case: Domain of RH is not empty, we have to take
                tree that contains minimal element of RH" 
              with Suc.prems(2) obtain x where MIN: "rh_min l x"
                by (force dest: cons_h_ex_rh_min)
              hence MIDL: "x\<in>dom l" by (auto split: option.split_asm)
              from ash_find_l_t[OF Suc.prems(1) MIDL] 
              obtain h1 t h2 l1 u1 e1 l2 u2 e2 where 
                FT_simps[simp]: "h = h1 @ t # h2" "l = l1 ++ l2" 
                                "u = u1 \<union> u2" "e = e1 ++ e2" and
                FT: "as t = Some (l1, u1, e1)" 
                    "ash h1 \<parallel> ash h2 = Some (l2, u2, e2)" and 
                MIDL1: "x \<in> dom l1" and
                FT_DISJ: "dom l1 \<inter> dom l2 = {}" "dom e1 \<inter> dom e2 = {}" .
              obtain x' t' where TFMT[simp]: "t=NNOSPAWN (LAcq x') t'" 
                using FT(1) MIDL1
                by (subgoal_tac "t\<in>set h") 
                   (erule acquire(1), auto split: option.split_asm)
              have G1: "NNOSPAWN (LAcq x') t' \<in> set h" by simp
              from Suc.prems(3) obtain \<mu>1 xs \<mu>2 where 
                [simp]: "\<mu>=\<mu>1@xs#\<mu>2" and 
                  WN_SPLIT: "wn_h h1 \<mu>1" "wn_t' t xs" "wn_h h2 \<mu>2" and
                  WN_DISJ: "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}"
                           "set xs \<inter> locks_\<mu> \<mu>2 = {}"
                by (fastsimp elim: wn_h_append_h wn_h_prepend_h)
              from WN_SPLIT(2) have WN': "wn_t' t' (x'#xs)" by simp

              from FT(1) obtain l1' u1' e1' where 
                AS: "as t' = Some(l1',u1',e1')" and 
                UU: "dom l1 \<subseteq> dom l1'" "x'\<notin>dom l1" 
                by (force split: eahl_splits)
              from UU MIDL1 have MIDL': "x\<in>dom l1'" by auto
              from MIDL1 UU have MNE: "x\<noteq>x'" by auto
              from wn_t_dom_l_stack'[OF WN' AS MIDL'] 
              obtain xs1 xs2 where 
                "x'#xs = xs1@x#xs2" "set xs1 \<subseteq> dom l1'" 
                "\<forall>x'\<in>set xs1. l1' x' \<le> l1' x \<and> x \<notin> the (l1' x') \<and> x' \<notin> the (l1' x')"
                by blast
              then obtain Xu where L1'X': "l1' x' = Some Xu" "Some Xu \<le> l1' x" 
                using MNE by (cases xs1) auto
              from dom_l_closing[OF AS WN', OF L1'X'(1)] have 
                G2: "closing x' t' = Some Xu" .
              from L1'X'(1) FT(1) AS have 
                L1FMT[simp]: "l1 = l_add_use (l_remove l1' x') {x'}" and 
                X'IU: "x'\<in>u" 
                by (auto split: eahl_splits)
              from MNE MIDL' have 
                "l1' x \<le> l1 x" and 
                X'IL1X: "x' \<in> the (l1 x)" 
                by (auto simp add: l_add_use_def split: option.split)
              with L1'X' have "Some Xu \<le> l1 x" by auto
              with FT_DISJ MIDL1 have 
                XULE: "Some Xu \<le> l x" 
                by (auto simp del: L1FMT simp add: map_add_def split: option.split)
              with MIN have "the (l x) \<inter> dom l = {}" by auto
              moreover from XULE MIDL have "Xu \<subseteq> the (l x)" 
                by (auto simp add: le_option_def split: option.split_asm)
              moreover from X'IL1X FT_DISJ MIDL1 have "x'\<in>the (l x)" 
                by (auto simp add: map_add_def split: option.split)
              ultimately have G3: "insert x' Xu \<inter> dom l = {}" by auto
              from that[OF G1 G2 G3] show ?thesis .
            qed
            
            -- "Split h (This duplicates some work done in the complex case of the 
                proof above)"
            from in_set_list_format[OF USE(1)] obtain h1 h2 where 
              HFMT[simp]: "h=h1@(NNOSPAWN (LAcq x) t)#h2" .
            from Suc.prems(3) obtain \<mu>1 xs \<mu>2 where 
              [simp]: "\<mu>=\<mu>1@xs#\<mu>2" and 
                WN_SPLIT: "wn_h h1 \<mu>1" "wn_t' (NNOSPAWN (LAcq x) t) xs" 
                          "wn_h h2 \<mu>2" and
                WN_DISJ: "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" 
                         "set xs \<inter> locks_\<mu> \<mu>2 = {}"
              by (fastsimp elim: wn_h_append_h wn_h_prepend_h)
            from WN_SPLIT(2) have WN': "wn_t' t (x#xs)" by simp
           
            -- "Split acquisition structure according to splitting of h"
            from Suc.prems(1) obtain l1 u1 e1 l2 u2 e2 where 
              AS_SPLIT: "as (NNOSPAWN (LAcq x) t) = Some (l1,u1,e1)" 
                        "ash h1 \<parallel> ash h2 = Some (l2,u2,e2)" and 
              [simp]: "l=l1++l2" "u=u1\<union>u2" "e=e1++e2" and
                AS_DISJ: "dom l1 \<inter> dom l2 = {}" "dom e1 \<inter> dom e2 = {}"
            proof -
              have "as (NNOSPAWN (LAcq x) t) \<parallel> (ash h1 \<parallel> ash h2) = ash h" 
                by auto
              also have "\<dots> = Some (l,u,e)" using Suc.prems(1) .
              finally show ?thesis by (erule_tac as_comp_SomeE) (blast intro!: that)
            qed
            
            -- "Obtain facts about new tree's acquisition structure"
            from wn_closing_as_fmt[OF WN_SPLIT(2) AS_SPLIT(1) USE(2)] 
            obtain l1' u1' where 
              S: "as t = Some (l1', u1', e1)" "l1' \<le> l1(x \<mapsto> Xu)" 
                 "u1 = insert x u1'" "dom l1' = insert x (dom l1)" .

            from USE(3) have XNIDL: "x\<notin>dom l" by simp
            from S(3) XNIDL Suc.prems(2) have XNILM: "x\<notin>locks_\<mu> \<mu>" by auto

            -- "Construct new hedge's acquisition structure"
            have "ash (h1@t#h2) = as t \<parallel> (ash h1 \<parallel> ash h2)" by simp
            also have "\<dots> = as_comp (l1',u1',e1) (l2,u2,e2)" 
              by (simp add: S(1) AS_SPLIT(2))
            also have "\<dots> = Some (l1'++l2,u1'\<union>u2, e)" 
              using XNIDL S(4) AS_DISJ by auto
            finally have 
              ASH': "ash (h1 @ t # h2) = Some (l1' ++ l2, u1' \<union> u2, e)" .
            
            -- "Collect facts for induction hypothesis"
            from XNILM WN_DISJ WN_SPLIT WN' have 
              WNH': "wn_h (h1@t#h2) (\<mu>1@(x#xs)#\<mu>2)"
              by (auto intro!: wn_h_appendI wn_h_prependI) 

            have CONS': "cons_as (l1' ++ l2, u1' \<union> u2, e) (locks_\<mu> (\<mu>1@(x#xs)#\<mu>2))" 
            proof -
              have CONSL': "cons_h (l1'++l2)" proof -
                from S(2) have LLE: "l1'++l2 \<le> l(x \<mapsto> Xu)" 
                  using XNIDL by (rule_tac le_funI, drule_tac x=xa in le_funD) 
                                 (auto simp add: map_add_def split: option.split)
                from Suc.prems(2) have CL: "cons_h l" by simp
                from wn_closing_ni[where ?\<mu>1.0="[]", simplified, OF WN' USE(2)] 
                have "x\<notin>Xu" .
                with cons_h_update[OF CL, of Xu x] USE(3) 
                have "cons_h (l(x\<mapsto>Xu))" by auto
                with cons_h_antimono[OF LLE] show ?thesis by simp
              qed

              from Suc.prems(2) have 1: "(locks_\<mu> \<mu> - dom l) \<inter> (u\<union>dom e) = {}" 
                by auto
              from S(4) have 
                2: "(locks_\<mu> \<mu> - dom l) \<supseteq> 
                      (locks_\<mu> (\<mu>1@(x#xs)#\<mu>2) - dom (l1' ++ l2))" 
                by auto
              from S(3) have 3: "(u\<union>dom e) \<supseteq> u1' \<union> u2 \<union> dom e" by auto
              from disjoint_mono[OF 2 3 1] have 
                "(locks_\<mu> (\<mu>1@(x#xs)#\<mu>2) - dom (l1' ++ l2)) \<inter> 
                   (u1' \<union> u2 \<union> dom e) = {}" .
              moreover from Suc.prems(2) have "cons_h e" by auto
              moreover note CONSL'
              ultimately show ?thesis by (auto)
            qed

            have SIZE: "hedge_size (h1@t#h2) \<le> s" using Suc.prems(4) by simp

            -- "Apply induction hypothesis"
            from Suc.hyps[OF SIZE ASH' CONS' WNH'] obtain w where 
              IHAPP: "lsched (h1 @ t # h2) (locks_\<mu> (\<mu>1 @ (x # xs) # \<mu>2)) w" 
              by blast

            -- "Show that we can schedule the first step"
            have LV: 
              "lock_valid (locks_\<mu> \<mu>) (LAcq x) (locks_\<mu> (\<mu>1@(x#xs)#\<mu>2))"
              using XNILM by simp
            from lsched_nospawn[OF IHAPP LV] have ?thesis by auto
          } moreover {
            assume C: "\<forall>t\<in>set h. closing' t = None" 
              -- "All the acquisitions at the roots of the hedge are final."
            
            -- "The release history is empty, and any used lock occurs after a final 
              acquisition"
            have "l=empty \<and> u=\<Union>ran e \<and> 
                  \<Union>ran (e |` rootlocks h) = \<Union>ran e \<and> rootlocks h \<subseteq> dom e" 
              by (blast intro!: ncl_as_fmt[OF Suc.prems(3,1) _ C] intro: acquire(1))
            hence 
              [simp]: "l=empty" and 
                NCL: "u=\<Union>ran e" and 
                XMS: "\<Union>ran (e |` rootlocks h) = \<Union>ran e" "rootlocks h \<subseteq> dom e" 
              by auto
            
            -- "There is at least one tree starting with an acquisition, thus the 
                acquisition history is not empty"
            have RLNE:  "rootlocks h \<noteq> {}" and ENE: "e\<noteq>empty" proof -
              obtain t' x h1 h2 where 
                HFMT[simp]: "h=h1@(NNOSPAWN (LAcq x) t')#h2" 
                by (blast intro: acquire(2) elim: in_set_list_format)
              thus "rootlocks h \<noteq> {}" by auto
              with XMS(2) show "e\<noteq>empty" by auto
            qed

            -- "We can obtain a minimal lock that is acquired at a root of some tree"
            obtain x where XIR: "x\<in>rootlocks h" and MIN: "ah_min e x" proof -
              have 1: "e |` rootlocks h \<noteq> empty" using XMS(2) RLNE 
                by (subgoal_tac "dom (e |` rootlocks h) \<noteq> {}") fastsimp+
              from cons_h_ex_ah_min[OF 1 cons_h_antimono[of "e|`rootlocks h" e]] 
                   Suc.prems(2) 
              obtain x where "ah_min (e |` rootlocks h) x" 
                by auto
              with XMS(1) show ?thesis by (auto intro!: that)
            qed

            -- {* Find the tree with @{term x} at the root *}
            from rootlocks_split[OF XIR] obtain h1 t h2 where 
              HFMT[simp]: "h=h1@NNOSPAWN (LAcq x) t#h2" .
            -- {* Split lock-stacks and acquisistion structures *}
            from wn_ash_split_aux[OF Suc.prems(3,1) HFMT] 
            obtain \<mu>1 xs \<mu>2 l1 u1 e1 l2 u2 e2 where
              SPLIT_simps[simp]: "\<mu> = \<mu>1 @ xs # \<mu>2" "u = u1 \<union> u2" 
                                 "e = e1 ++ e2" and
              WNS: "wn_t' (NNOSPAWN (LAcq x) t) xs" "wn_h h1 \<mu>1" 
                   "wn_h h2 \<mu>2" and
              ASS: "as (NNOSPAWN (LAcq x) t) = Some (l1, u1, e1)" 
                   "ash h1 \<parallel> ash h2 = Some (l2, u2, e2)" and
              DISJ: "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" 
                    "set xs \<inter> locks_\<mu> \<mu>2 = {}" "dom l1 \<inter> dom l2 = {}" 
                    "dom e1 \<inter> dom e2 = {}" and 
              LL: "l = l1 ++ l2" 
              .
            from LL have [simp]: "l1=empty" "l2=empty" by auto

            -- {* Get acquisition structure of @{term t} *}
            obtain e1' where 
              AS': "as t = Some (empty,u1,e1')" "e1 = e1'(x \<mapsto> u1)" "x \<notin> dom e1'" 
              by (rule_tac 
                ncl_as_fmt_single'[OF WNS(1)
                                      C[rule_format, of "NNOSPAWN (LAcq x) t"]
                                      ASS(1)]
                 )
                 (simp)

            -- {* Get acqusition structure of new hedge *}
            have ASH': "ash (h1@t#h2) = Some (empty,u,e1'++e2)" proof -
              from AS'(2,3) DISJ have D': "dom e1' \<inter> dom e2 = {}" by simp
              have "ash (h1@t#h2) = as t \<parallel> (ash h1 \<parallel> ash h2)" by simp
              also from DISJ D' AS' ASS(2) have "\<dots> = Some (empty,u,e1'++e2)" 
                by simp
              finally show ?thesis .
            qed

            -- {* The new hedge is well-nested *}
            from AS'(2) Suc.prems(2) have XNILM: "x\<notin>locks_\<mu> \<mu>" by auto
            have WN': "wn_h (h1@t#h2) (\<mu>1@(x#xs)#\<mu>2)" 
              using WNS DISJ XNILM by (auto intro!: wn_h_appendI wn_h_prependI)

            -- {* The new acquisition history is consistent *}
            have CONS': "cons_as (empty,u,e1'++e2) (locks_\<mu> (\<mu>1@(x#xs)#\<mu>2))" 
            proof -
              have "cons_h (e1'++e2)" proof -
                from AS'(2,3) have "e1' \<le> e1" by (simp add: le_fun_def dom_def)
                hence 1: "e1'++e2 \<le> e" by (auto intro!: map_add_first_le)
                from cons_h_antimono[OF 1] Suc.prems(2) show ?thesis by auto
              qed 
              moreover
              have "insert x (locks_\<mu> \<mu>) \<inter> (dom (e1'++e2) \<union> u) = {}" proof -
                from AS' have DEF: "dom e = insert x (dom (e1' ++ e2))" by auto
                from Suc.prems(2) have DJO: "locks_\<mu> \<mu> \<inter> (dom e \<union> u) = {}" 
                  by auto
                have 1: "(dom (e1' ++ e2) \<union> u) \<subseteq> dom e \<union> u" using DEF by auto
                from disjoint_mono[of "locks_\<mu> \<mu>" "locks_\<mu> \<mu>", OF _ 1 DJO] have 
                  "locks_\<mu> \<mu> \<inter> (dom (e1' ++ e2) \<union> u) = {}" 
                  by simp
                moreover from AS' DISJ have "x\<notin>dom (e1'++e2)" by auto
                moreover from MIN NCL have "x\<notin>u" by simp
                ultimately show ?thesis by simp
              qed
              ultimately show ?thesis by fastsimp
            qed

            -- {* Now we can apply the induction hypothesis and finnish the proof *}
            have SIZE: "hedge_size (h1@t#h2) \<le> s" using Suc.prems(4) by simp

            from Suc.hyps[OF SIZE ASH' CONS' WN'] obtain w where 
              IHAPP: "lsched (h1 @ t # h2) (locks_\<mu> (\<mu>1 @ (x#xs) # \<mu>2)) w" 
              by blast
            moreover have "lock_valid (locks_\<mu> \<mu>) (LAcq x) (locks_\<mu> (\<mu>1@(x#xs)#\<mu>2))"
              using XNILM by simp
            ultimately have "lsched (h) (locks_\<mu> \<mu>) ((LAcq x)#w)" 
              by (auto intro: lsched.intros)
            hence ?thesis by blast
          } ultimately show ?thesis by force
        qed
      qed
    } 
    with A show ?thesis by blast
  qed

  text {*
    The following is the main theorem of this section. It states the correctness of the
    acquisition structure construction. For all non-empty hedges that are well-nested
    w.r.t. a list of lock-stacks with locks @{text X}, the existence of a schedule 
    starting with locks @{text X} is equivalent to the conistency of the hedge's 
    acquisition history w.r.t. @{text X}.
    *}

  lemma acqh_correct': 
    fixes h::"('P,'\<Gamma>,'L,'X) lex_hedge" 
    shows "\<lbrakk>wn_h h \<mu>\<rbrakk> \<Longrightarrow> 
    (\<exists>w. lsched h (locks_\<mu> \<mu>) w) \<longleftrightarrow> 
      (\<exists>l u e. ash h = Some (l, u, e) \<and> cons_as (l, u, e) (locks_\<mu> \<mu>)
    )"
    using acqh_sound acqh_precise by blast

  theorem acqh_correct: 
    fixes h::"('P,'\<Gamma>,'L,'X) lex_hedge" 
    assumes WN: "wn_h h \<mu>"
    shows "(\<exists>w. lsched h (locks_\<mu> \<mu>) w) \<longleftrightarrow> cons (ash h) (locks_\<mu> \<mu>)"
    using WN
    (* FIXME: Why doesn't apply (cases "ash h" rule: cons.cases) work ? *)
    apply (simp only: acqh_correct')
    apply (cases "ash h")
    apply simp
    apply (case_tac a)
    apply (case_tac b)
    apply simp
    done

end

end
