header "Hedge Constraints for Acquisition Histories"
theory As_hc
imports Acqh WellNested DPN_c0 Specification
begin
text_raw {*\label{thy:As_hc}*}

text {*
  This theory formulates the set of execution hedges that have a lock-sensitive 
  schedule, and shows how to use hedge-constrained predecessor set computations to 
  compute property specifications based on cascaded predecessor sets.
  *}

subsection "Locks Encoded in Control State"
text {*
  For this section, we make the assumption that the set of locks is encoded in the 
  control state of the DPN. We formalize this by means of a locale.
  *}

locale EncodedLDPN = LDPN +
  -- "The states of the DPN are tuples of some states @{typ 'P} and sets of locks:"
  constrains \<Delta> :: "('P\<times>'X set,'\<Gamma>,'L,'X::finite) ldpn"
  constrains c0 :: "('P\<times>'X set,'\<Gamma>) conf"
  constrains \<mu>0 :: "'X list list"
  -- "A step of the DPN transforms the locks as expected:"
  assumes encoding_correct_nospawn: 
    "((p,X),\<gamma> \<hookrightarrow>\<^sub>l (p',X'),w)\<in>\<Delta> \<Longrightarrow> lock_valid X l X'"
  assumes encoding_correct_spawn1: 
    "((p,X),\<gamma> \<hookrightarrow>\<^sub>l (ps,Xs),ws \<sharp> (p',X'),w) \<in> \<Delta> \<Longrightarrow> lock_valid X l X'"

  -- "A freshly spawned process initially owns no locks:"
  assumes encoding_correct_spawn2:   
    "((p,X),\<gamma> \<hookrightarrow>\<^sub>l (ps,Xs),ws \<sharp> (p',X'),w) \<in> \<Delta> \<Longrightarrow> Xs={}"
begin

  lemmas encoding_correct_spawn = encoding_correct_spawn1 encoding_correct_spawn2
  lemmas encoding_correct = encoding_correct_nospawn encoding_correct_spawn

  lemma encoding_correct_nospawn': 
    "(p,\<gamma> \<hookrightarrow>\<^sub>l p',w)\<in>\<Delta> \<Longrightarrow> lock_valid (snd p) l (snd p')"
    by (cases p, cases p') (auto intro: encoding_correct_nospawn)

  lemma encoding_correct_spawn': 
    assumes A: "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w) \<in> \<Delta>" 
    shows "lock_valid (snd p) l (snd p')" "snd ps={}"
    using A encoding_correct_spawn by (cases p, cases p', cases ps, force)+

  lemma encoding_correct_spawn2':
    "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w) \<in> \<Delta> \<Longrightarrow> snd ps = {}"
    using encoding_correct_spawn by (cases p, cases p', cases ps, force)+
   

  lemma ec_preserve_singlestep: 
    assumes 
      A: "((c,locks_\<mu> \<mu>),l,(c',X'))\<in>ldpntr \<Delta>" "wn_c \<Delta> c \<mu>" 
         "map (snd\<circ>fst) c = map set \<mu>" and 
      C: "!!\<mu>'. \<lbrakk> wn_c \<Delta> c' \<mu>'; X'=locks_\<mu> \<mu>'; 
                  map (snd\<circ>fst) c' = map set \<mu>'
                \<rbrakk> \<Longrightarrow> P"
    shows P
  proof -
    from A have 
      TR: "(c,l,c')\<in>dpntr \<Delta>" and 
      LV: "lock_valid (locks_\<mu> \<mu>) l X'" 
      by (auto simp add: ldpntr_def)
    from TR show ?thesis proof (cases rule: dpntr.cases)
      case (dpntr_no_spawn p \<gamma> _ p' w c1 r c2) 
      hence 
        FMT[simp]: "c = c1 @ (p, \<gamma> # r) # c2" "c' = c1 @ (p', w @ r) # c2" and 
        R: "(p,\<gamma> \<hookrightarrow>\<^sub>l p',w) \<in> \<Delta>" 
        by auto
      from wn_c_split_aux[OF A(2) FMT(1)] obtain \<mu>1 xs \<mu>2 where 
        [simp]: "\<mu> = \<mu>1 @ xs # \<mu>2" and 
          WNS: "wn_\<pi> \<Delta> (p, \<gamma> # r) xs" "wn_c \<Delta> c1 \<mu>1" "wn_c \<Delta> c2 \<mu>2" and
          DISJ: "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" 
                "set xs \<inter> locks_\<mu> \<mu>2 = {}" 
        .
      from A(3) wn_c_length[OF WNS(2)] wn_c_length[OF WNS(3)] have 
        ECS: "map (snd\<circ>fst) c1 = map set \<mu>1" "snd p = set xs" 
             "map (snd\<circ>fst) c2 = map set \<mu>2" 
        by auto
      obtain xs' where 
        "wn_\<pi> \<Delta> (p',w@r) xs'" "X'=(locks_\<mu> (\<mu>1@xs'#\<mu>2))" 
        "locks_\<mu> \<mu>1 \<inter> set xs' = {}" "set xs' \<inter> locks_\<mu> \<mu>2 = {}" "snd p' = set xs'" 
      proof (cases l)
        case LNone[simp] 
        from DISJ LV encoding_correct_nospawn'[OF R] ECS(2) show ?thesis 
          by (rule_tac that[OF wn_\<pi>_none[OF R[simplified] WNS(1)]]) simp_all
      next
        case (LAcq x)[simp] 
        from that[OF wn_\<pi>_acq[OF R[simplified] WNS(1)]] LV DISJ 
             encoding_correct_nospawn'[OF R] ECS(2) 
        show ?thesis by auto
      next
        case (LRel x)[simp] 
        from wn_\<pi>_rel[OF R[simplified] WNS(1)] obtain xs' where 
          [simp]: "xs=x#xs'" and 
            1: "x\<notin>set xs'" and 
            2: "wn_\<pi> \<Delta> (p',w@r) xs'" 
          .
        from 1 LV DISJ encoding_correct_nospawn'[OF R] ECS(2) show ?thesis 
          by (rule_tac that[OF 2]) auto
      qed
      with WNS(2,3) DISJ(2) ECS(1,3) show P 
        by (rule_tac \<mu>'="\<mu>1@xs'#\<mu>2" in C) (auto intro!: wn_c_appendI wn_c_prependI)
    next
      case (dpntr_spawn p \<gamma> _ ps ws p' w c1 r c2) hence 
        FMT[simp]: "c = c1 @ (p, \<gamma> # r) # c2" 
                   "c' = c1 @ (ps, ws) # (p', w @ r) # c2" and 
        R: "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w) \<in> \<Delta>" 
        by auto
      from R obtain nlab where [simp]: "l=LNone nlab" by (cases l) auto
      from wn_c_split_aux[OF A(2) FMT(1)] obtain \<mu>1 xs \<mu>2 where 
        [simp]: "\<mu> = \<mu>1 @ xs # \<mu>2" and 
          WNS: "wn_\<pi> \<Delta> (p, \<gamma> # r) xs" "wn_c \<Delta> c1 \<mu>1" "wn_c \<Delta> c2 \<mu>2" and
          DISJ: "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" 
                "set xs \<inter> locks_\<mu> \<mu>2 = {}" 
        .
      from A(3) wn_c_length[OF WNS(2)] wn_c_length[OF WNS(3)] have 
        ECS: "map (snd\<circ>fst) c1 = map set \<mu>1" "snd p = set xs" 
             "map (snd\<circ>fst) c2 = map set \<mu>2" 
        by auto
      from wn_\<pi>_spawn1[OF R WNS(1)] wn_\<pi>_spawn2[OF R WNS(1)] 
           WNS(2,3) DISJ
      have "wn_c \<Delta> c' (\<mu>1@[]#xs#\<mu>2)"
        by (auto intro!: wn_c_appendI wn_c_prependI)
      thus ?thesis 
        using LV encoding_correct_spawn'[OF R] ECS 
        by (rule_tac \<mu>'="\<mu>1@[]#xs#\<mu>2" in C) auto
    qed
  qed  

  lemma ec_preserve: 
    assumes 
      A: "((c,locks_\<mu> \<mu>),ll,(c',X'))\<in>ldpntrc \<Delta>" "wn_c \<Delta> c \<mu>" 
               "map (snd\<circ>fst) c = map set \<mu>" and 
      C: "!!\<mu>'. \<lbrakk>X'=locks_\<mu> \<mu>'; wn_c \<Delta> c' \<mu>'; map (snd\<circ>fst) c' = map set \<mu>'\<rbrakk> \<Longrightarrow> P" 
    shows "P"
  proof -
    {
      fix c X \<mu> ll c' X' P
      assume 
        A: "((c,X),ll,(c',X'))\<in>ldpntrc \<Delta>" "wn_c \<Delta> c \<mu>" 
                "map (snd\<circ>fst) c = map set \<mu>" "X=locks_\<mu> \<mu>" and 
        C: "!!\<mu>'. \<lbrakk> X'=locks_\<mu> \<mu>'; wn_c \<Delta> c' \<mu>'; 
                    map (snd\<circ>fst) c' = map set \<mu>'
                  \<rbrakk> \<Longrightarrow> P"
      hence P 
      proof (induct arbitrary: \<mu> P rule: trcl_pair_induct)
        case empty thus ?case by auto
      next
        case (cons c x l ch Xh ll c' X' \<mu> P) note [simp]=`x=locks_\<mu> \<mu>`
        from ec_preserve_singlestep[OF cons.hyps(1)[simplified] cons.prems(1,2)] 
        obtain \<mu>' where 
          P: "wn_c \<Delta> ch \<mu>'" "map (snd \<circ> fst) ch = map set \<mu>'" "Xh=locks_\<mu> \<mu>'" 
          .
        from cons.hyps(3)[OF P] cons.prems(4) show ?case by blast
      qed
    } with A C show ?thesis by blast
  qed


  text {* The following abbreviates the locks owned by a configuration: *}
  abbreviation "locks_c c == list_collect_set (snd\<circ>fst) c"

  lemma locks_\<mu>_mapset: "locks_\<mu> \<mu> = \<Union>set (map set \<mu>)"
    by (auto simp add: list_collect_set_as_map)

  lemma locks_c_mapset: "locks_c c = \<Union>set (map (snd\<circ>fst) c)"
    by (auto simp add: list_collect_set_as_map)


end


locale EncodedLDPN_c0 = EncodedLDPN + LDPN_c0 +
  -- "The states of the DPN are tuples of some states @{typ 'P} and sets of locks:"
  constrains \<Delta> :: "('P\<times>'X set,'\<Gamma>,'L,'X::finite) ldpn"
  constrains c0 :: "('P\<times>'X set,'\<Gamma>) conf"
  constrains \<mu>0 :: "'X list list"

  -- "The locks encoded in the initial configuration correspond to the locks in the 
    initial list of lock-stacks:"
  assumes encoding_correct_start: 
    "map (snd\<circ>fst) c0 = map set \<mu>0"

begin

  text {* Reachable configurations are well-nested w.r.t. a lock-stack corresponding to the locks encoded in the control states of the processes *}
  lemma reachable_ec: 
    "\<lbrakk> (c,X)\<in>reachablels; 
       !!\<mu>. \<lbrakk>wn_c \<Delta> c \<mu>; X=locks_\<mu> \<mu>; map (snd\<circ>fst) c = map set \<mu>\<rbrakk> \<Longrightarrow> P
     \<rbrakk> \<Longrightarrow> P"
    apply (unfold reachablels_def)
    apply simp
    apply (erule exE)
    apply (erule ec_preserve)
    apply (rule wellnested)
    apply (rule encoding_correct_start)
    apply blast
    done

  text {*
    Due to our assumptions, a reachable configuration always encodes the locks that 
    are also used by the lock-sensitive semantics.
    *}
  theorem reachable_locks: "(c,X)\<in>reachablels \<Longrightarrow> locks_c c = X"
    by (erule reachable_ec) (auto simp add: locks_\<mu>_mapset locks_c_mapset)

  subsection "Characterizing Schedulable Execution Hedges"
  text {*
    In order to characterize schedulable execution hedges, we have to first 
    characterize the locks allocated at the roots of an execution hedge. This can be 
    done by deriving the locks at the roots from the control states annotated at the 
    leafs.
    *}

  fun lock_eff :: "('L,'X) lockstep \<Rightarrow> 'X set \<Rightarrow> 'X set" where
    "lock_eff (LNone nlab) X = X" |
    "lock_eff (LAcq x) X = insert x X" |
    "lock_eff (LRel x) X = X - {x}"

  fun lock_eff_inv :: "('L,'X) lockstep \<Rightarrow> 'X set \<Rightarrow> 'X set" where
    "lock_eff_inv (LNone nlab) X = X" |
    "lock_eff_inv (LAcq x) X = X - {x}" |
    "lock_eff_inv (LRel x) X = insert x X"


  fun rlocks_t :: "('P\<times>'X set,'\<Gamma>,'L,'X) lex_tree \<Rightarrow> 'X set" where
    "rlocks_t (NLEAF \<pi>) = (case \<pi> of ((p,X),w) \<Rightarrow> X)" |
    "rlocks_t (NNOSPAWN l t) = lock_eff_inv l (rlocks_t t)" |
    "rlocks_t (NSPAWN l ts t) = lock_eff_inv l (rlocks_t t)"

  abbreviation rlocks_h :: "('P\<times>'X set,'\<Gamma>,'L,'X) lex_hedge \<Rightarrow> 'X set list" where 
    "rlocks_h h == map rlocks_t h" 

  lemma tsem_locks: "tsem \<Delta> \<pi> t c' \<Longrightarrow> snd (fst \<pi>) = rlocks_t t"
    apply (induct rule: tsem.induct)
    apply auto [1]
    apply (drule encoding_correct_nospawn')
    apply (case_tac l)
    apply (auto) [3]
    apply (drule encoding_correct_spawn')
    apply (case_tac l)
    apply (auto) [3]
    done

  lemma hsem_locks: "hsem \<Delta> c h c' \<Longrightarrow> map (snd\<circ>fst) c = rlocks_h h"
    by (induct rule: hsem.induct) (auto dest: tsem_locks)

  
  text {*
    Next, we have to characterize the execution hedges with consistent acquisition 
    histories w.r.t. the set of allocated locks.
    *}
  definition "Hls h == cons (ash h) (\<Union>set (rlocks_h h))"

  theorem reachable_hls_char: 
    assumes A: "(c,X)\<in>reachablels" "hsem \<Delta> c h c'" 
    shows "(\<exists>w. lsched h X w) \<longleftrightarrow> Hls h"
  proof -
    from reachable_ec[OF A(1)] obtain \<mu> where 
      [simp]: "X = locks_\<mu> \<mu>" and 
        EC: "wn_c \<Delta> c \<mu>" "map (snd \<circ> fst) c = map set \<mu>" 
      .
    from EC(1) A(2) have WNH: "wn_h h \<mu>" 
      by (auto simp add: wnc_eq_wnch wn_c_h_def)  
    have "(\<exists>w. lsched h X w) \<longleftrightarrow> (\<exists>w. lsched h (locks_\<mu> \<mu>) w)" by simp
    also from acqh_correct[OF WNH] have "\<dots> = cons (ash h) (locks_\<mu> \<mu>)" .
    also have "(locks_\<mu> \<mu>) = \<Union>set (rlocks_h h)" 
      by (simp only: hsem_locks[OF A(2)] locks_\<mu>_mapset EC(2)[symmetric])
    finally show ?thesis by (unfold Hls_def)
  qed

  text {* Now we can put it all together and show correctness of lock-sensitive 
    predecessor computation *}
  lemma lsprestar1:
    assumes 
    REACH:"(c,X)\<in>reachablels" and
    PRE: "c\<in>prehc \<Delta> Hls C'"
    shows "\<exists>c'\<in>C'. \<exists>ll X'. ((c,X),ll,(c',X'))\<in>ldpntrc \<Delta>"
  proof -
    from PRE obtain h c' where A: "c'\<in>C'" "h\<in>Hls" "hsem \<Delta> c h c'" 
      by (auto elim: prehcE)
    from reachable_hls_char[OF REACH A(3)] A(2) obtain ll where 
      B: "lsched h X ll" 
      by (auto simp add: mem_def)
    from lsched_correct2[OF B A(3)] A(1) show ?thesis by blast
  qed

  lemma lsprestar2:
    assumes 
    REACH:"(c,X)\<in>reachablels" and
    MEM: "c'\<in>C'" and
    PATH: "((c,X),ll,(c',X'))\<in>ldpntrc \<Delta>"
    shows "c\<in>prehc \<Delta> Hls C'"
  proof -
    from lsched_correct1[OF PATH] obtain h where 
      A: "hsem \<Delta> c h c'" "lsched h X ll" 
      by blast
    from reachable_hls_char[OF REACH A(1)] A(2) have B: "Hls h" by blast
    from prehcI[OF _ MEM A(1)] B show ?thesis by (auto simp add: mem_def)
  qed

  theorem lsprestar:
    assumes REACH:"(c,X)\<in>reachablels"
    shows "c\<in>prehc \<Delta> Hls C' \<longleftrightarrow> (\<exists>c'\<in>C'. \<exists>ll X'. ((c,X),ll,(c',X'))\<in>ldpntrc \<Delta>)"
    using REACH lsprestar1 lsprestar2 by blast


  subsection {* Checking Specifications Using @{term "prehc \<Delta> Hls"} *}
  text {*
    We now show that we can use our construction to check for property specifications
    (cf. Specification.thy).
    *}

  text {*
    We first have to construct a hedge-constraint for execution hedges that contain a 
    restricted set of labels.
    *}
  fun isLab :: "('L,'X) lockstep set \<Rightarrow> ('Q,'\<Gamma>,'L,'X) lex_tree \<Rightarrow> bool" where
    "isLab L (NLEAF \<pi>) \<longleftrightarrow> True" |
    "isLab L (NNOSPAWN l t) \<longleftrightarrow> l\<in>L \<and> isLab L t" |
    "isLab L (NSPAWN l ts t) \<longleftrightarrow> l\<in>L \<and> isLab L ts \<and> isLab L t"

  abbreviation "HLab L == { h . list_all (isLab L) h}"

  lemma final_h_is_lab[simp]: "final h \<Longrightarrow> list_all (isLab L) h" 
    apply (induct h)
    apply simp
    apply (case_tac a)
    apply auto
    done

  lemma HLab_correct: "sched h ll \<Longrightarrow> h\<in>HLab L \<longleftrightarrow> ll\<in>lists L"
    by (induct rule: sched.induct) (auto simp add: lists.Nil)

  lemmas HLab_correct' = HLab_correct[OF lsched_is_sched]


  text {*
    Then we can show how to check property specifications using @{const prehc}.
    *}
  fun mc_pre :: "('P\<times>'X set,'\<Gamma>,'L,'X) spec \<Rightarrow> ('P\<times>'X set,'\<Gamma>) conf set" where
    "mc_pre [] = UNIV" |
    "mc_pre (SPEC_RESTRICT C # \<Phi>) = C \<inter> mc_pre \<Phi>" |
    "mc_pre (SPEC_STEP L # \<Phi>) = prehc \<Delta> (Hls \<inter> Hpre \<inter> HLab L) (mc_pre \<Phi>)" |
    "mc_pre (SPEC_STEPS L # \<Phi>) = prehc \<Delta> (Hls \<inter> HLab L) (mc_pre \<Phi>)"
  

  lemma mc_pre_correct_aux: 
    "(c,X)\<in>reachablels \<Longrightarrow> c\<in>mc_pre \<Phi> \<longleftrightarrow> (c,X)\<in>Domain (spec_tr \<Phi>)"
  proof (induct \<Phi> arbitrary: c X)
    case Nil thus ?case by auto
  next
    case (Cons A \<Phi>)
    show ?case proof (cases A)
      case (SPEC_RESTRICT C) with Cons show ?thesis by auto
    next
      case (SPEC_STEP L)[simp]
      show ?thesis proof (auto simp add: prehc_def)
        case (goal1 h c')
        from reachable_hls_char[OF Cons.prems goal1(5)] goal1(1) obtain w where 
          LS: "lsched h X w" by (fastsimp simp add: mem_def)
        from Hpre_length1[OF goal1(2) lsched_is_sched[OF LS]] have 
          LEN: "length w = 1" .
        from HLab_correct'[OF LS] goal1(3) have IL: "w\<in>lists L" by simp
        from lsched_correct2[OF LS goal1(5)] obtain X' where 
          P: "((c, X), w, (c', X')) \<in> ldpntrc \<Delta>" 
          ..
        with LEN IL obtain a where 
          [simp]: "w=[a]" and 
            P1: "a\<in>L" "((c, X), a, (c', X')) \<in> ldpntr \<Delta>" 
          by (cases w) auto
        from P Cons.prems have P2: "(c',X')\<in>reachablels" 
          by (unfold reachablels_def) (auto dest: trcl_concat trcl_one_elem)
        from Cons.hyps[OF P2] goal1(4) have 
          "(c', X') \<in> Domain (LDPN.spec_tr \<Delta> \<Phi>)" 
          by simp
        thus ?case using P1 by force
      next
        case (goal2 c' X' l ch Xh)
        from goal2(2) Cons.prems have REACH: "(ch,Xh)\<in>reachablels" 
          by (unfold reachablels_def) (auto dest: trcl_concat trcl_one_elem)
        from Cons.hyps[OF REACH] goal2(3) have IHAPP: "ch\<in>mc_pre \<Phi>" by auto
        from lsched_correct1[OF trcl_one_elem[OF goal2(2)]] obtain h where 
          H: "hsem \<Delta> c h ch" "lsched h X [l]" 
          by blast
        from Hpre_length2[OF lsched_is_sched[OF H(2)]] have 
          HPRE: "h\<in>Hpre" 
          by simp
        from reachable_hls_char[OF Cons.prems H(1)] H(2) have 
          HLS: "h\<in>Hls" 
          by (auto simp add: mem_def)
        from HLab_correct'[OF H(2), of L] goal2(1) have 
          "list_all (isLab L) h" 
          by auto
        with HLS HPRE IHAPP H(1) show ?case by blast
      qed
    next
      case (SPEC_STEPS L)[simp]
      show ?thesis proof (auto simp add: prehc_def)
        case (goal1 h c')
        from reachable_hls_char[OF Cons.prems goal1(4)] goal1(1) obtain w where 
          LS: "lsched h X w" 
          by (fastsimp simp add: mem_def)
        from HLab_correct'[OF LS] goal1(2) have IL: "w\<in>lists L" by simp
        from lsched_correct2[OF LS goal1(4)] obtain X' where 
          P: "((c, X), w, (c', X')) \<in> ldpntrc \<Delta>" ..
        from P Cons.prems have P2: "(c',X')\<in>reachablels" 
          by (unfold reachablels_def) (auto dest: trcl_concat)
        from Cons.hyps[OF P2] goal1(3) 
        have "(c', X') \<in> Domain (LDPN.spec_tr \<Delta> \<Phi>)" by simp
        thus ?case using IL P by force
      next
        case (goal2 c' X' ll ch Xh)
        from goal2(2) Cons.prems have REACH: "(ch,Xh)\<in>reachablels" 
          by (unfold reachablels_def) (auto dest: trcl_concat)
        from Cons.hyps[OF REACH] goal2(3) have IHAPP: "ch\<in>mc_pre \<Phi>" by auto
        from lsched_correct1[OF goal2(2)] obtain h where 
          H: "hsem \<Delta> c h ch" "lsched h X ll" 
          by blast
        from reachable_hls_char[OF Cons.prems H(1)] H(2) have HLS: "h\<in>Hls" 
          by (auto simp add: mem_def)
        from HLab_correct'[OF H(2), of L] goal2(1) 
        have "list_all (isLab L) h" by auto
        with HLS IHAPP H(1) show ?case by blast
      qed
    qed
  qed

  theorem mc_pre_correct: "c0\<in>mc_pre \<Phi> \<longleftrightarrow> model_check_ref \<Phi>"
    using mc_pre_correct_aux[of c0 "locks_\<mu> \<mu>0" \<Phi>, simplified]
    by (unfold model_check_ref_def)

end

end
