header "Computation of Hedge-Constrained Predecessor Sets"
theory CrossProd
imports ListSemantics HedgeAutomata
begin
text_raw {*\label{thy:CrossProd}*}

text {*
  In this section we show how to compute predecessor sets with regular hedge 
  constraints. 
  The computation is done by reduction to the computation of the unconstrained 
  predecessor set. The reduction uses a cross-product like approach, computing a 
  product-DPN of the original DPN and the hedge automaton, and a product regular set 
  of the original regular set and the hedge-automaton's leaf rules.

  This theory uses a list-based representation of DPN-configurations, where the type 
  of a configuration is a list of control- and stack-symbols. This type is less 
  structured than the original type of configurations, that is lists of pairs of 
  control symbol and stack. However, it admits handling configurations as words, and
  sets of configurations as (regular) languages.

  This theory does not use a formalization of regular languages, nor does it generate 
  executable code. Instead, regular sets are modeled as sets. The effectiveness proofs
  show representations that only contain operations well-known to preserve regularity.
  However, an implementation of those operations is not formalized.
*}

text {*
  The cross-product DPN simulates the rules of the hedge-automaton via its 
  transitions, the current state of the hedge automaton is stored in the DPN's state:
*}
inductive_set 
  xdpn :: "('P,'\<Gamma>,'L) dpn \<Rightarrow> ('S,'P,'\<Gamma>,'L) ha_rule set \<Rightarrow> ('P\<times>'S,'\<Gamma>,'L) dpn" 
  for \<Delta> H where
  xdpn_nospawn: 
    "\<lbrakk> (p,\<gamma> \<hookrightarrow>\<^sub>l p',w)\<in>\<Delta>; HAR_NOSPAWN s l s'\<in>H \<rbrakk> \<Longrightarrow> 
       ((p,s),\<gamma> \<hookrightarrow>\<^sub>l (p',s'),w) \<in> xdpn \<Delta> H" |
  xdpn_spawn: 
    "\<lbrakk> (p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w)\<in>\<Delta>; HAR_SPAWN s l ss s'\<in>H \<rbrakk> \<Longrightarrow> 
       ((p,s),\<gamma> \<hookrightarrow>\<^sub>l (ps,ss),ws \<sharp> (p',s'),w)\<in>xdpn \<Delta> H"

text {*
  The @{thm [source] xdpn_nospawn}-rule adds a transition rule to the cross-product 
  DPN for each original non-spawning transition rule and hedge automaton rule that
  could be used to label the node generated by this transition rule. Analogously, 
  the @{thm [source] xdpn_spawn}-rule adds a transition rule to the cross-product DPN
  for spawning rules.
*}


text {*
  We now define operators to map configurations of the cross-product DPN to 
  configurations of the original DPN and sequences of states of the hedge automaton.
*}
abbreviation 
  proj_c1 :: "('P\<times>'S,'\<Gamma>) conf \<Rightarrow> ('P,'\<Gamma>) conf" where 
  "proj_c1 cx == map (\<lambda>((p,s),w). (p,w)) cx"
abbreviation 
  proj_c2 :: "('P\<times>'S,'\<Gamma>) conf \<Rightarrow> 'S list" where 
  "proj_c2 cx == map (\<lambda>((p,s),w). s) cx"

text {*
  We also have to define a mapping for execution hedges, because the labeling of the 
  leafs is different:
*}
fun proj_t1 :: "('P\<times>'S,'\<Gamma>,'L) ex_tree \<Rightarrow> ('P,'\<Gamma>,'L) ex_tree" where
  "proj_t1 (NLEAF ((p,s),w)) = NLEAF (p,w)" |
  "proj_t1 (NNOSPAWN l t) = NNOSPAWN l (proj_t1 t)" |
  "proj_t1 (NSPAWN l ts t) = NSPAWN l (proj_t1 ts) (proj_t1 t)"


text {*
  Next we define how to transform the target set, that contains the configurations of 
  that we want to compute the predecessors.

  The new target set contains the configurations of the original target set with all 
  labelings that may be done by leaf-rules of the hedge automaton:
*}

-- "Process labeled by a leaf-rule: "
abbreviation 
  "xdpnCLP H == { ((p,s),w). \<exists>W. HAR_LEAF s p W \<in> H \<and> w\<in>W }" 

-- "Configuration labeled by leaf-rules:"
abbreviation 
  "xdpnCL H == { cx . (\<forall>((p,s),w)\<in>set cx. ((p,s),w) \<in> xdpnCLP H ) }" 

-- "New target set:"
definition 
  "xdpnC C H == { cx . proj_c1 cx \<in> C } \<inter> xdpnCL H" 
 

text {*
  Finally we define how to transform the computed predecessor set in order to get a 
  set of configurations of the original DPN. This phase consists of two operations: 
  First, we have to restrict the configurations to those that are accepted by the 
  hedge automaton's initial language, and then we have to project away the 
  hedge-automaton's states to get a configuration of the original DPN. In the following 
  definition, these two steps are combined:
*}
definition 
  projH :: "'S list set \<Rightarrow> ('P\<times>'S,'\<Gamma>) conf set \<Rightarrow> ('P,'\<Gamma>) conf set" where 
  "projH H0 Cx == { proj_c1 cx | cx. cx\<in>Cx \<and> proj_c2 cx \<in> H0 }"


subsection {* Correctness of Reduction *}
text {*
  In this section we show that our reduction is correct, i.e. that we really get the 
  hedge-constrained predecessor set by computing the predecessor set of the 
  cross-product DPN and a transformed target set, and then applying the 
  @{const projH}-operator to the result.
*}

text {*
  We first need to introduce a combination operator that combines an original DPN's 
  configuration and a list of hedge automaton states to a cross-product DPN's 
  configuration.
*}

abbreviation "cxs c \<sigma> == zipf (\<lambda>(p,w) s. ((p,s),w)) c \<sigma>"

lemma proj_cxs1[simp]: "length c = length \<sigma> \<Longrightarrow> proj_c1 (cxs c \<sigma>) = c"
  by (induct rule: list_induct2) auto

lemma proj_cxs2[simp]: "length c = length \<sigma> \<Longrightarrow> proj_c2 (cxs c \<sigma>) = \<sigma>"
  by (induct rule: list_induct2) auto

lemma cxs_proj[simp]: "cxs (proj_c1 cx) (proj_c2 cx) = cx"
  by (induct cx) auto

lemma xdpnc_proj: "cx \<in> xdpnC C H \<Longrightarrow> proj_c1 cx \<in> C"
  by (unfold xdpnC_def) auto

text {*
  We now prove the two directions of our main goal. 
  Each direction requires 2 lemmas, the first one for a single tree and the second 
  one for a hedge.
*}

lemmas tsem_induct_x = 
  tsem.induct[ where ?x1.0 = "((p,s),w)", split_format (complete), 
               consumes 1, case_names tsem_leaf tsem_nospawn tsem_spawn
             ]

lemmas tsem_induct_p = 
  tsem.induct[ where ?x1.0 = "(p,w)", split_format (complete), 
               consumes 1, case_names tsem_leaf tsem_nospawn tsem_spawn
             ]

lemma xdpn_correct1_t: 
  "\<lbrakk>tsem (xdpn \<Delta> H) ((p,s),w) t c'; c'\<in>xdpnCL H\<rbrakk> \<Longrightarrow> 
    tsem \<Delta> (p,w) (proj_t1 t) (proj_c1 c') \<and> lab H (proj_t1 t) s"
proof (induct arbitrary: C rule: tsem_induct_x)
  case (tsem_leaf p s w) thus ?case by (auto intro: lab.intros)
next
  case (tsem_nospawn p s \<gamma> l p' s' w r t c' ) thus ?case
    by (auto elim: xdpn.cases intro: lab.intros tsem.intros)
next
  case (tsem_spawn p s \<gamma> l ps ss ws p' s' w ts cs r t c') thus ?case
    by (auto elim: xdpn.cases intro: lab.intros tsem.intros)
qed

lemma xdpn_correct1: 
  "\<lbrakk> hsem (xdpn \<Delta> H) c h c'; c'\<in>xdpnCL H \<rbrakk> \<Longrightarrow> 
     hsem \<Delta> (proj_c1 c) (map proj_t1 h) (proj_c1 c') \<and> 
     labh H (map proj_t1 h) (proj_c2 c)"
proof (induct arbitrary: C' rule: hsem.induct)
  case hsem_empty thus ?case by auto
next
  case (hsem_cons \<pi> t cf' c h c') 
  obtain p s w where [simp]: "\<pi>=((p,s),w)" by (cases \<pi>) auto
  from hsem_cons.prems have CLHS: "cf'\<in>xdpnCL H" "c'\<in>xdpnCL H" by auto
  from xdpn_correct1_t[OF hsem_cons.hyps(1)[simplified] CLHS(1)] 
       hsem_cons.hyps(3)[OF CLHS(2)] 
  show ?case by (auto intro: labh.intros hsem.intros)
qed

lemma xdpn_correct2_t: 
  "\<lbrakk>tsem \<Delta> (p,w) t c'; lab H t s\<rbrakk> \<Longrightarrow> 
     \<exists>tx cx'. tsem (xdpn \<Delta> H) ((p,s),w) tx cx' \<and> 
              cx'\<in>xdpnCL H \<and> proj_t1 tx = t \<and> 
              proj_c1 cx' = c'"
proof (induct arbitrary: s rule: tsem_induct_p)
  case (tsem_leaf p w s) thus ?case
    apply (rule_tac x="NLEAF ((p,s),w)" in exI)
    apply (rule_tac x="[((p,s),w)]" in exI)
    by (auto elim: lab.cases)
next
  case (tsem_nospawn p \<gamma> l p' w r t c' s) 
  from tsem_nospawn.prems obtain s' where 
    HRULE: "HAR_NOSPAWN s l s'\<in>H" "lab H t s'" 
    by (auto elim: lab.cases)
  from tsem_nospawn.hyps(3)[OF HRULE(2)] obtain tx cx' where 
    IHAPP: "tsem (xdpn \<Delta> H) ((p', s'), w @ r) tx cx'" 
           "cx' \<in> xdpnCL H" "proj_t1 tx = t" "proj_c1 cx' = c'" 
    by blast
  from tsem.intros(2)[OF xdpn_nospawn[OF tsem_nospawn.hyps(1) HRULE(1)] IHAPP(1)] 
  have "tsem (xdpn \<Delta> H) ((p, s), \<gamma> # r) (NNOSPAWN l tx) cx'" .
  thus ?case using IHAPP(2,3,4) by fastsimp
next
  case (tsem_spawn p \<gamma> l ps ws p' w ts cs r t c' s) 
  from tsem_spawn.prems obtain ss s' where 
    HRULE: "HAR_SPAWN s l ss s'\<in>H" "lab H ts ss" "lab H t s'" 
    by (auto elim: lab.cases)
  from tsem_spawn.hyps(3)[OF HRULE(2)] tsem_spawn.hyps(5)[OF HRULE(3)] 
  obtain txs cxs tx cx' where 
    IHAPPS: "tsem (xdpn \<Delta> H) ((ps, ss), ws) txs cxs" 
            "cxs \<in> xdpnCL H" "proj_t1 txs = ts" "proj_c1 cxs = cs" and
    IHAPP: "tsem (xdpn \<Delta> H) ((p', s'), w @ r) tx cx'" "cx' \<in> xdpnCL H" 
           "proj_t1 tx = t" "proj_c1 cx' = c'" 
    by blast
  from tsem.intros(3)[ OF xdpn_spawn[OF tsem_spawn.hyps(1) HRULE(1)] 
                         IHAPPS(1) IHAPP(1) ]
  have "tsem (xdpn \<Delta> H) ((p, s), \<gamma> # r) (NSPAWN l txs tx) (cxs @ cx')" .
  thus ?case using IHAPPS(2,3,4) IHAPP(2,3,4) by fastsimp
qed


lemma xdpn_correct2: 
  "\<lbrakk> hsem \<Delta> c h c'; labh H h \<sigma> \<rbrakk> \<Longrightarrow> 
     \<exists>hx cx'. hsem (xdpn \<Delta> H) (cxs c \<sigma>) hx cx' \<and> 
              cx'\<in>xdpnCL H \<and> 
              (map proj_t1 hx) = h \<and> 
              proj_c1 cx' = c'"
proof (induct arbitrary: \<sigma> rule: hsem.induct)
  case hsem_empty thus ?case by (auto)
next
  case (hsem_cons \<pi> t cf' c h c' \<sigma>)
  from hsem_cons.prems obtain s \<sigma>s where 
    [simp]: "\<sigma>=s#\<sigma>s" and 
      LS: "lab H t s" "labh H h \<sigma>s" 
    by (fastsimp elim: labh.cases)
  from hsem_cons.hyps(3)[OF LS(2)] obtain hx cx' where 
    IHAPP: "hsem (xdpn \<Delta> H) (cxs c \<sigma>s) hx cx'" 
           "cx' \<in> xdpnCL H" 
           "map proj_t1 hx = h" 
           "proj_c1 cx' = c'" 
    by blast
  moreover obtain p w where [simp]: "\<pi>=(p,w)" by (cases \<pi>) auto
  from xdpn_correct2_t[OF hsem_cons.hyps(1)[simplified] LS(1)] 
  obtain tx cfx' where 
    "tsem (xdpn \<Delta> H) ((p, s), w) tx cfx'" 
    "cfx' \<in> xdpnCL H" 
    "proj_t1 tx = t" 
    "proj_c1 cfx' = cf'" 
    by blast
  ultimately show ?case
    apply (rule_tac x="tx#hx" in exI)
    apply (rule_tac x="cfx'@cx'" in exI)
    by (auto intro: hsem.intros)
qed

text {*
  Finally we use the lemmas proven above to show our main goal, i.e. a representation
  of the hedge-constrained predecessor set w.r.t. the language of a hedge automaton
  by means of the sequential @{term "pre\<^sup>*"}-operator and the cross-product 
  construction.
*}
theorem xdpn_correct: 
  "prehc \<Delta> (langh (H0,H)) C' = projH H0 ( pre\<^sup>* (xdpn \<Delta> H) (xdpnC C' H) )"
proof (intro equalityI subsetI)
  fix c
  assume A: "c \<in> prehc \<Delta> (langh (H0, H)) C'"
  then obtain c' h where 
    D: "c'\<in>C'" "hsem \<Delta> c h c'" "h\<in>langh (H0,H)" 
    by (unfold prehc_def) auto
  then obtain \<sigma> where DD: "\<sigma>\<in>H0" "labh H h \<sigma>" by (unfold langh_def) auto

    -- {* We need the following later in order to reason about the (underdefined) 
      @{const cxs}-operator: *}
  from hsem_length[OF D(2)] labh_length[OF DD(2)] have 
    [simp]: "length c = length \<sigma>" 
    by simp 
  from xdpn_correct2[OF D(2) DD(2)] obtain hx cx' where 
    M: "hsem (xdpn \<Delta> H) (cxs c \<sigma>) hx cx'" 
       "cx' \<in> xdpnCL H" 
       "map proj_t1 hx = h" 
       "proj_c1 cx' = c'" 
    by blast
  from M(2,4) D(1) have "cx'\<in>xdpnC C' H" by (unfold xdpnC_def) auto
  hence "cxs c \<sigma> \<in> pre\<^sup>* (xdpn \<Delta> H) (xdpnC C' H)" 
    by (rule_tac obtain_schedule[OF M(1)]) (auto simp add: pre_star_def)
  with DD(1) show "c \<in> projH H0 (pre\<^sup>* (xdpn \<Delta> H) (xdpnC C' H))"
    apply (unfold projH_def)
    apply auto
    apply (rule_tac x="cxs c \<sigma>" in exI)
    apply auto
    done
next
  fix c
  assume A: "c \<in> projH H0 (pre\<^sup>* (xdpn \<Delta> H) (xdpnC C' H))"  
  then obtain cx where 
    D: "c=proj_c1 cx" "proj_c2 cx \<in> H0" "cx\<in>pre\<^sup>* (xdpn \<Delta> H) (xdpnC C' H)" 
    by (unfold projH_def) auto
  then obtain ll cx' where 
    DD: "cx'\<in>(xdpnC C' H)" "(cx, ll, cx')\<in>dpntrc (xdpn \<Delta> H)" 
    by (unfold pre_star_def) auto
  then obtain hx where DDH: "hsem (xdpn \<Delta> H) cx hx cx'" 
    by (auto simp add: sched_correct)
  from DD(1) have CL: "cx'\<in>xdpnCL H" "proj_c1 cx' \<in> C'" 
    by (unfold xdpnC_def) auto
  from xdpn_correct1[OF DDH CL(1)] have 
    M: "hsem \<Delta> (proj_c1 cx) (map proj_t1 hx) (proj_c1 cx')" 
       "labh H (map proj_t1 hx) (proj_c2 cx)" 
    by auto
  from D(2) M(2) have "(map proj_t1 hx)\<in>langh (H0,H)" 
    by (unfold langh_def) auto
  with M(1) D(1) CL(2) show "c \<in> prehc \<Delta> (langh (H0, H)) C'" 
    by (unfold prehc_def) auto
qed


subsection "Effectiveness of Reduction"
text {*
  In this section we give indication that the cross-product construction is 
  computable for regular target sets.
  
  The new set of rules @{text xdpn} can be computed if the set of dpn rules and the 
  set of hedge automaton transitions are finite, as the definition of @{text xdpn} is 
  not recursive and each LHS depends on only one element of each set. However, as said 
  above, we do not provide executable code here.

  In \cite{BMT05}, a configuration is represented as a sequence of control and stack
  symbols, each process starting with a control symbol followed by its stack.
  For sequences that start with a control symbol, this representation is isomorphic 
  to our representation (cf. Section~\ref{ListSemantics:stepr}).
  As regular sets of configurations are best defined on this list-based semantics, we 
  also show the effectiveness of our construction on the list-based semantics.

  This section, especially the proofs of the Theorems, are rather technical. The 
  theorems itself show how to compute the new target configuration and the
  projection from the computed predecessor set using only operations well-known to 
  preserve regularity (in this case intersection, union, concatenation, star, and 
  substitution) as well as some sets that are obviously regular. However, no formal 
  proof of regularity or effectiveness is given.
*}

subsubsection {* Definitions *}

text {* This function defines the projection operator from the extended to the 
  original configuration: *}
fun fp_cl1 where
  "fp_cl1 (CTRL (p,s)) = CTRL p" |
  "fp_cl1 (STACK \<gamma>) = STACK \<gamma>"

text {* This function maps a hedge-automaton state to the regular set of all process 
  configurations labeled with that state. Note that the sets 
  @{term "{ [CTRL (p,s)] | p. True }"} and @{term "{[STACK \<gamma>] | \<gamma>. True}"} are 
  obviously regular.
 *}
definition fp_inv_subst2 where
  "fp_inv_subst2 s = conc { [CTRL (p,s)] | p. True } (star {[STACK \<gamma>] | \<gamma>. True})"

text {*
  The projection operator can be written using substitution, projection 
  (a special form of substitution), and intersection.

  The intuitive idea is, that @{term "subst fp_inv_subst2 H0"} is the set of all 
  configurations with a hedge-automaton labeling sequence that is accepted by 
  @{text H0}. 
*}
definition projH_cl :: "'S list set \<Rightarrow> ('Q\<times>'S,'\<Gamma>) cl set \<Rightarrow> ('Q,'\<Gamma>) cl set" where
  "projH_cl H0 Clx = lang_proj fp_cl1 ( subst fp_inv_subst2 H0 \<inter> (Clx) )"

text {*
  The derivation of the new target set is done by first characterizing all sets of 
  cross-product configurations whose leafs are labeled correctly according to the leaf 
  rules of the hedge automaton. Note that there are only finitely many leaf-rules, 
  hence the union below is over a finite set. Moreover, the language @{term W} at a
  leaf rule is regular by default, the operation @{text "map STACK ` _"} is a 
  projection and the operation @{text "op # (CTRL (p,s)) ` _"} is a concatenation. 
  Hence all the operations below are effective.
*}
definition xdpnCL_cl :: "('S,'P,'\<Gamma>,'L) ha_rule set \<Rightarrow> ('P\<times>'S,'\<Gamma>) cl set" where
  "xdpnCL_cl H = star ( \<Union>{ op # (CTRL (p,s)) ` (map STACK ` W) | 
                             s p W. HAR_LEAF s p W \<in> H }
                      )"

text {*
  Having characterized all configurations that are correctly labeled, one gets the 
  new target set by intersecting them with all configurations that correspond to the
  old target set:
*}
definition xdpnC_cl 
  :: "('P,'\<Gamma>) cl set \<Rightarrow> ('S,'P,'\<Gamma>,'L) ha_rule set \<Rightarrow> ('P\<times>'S,'\<Gamma>) cl set" 
  where
  "xdpnC_cl Cl H = lang_inv_proj fp_cl1 Cl \<inter> xdpnCL_cl H"

text {*
  In order to compute @{term "prehc \<Delta> (langh (H0,H)) C'"}, we map C' to its 
  corresponding regular set of list-based configurations @{term "c2cl`C'"} and apply
  the list-based operations for cross-product, predecessor set and projection on it:
*}
definition prehc_cl 
  :: "('Q,'\<Gamma>,'L) dpn \<Rightarrow> ('S,'Q,'\<Gamma>,'L) ha \<Rightarrow> ('Q,'\<Gamma>) cl set \<Rightarrow> ('Q,'\<Gamma>) cl set" 
  where
  "prehc_cl \<Delta> HA Cl' = 
    projH_cl (fst HA) (pre\<^sup>*\<^bsub>cl\<^esub> (xdpn \<Delta> (snd HA)) (xdpnC_cl Cl' (snd HA)))"

subsubsection {* Theorems *}

(* text_raw {* \paragraph{Projection to First Component} *} *)

lemma fp_cl1_map_stack_id[simp]: "map fp_cl1 (map STACK w) = map STACK w"
  by (induct w) auto

lemma fp_cl1_stack_id[simp]: "fp_cl1 s = STACK \<gamma> \<longleftrightarrow> s=STACK \<gamma>"
  by (cases s) auto

lemma fp_cl1_eq_map_stack[simp]: 
  "map fp_cl1 la = map STACK w \<longleftrightarrow> la=map STACK w"
  apply (induct w arbitrary: la)
  apply simp
  apply (case_tac la)
  apply auto
  done

lemma star_STACK[simplified,simp]: 
  "star {[STACK \<gamma>] | \<gamma>. True} = {map STACK w | w. True}"
  apply auto
proof -
  case goal1 thus ?case
    apply (induct rule: star.induct)
    apply auto
    apply (rule_tac x="\<gamma>#w" in exI)
    apply simp
    done
next
  case goal2 thus ?case 
    apply (induct w)
    apply (auto intro: star.ConsI[of "[a]", simplified, standard])
    done
qed

lemma proj_c1_effective: "c2cl (proj_c1 c) = map fp_cl1 (c2cl c)"
  by (induct c) auto

(* text_raw {* \paragraph{Inverse Projection to Second Component} *}*)

lemma fp_inv_subst2I[intro!, simp]: 
  "CTRL (p,s)#map STACK w \<in> fp_inv_subst2 s" 
proof -
  have 1: "[CTRL (p,s)] \<in> { [CTRL (p,s)] | p. True }" by auto
  have 2: "map STACK w \<in> (star {[STACK \<gamma>] | \<gamma>. True})" by auto
  from concI[OF 1 2] show ?thesis by (auto simp add: fp_inv_subst2_def)
qed

lemma fp_inv_subst2E: 
  "\<lbrakk>cl\<in>fp_inv_subst2 s; !!p w. cl=CTRL (p,s)#map STACK w \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
  apply (unfold fp_inv_subst2_def)
  apply (erule concE)
  apply fastsimp
  done

(* text_raw {* \paragraph{Projection of Predecessor Set} *} *)

text {*
  Idea of the operation on the original representations of configurations:
*}
lemma projH_effective': 
  "projH H0 Cx = lang_proj (\<lambda>((p,s),w). (p,w)) 
                           ( lang_inv_proj (\<lambda>((p,s),w). s) H0 \<inter> Cx )"
  by (unfold projH_def lang_proj_def lang_inv_proj_def) auto

text {*
  Correctness of the list-level operation:
*}
theorem projH_effective: "c2cl ` projH H0 Cx = projH_cl H0 (c2cl ` Cx)"
  apply (unfold projH_effective' lang_proj_def lang_inv_proj_def projH_cl_def)
  apply auto
proof -
  case (goal1 cx) thus ?case proof (induct cx arbitrary: Cx H0)
    case Nil thus ?case
      by (force simp add: subst_def subst_word_def)
  next
    case (Cons \<pi>x cx) 
    obtain p s w where [simp]: "\<pi>x=((p,s),w)" by (cases \<pi>x) auto
    from Cons.prems[simplified] have 
      P: "cx\<in>{ cx' . ((p,s),w)#cx'\<in>Cx }" "proj_c2 cx \<in> { ss . s#ss\<in>H0 }" 
      by auto
    from Cons.hyps[OF P] show ?case
      apply auto
    proof -
      case goal1
      from imageI[OF goal1(3), of c2cl, simplified] have 
        "CTRL (p, s) # map STACK w @ c2cl xa \<in> c2cl ` Cx" .
      moreover from goal1(2) have 
        "CTRL (p, s) # map STACK w @ c2cl xa \<in> subst fp_inv_subst2 H0"
        apply (auto simp add: subst_def subst_word_def)
        apply (rule_tac x="s#x" in bexI)
        apply auto
        apply (simp only: append.simps(2)[symmetric])
        apply (rule concI)
        apply auto
        done
      ultimately have 
        "CTRL (p, s) # map STACK w @ c2cl xa \<in> 
          subst fp_inv_subst2 H0 \<inter> c2cl ` Cx" 
        by blast
      from imageI[OF this, of "map fp_cl1"] show ?case by simp
    qed
  qed
next
  case (goal2 cx) thus ?case 
  proof (induct cx arbitrary: Cx H0)
    case Nil thus ?case
      apply (auto simp add: subst_def subst_word_def fp_inv_subst2_def)
      apply (case_tac x)
      apply (auto simp add: conc_def)
      done
  next
    case (Cons \<pi>x cx Cx H0) 
    obtain p s w where [simp]: "\<pi>x=((p,s),w)" by (cases \<pi>x) auto
    from Cons.prems[simplified] have 
      "CTRL (p, s) # map STACK w @ c2cl cx \<in> subst fp_inv_subst2 H0" 
      "((p, s), w) # cx \<in> Cx" 
      by auto
    hence 
      P: "c2cl cx \<in> 
          { cl. CTRL (p, s) # map STACK w @ cl \<in> subst fp_inv_subst2 H0 }" 
      "cx \<in> { cx . ((p,s),w)#cx\<in>Cx }" 
      by auto
    from P(1) have P': "c2cl cx \<in> subst fp_inv_subst2 { ss . s#ss\<in>H0 }"
      apply (auto simp add: subst_def subst_word_def)
      apply (case_tac x)
      apply simp
      apply simp
      apply (erule concE)
      apply auto
      apply (erule fp_inv_subst2E)
      apply auto
      apply (rule_tac x="list" in exI)
      apply auto
    proof -
      case (goal1 list b wa) hence "wa=w \<and> b=c2cl cx"
        apply (cases list)
        apply simp
        apply (cases cx)
        apply simp_all
        apply (erule concE)
        apply auto
        apply (erule fp_inv_subst2E)
        apply simp
        apply (cases cx)
        apply simp_all
        apply (erule fp_inv_subst2E)
        apply simp
        apply (cases cx)
        apply auto
        done
      thus "c2cl cx \<in> conc_list (map fp_inv_subst2 list)" using goal1(2) by simp
    qed
    from Cons.hyps[OF P' P(2)] show ?case by force
  qed
qed


(*text_raw {* \paragraph{Construction of New Target Set} *} *)
 
lemma c2cl_empty_rev: "[] = c2cl []" by simp

theorem xdpnCL_effective: "c2cl ` (xdpnCL H) = xdpnCL_cl H"
  apply (unfold c2cl_def_raw xdpnCL_cl_def)
  apply safe
proof -
  case goal1 thus ?case proof (induct c)
    case Nil thus ?case by simp
  next
    case (Cons \<pi> c)
    from Cons have 
      IHAPP: "c2cl_abbrv c \<in> 
        RegSet.star ( \<Union>{op # (CTRL (p, s)) ` map STACK ` W | 
                          s p W. HAR_LEAF s p W \<in> H}
                    )" 
      by auto
    moreover from Cons.prems have 
      "pc2cl \<pi> \<in> ( \<Union>{op # (CTRL (p, s)) ` map STACK ` W | 
                        s p W. HAR_LEAF s p W \<in> H}
                 )" 
      by (auto) (auto simp add: split_paired_all)
    ultimately show ?case by auto
  qed
next
  case goal2 thus ?case proof (induct rule: star.induct)
    case NilI have "[]\<in>xdpnCL H" by auto
    thus ?case by (blast intro: c2cl_empty_rev[unfolded c2cl_def])
  next
    case (ConsI \<pi>l cl)
    from ConsI.hyps(1) obtain p s w W where 
      [simp]: "\<pi>l = CTRL (p,s)# map STACK w" and 
        P: "w\<in>W" "HAR_LEAF s p W \<in> H" 
      by auto
    hence 
      [simp]: "\<pi>l=pc2cl ((p,s),w)" and 
        C1: "[((p,s),w)]\<in>xdpnCL H" 
      by auto
    from ConsI.hyps(3) obtain c where 
      [simp]: "cl = c2cl_abbrv c" and 
        C2: "c\<in>xdpnCL H" 
      by auto
    from C1 C2 have "((p,s),w)#c\<in>xdpnCL H" by auto
    moreover have "\<pi>l@cl = c2cl_abbrv (((p,s),w)#c)" by auto
    ultimately show ?case by blast
  qed
qed
    
(* FIXME: This proof looks too complicated !? *)
lemma inv_proj_c1_effective: 
  "c2cl ` { cx . proj_c1 cx \<in> C } = lang_inv_proj fp_cl1 (c2cl ` C)"
  apply (unfold c2cl_def_raw)
  apply safe
proof -
  case goal1
  thus ?case proof (induct c arbitrary: C)
    case Nil hence "[]\<in>C" by auto
    thus ?case
      by (auto simp add: lang_inv_proj_def) 
         (blast intro: c2cl_empty_rev[unfolded c2cl_def])
  next
    case (Cons \<pi> c)
    then obtain p s w where [simp]: "\<pi>=((p,s),w)" by (cases \<pi>) auto
    from Cons.prems have P: "proj_c1 c \<in> { c1 . (p,w)#c1 \<in> C }" by auto
    from Cons.hyps[OF P] show ?case
      apply (auto simp add: lang_inv_proj_def)
      apply (drule_tac f=c2cl_abbrv in imageI)
      apply simp
      done
  qed
next
  case (goal2 cl) thus ?case
    apply (auto simp add: lang_inv_proj_def)
  proof -
    case goal1 thus ?thesis
    proof (induct c arbitrary: C cl)
      case Nil hence [simp]: "cl=[]" by (cases cl) auto
      from Nil(2) have "[]\<in>{cx. proj_c1 cx \<in> C}" by simp
      thus ?case by (drule_tac f=c2cl_abbrv in imageI) simp
    next
      case (Cons \<pi> c)
      obtain p w where [simp]: "\<pi>=(p,w)" by (cases \<pi>) auto
      from Cons.prems have P1: "c\<in>{ c . \<pi>#c \<in> C }" by simp
      from Cons.prems(1)[simplified] obtain s cl' where 
        [simp]: "cl=CTRL (p,s) # map STACK w @ cl'" and 
          P2: "map fp_cl1 cl' = c2cl_abbrv c"
        apply -
        apply (elim map_eq_consE map_eq_concE)
        apply (case_tac a)
        apply fastsimp
        apply simp
        done
      from Cons.hyps[OF P2 P1] show ?case
        apply auto
      proof -
        case (goal1 cx) hence "((p,s),w)#cx \<in> {cx. proj_c1 cx \<in> C}" by auto
        thus ?case by (drule_tac f=c2cl_abbrv in imageI) auto
      qed
    qed
  qed
qed

theorem xdpnC_effective: "c2cl ` (xdpnC C H) = xdpnC_cl (c2cl ` C) H"
  apply (unfold xdpnC_def xdpnC_cl_def)
  apply (simp only: c2cl_img_Int)
  apply (simp only: inv_proj_c1_effective xdpnCL_effective)
  done  

(* text_raw {* \paragraph{Constrained Predecessor Set} *} *)
theorem prehc_effective: 
  "c2cl ` prehc \<Delta> (langh (H0,H)) C' = prehc_cl \<Delta> (H0,H) (c2cl ` C')"
  apply (simp add: xdpn_correct prehc_cl_def)
  apply (simp add: xdpnC_effective[symmetric] precl_star_is_pre_star projH_effective)
  done

subsection {* What Does This Proof Tell You ?*}
text {*
  In order to believe that our construction is effective, you have to believe
  that the RHS of Theorem~@{thm [source] prehc_effective} is really effective.

  The effectiveness of the $pre^*$ - computation is shown in \cite{BMT05}, and we 
  have also an unpublished formal proof of the algorithm presented there. We are 
  planning to adapt this proof to our model definition and the latest Isabelle version
  in near future, and then publish it.

  The effectiveness of the involved automata computations is well-known. In a future
  version of this formalization, we plan to formalize or adopt an automata library 
  and use it to generate executable code.
*}

end
