header "Automata for Execution Hedges"
theory HedgeAutomata
imports Main Semantics
begin
text_raw {*\label{thy:HedgeAutomata}*}

text {*
  In this section we define hedge automata that accept execution hedges.

  A hedge automaton consists of a set of states, 
  an regular {\em initial language} of state sequences and a set of transitions.
  Transitions are either leaf transitions that label a leaf node with a state if the 
  configuration at the leaf node is contained in some (regular) language, or
  non-spawning or spawning transitions, that label a spawning or non-spawning node 
  respectively with a state depending on the states of the successor nodes.
*}

text {*
  In this formalization, we model the initial language and the regular languages at the leafs just at sets.
  However, if we want an executable representation, we need to model real automata there. This is planned to
  be done in the future.
*}

datatype ('S,'P,'\<Gamma>,'L) ha_rule = 
  HAR_LEAF 'S 'P "'\<Gamma> list set" |
  HAR_NOSPAWN 'S 'L 'S |
  HAR_SPAWN 'S 'L 'S 'S

types ('S,'P,'\<Gamma>,'L) ha = "'S list set \<times> ('S,'P,'\<Gamma>,'L) ha_rule set"

text {* In order to model acceptance of a hedge, we define a relation
  between trees and states with which we can label those trees. 
  We then extend this relation to hedges.
*}

inductive lab 
  :: "('S,'P,'\<Gamma>,'L) ha_rule set \<Rightarrow> ('P,'\<Gamma>,'L) ex_tree \<Rightarrow> 'S \<Rightarrow> bool" 
  for H where
  lab_leaf: 
    "\<lbrakk> HAR_LEAF s p W \<in> H; w\<in>W \<rbrakk> \<Longrightarrow> lab H (NLEAF (p,w)) s" |
  lab_nospawn: 
    "\<lbrakk> HAR_NOSPAWN s l s' \<in> H; lab H t s' \<rbrakk> \<Longrightarrow> lab H (NNOSPAWN l t) s" |
  lab_spawn: 
    "\<lbrakk> HAR_SPAWN s l ss s' \<in> H; lab H ts ss; lab H t s' \<rbrakk> \<Longrightarrow> 
        lab H (NSPAWN l ts t) s"

inductive labh :: "('S,'P,'\<Gamma>,'L) ha_rule set \<Rightarrow> ('P,'\<Gamma>,'L) ex_hedge \<Rightarrow> 'S list \<Rightarrow> bool" 
  for H where
  labh_empty[simp, intro!]: "labh H [] []" |
  labh_cons: "\<lbrakk> lab H t s; labh H h \<sigma>\<rbrakk> \<Longrightarrow> labh H (t#h) (s#\<sigma>)"

lemma labh_empty[simp]: 
  "labh H [] \<sigma> \<longleftrightarrow> \<sigma>=[]" 
  "labh H h [] \<longleftrightarrow> h=[]" 
  by (auto elim: labh.cases)

lemma labh_length: "labh H h \<sigma> \<Longrightarrow> length h = length \<sigma>" 
  by (induct rule: labh.induct) auto

text {*
  The language of a hedge automaton consists of those hedges whose roots can be labeled with a state sequence in the initial language. 
*}
definition langh :: "('S,'P,'\<Gamma>,'L) ha \<Rightarrow> ('P,'\<Gamma>,'L) ex_hedge set" where
  "langh HA == { h . \<exists>\<sigma>\<in>fst HA. labh (snd HA) h \<sigma> }"

end
