header "DPN Semantics on Lists"
theory ListSemantics
imports Semantics
begin
text_raw {*\label{thy:ListSemantics}*}

text {*
  The interleaving semantics works on configurations that are lists of process 
  configurations. 

  However, in \cite{BMT05} a DPN configuration is represented as a sequence of 
  control and stack symbols. Each process starts with a control symbol, followed by 
  its stack symbols. The configuration is simply a concatenation of processes. This 
  representation allows the notion of a regular set of configurations as a set of 
  configurations accepted by a FSM. 

  In this theory, we adopt this representation of configurations, define a semantics 
  directly over this representation, and show that this representation is isomorphic 
  to ours for sequences starting with a control symbol. Note that sequences starting 
  with a stack symbol have no meaningful interpretation, as each process's configuration
  has to start with a control symbol.
*}

subsection "Definitions"
text {*
  We separate stack and control symbols using a datatype with two constructors:
*}
datatype ('Q,'\<Gamma>) cl_item = CTRL 'Q | STACK '\<Gamma>
types ('Q,'\<Gamma>) cl = "('Q,'\<Gamma>) cl_item list"

text {*
  The mapping from configurations to list-based configurations is straightforward:
*}
fun pc2cl :: "('Q,'\<Gamma>) pconf \<Rightarrow> ('Q,'\<Gamma>) cl" where
  "pc2cl (p,w) = CTRL p # map STACK w"

definition c2cl :: "('Q,'\<Gamma>) conf \<Rightarrow> ('Q,'\<Gamma>) cl" where
  "c2cl c == concat (map pc2cl c)"

abbreviation c2cl_abbrv :: "('Q,'\<Gamma>) conf \<Rightarrow> ('Q,'\<Gamma>) cl" 
  -- "This abbreviation is just for convenience"
  where
  "c2cl_abbrv c == concat (map pc2cl c)"

text {*
  Valid single-process configurations are those that start with a control symbol 
  followed by a list of stack symbols:
*}
definition "pclvalid == {CTRL p#map STACK w | p w. True}"

text {*
  Valid configurations are those that start with a control symbol:
*}
definition "clvalid == {[]} \<union> {CTRL p#c | p c. True}"

text {*
  We also define the step relation directly on list representation of configurations:
*}
inductive_set cltr :: "('Q,'\<Gamma>,'L) dpn \<Rightarrow> (('Q,'\<Gamma>) cl \<times> 'L \<times> ('Q,'\<Gamma>) cl) set" 
  for \<Delta> where
  cltr_no_spawn: 
    "\<lbrakk> (p,\<gamma> \<hookrightarrow>\<^sub>l p',w)\<in>\<Delta> \<rbrakk> \<Longrightarrow> 
       ( c1@[CTRL p, STACK \<gamma>]@c2,
         l,
         c1@CTRL p'#(map STACK w)@c2
       ) \<in> cltr \<Delta>" |
  cltr_spawn: 
    "\<lbrakk> (p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w)\<in>\<Delta> \<rbrakk> \<Longrightarrow> 
       ( c1@[CTRL p, STACK \<gamma>]@c2,
         l,
         c1@CTRL ps#(map STACK ws)@CTRL p'#(map STACK w)@c2
       ) \<in> cltr \<Delta>"

subsection "Theorems"

lemma inj_STACK[simp, intro!]: "inj STACK" by (rule injI) auto


subsubsection "Representation of Single Processes"

lemma pc2cl_not_empty[simp]: "pc2cl \<pi> \<noteq> []" by (cases \<pi>) auto

lemma pc2cl_inj[simp, intro!]: "inj pc2cl"
  apply (rule injI)
  apply (case_tac x, case_tac y)
  apply simp
  done

lemmas pc2cl_inj_simp[simp] = inj_eq[OF pc2cl_inj]

lemma pc2cl_valid[intro!,simp]: "pc2cl \<pi> \<in> pclvalid"
  by (cases \<pi>) (auto simp add: pclvalid_def)

lemma pc2cl_surj: "\<lbrakk>\<pi>l\<in>pclvalid; !!\<pi>. \<pi>l=pc2cl \<pi> \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
  apply (unfold pclvalid_def)
  apply (cases \<pi>l)
  apply simp
  apply fastsimp
  done

subsubsection "Representation of Configurations "

text {* We start with a bunch of simplification rules and other auxilliary lemmas: *}

(* Stack and Control Symbols *)

lemma stack_no_ctrl1[simp]: 
  "map STACK w \<noteq> c1@CTRL p#c2" 
  by (auto elim!: map_eq_concE)

lemmas stack_no_ctrl2[simp] = stack_no_ctrl1[symmetric]

lemma map_stack_ne_cCc1 [simp]: 
  "map STACK w \<noteq> c@CTRL s#c'"
  apply (induct w arbitrary: c s c')
  apply auto
  apply (case_tac c)
  apply auto
  done
lemmas map_stack_ne_cCc2[simp] = map_stack_ne_cCc1[symmetric]

lemmas map_stack_ne_add_simps[simp] =
  map_stack_ne_cCc1[where c="[]", simplified]
  map_stack_ne_cCc1[where c="[a]", simplified, standard]

lemma map_STACK_eq_map_STACK_simp[simp]: 
  "map STACK w @CTRL p # cl = map STACK w' @ CTRL p' # cl' \<longleftrightarrow> 
    w'=w \<and> p'=p \<and> cl'=cl"
  apply (induct w arbitrary: w')
  apply (case_tac w')
  apply auto[2]
  apply (case_tac w')
  apply auto
done

(* Representation of Single Process *)

lemma map_stack_ne_pc2cl[simp]:
  "map STACK w \<noteq> c@pc2cl \<pi>@c'"
  "c@pc2cl \<pi>@c' \<noteq> map STACK w"
  by (cases \<pi>, auto)+

lemmas map_stack_ne_pc2cl_add_simps[simp] =
  map_stack_ne_pc2cl[where c="[]", simplified]
  (*map_stack_ne_pc2cl[where c="[]", simplified]*)

lemma map_STACK_eq_map_STACK_add_simps[simp]:
  "map STACK w @ CTRL p#cl = map STACK w'@pc2cl \<pi>'@cl' \<longleftrightarrow> 
    w=w' \<and> p=fst \<pi>' \<and> cl=map STACK (snd \<pi>')@cl'"
  "map STACK w'@pc2cl \<pi>'@cl' = map STACK w @ CTRL p#cl \<longleftrightarrow> 
    w=w' \<and> p=fst \<pi>' \<and> cl=map STACK (snd \<pi>')@cl'"
  by (cases \<pi>', auto)+

(* Representation of Configuration *)
lemma c2cl_simps[simp]:
  "c2cl [] = []"
  "c2cl (\<pi>#c) = pc2cl \<pi> @ c2cl c"
  "c2cl (c1@c2) = c2cl c1 @ c2cl c2"
  by (unfold c2cl_def) auto

lemma c2cl_empty[simp]: 
  "c2cl c = [] \<longleftrightarrow> c=[]"
  "[] = c2cl c \<longleftrightarrow> c=[]"
  by (cases c, auto)+

lemma c2cl_start_with_ctrl[simp]: 
  "c2cl c \<noteq> STACK \<gamma>#cl"
  "STACK \<gamma>#cl \<noteq> c2cl c"
  by (cases c, auto)+

lemma c2cl_start_with_ctrl_map:
  "w\<noteq>[] \<Longrightarrow> c2cl c \<noteq> map STACK w"
  "w\<noteq>[] \<Longrightarrow> map STACK w \<noteq> c2cl c"
  by (cases w, auto)+


lemma map_stack_c2cl_eq_simps[simp]: 
  "map STACK w @ c2cl c = map STACK w' @ c2cl c' \<longleftrightarrow> w=w' \<and> c2cl c=c2cl c'"
  apply (rule iffI)
  defer
  apply simp
  apply (induct w arbitrary: w')
  apply (case_tac w')
  apply auto
  apply (case_tac w')
  apply auto
  apply (case_tac w')
  apply auto
  done

lemma c2cl_s_cl_eqE: 
  "\<lbrakk> STACK \<gamma> # cl = map STACK w @ c2cl c; 
     !!wr. \<lbrakk> w=\<gamma>#wr; cl = map STACK wr @ c2cl c \<rbrakk> \<Longrightarrow> P
   \<rbrakk> \<Longrightarrow> P"
  by (cases w) auto

lemma c2cl_first_processE: 
  "\<lbrakk> c2cl c = CTRL p#cl2; 
     !!w c2 cl2'. \<lbrakk> c=(p,w)#c2; cl2=(map STACK w)@cl2'; c2cl c2=cl2' \<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
  apply (cases c)
  apply simp
  apply simp
  apply (case_tac a)
  apply simp
  apply blast
  done  

lemma c2cl_find_process1: 
  "\<lbrakk> c2cl c = cl1@CTRL p#cl2; 
     !!c1 w c2. \<lbrakk> c=c1@(p,w)#c2; cl2=(map STACK w)@c2cl c2; 
                  cl1=c2cl c1
                \<rbrakk> \<Longrightarrow> P
   \<rbrakk> \<Longrightarrow> P"
proof (induct cl1 arbitrary: c P rule: length_compl_induct)
  case Nil thus ?case by (force elim!: c2cl_first_processE)
next
  case (Cons e cl1') show ?case proof (cases e)
    case (STACK \<gamma>) with Cons.prems(1) have False by simp thus ?thesis ..
  next
    case (CTRL p')[simp] 
    from Cons.prems(1) have E: "c2cl c = CTRL p' # (cl1'@CTRL p#cl2)" by simp
    from c2cl_first_processE[OF E] obtain w c2 cl2' where 
      [simp]: "c = (p', w) # c2" and 
        S: "cl1' @ CTRL p # cl2 = map STACK w @ cl2'" "c2cl c2 = cl2'" 
      .
    obtain cl1'2 where [simp]: "cl1'=map STACK w @ cl1'2" 
    proof -
      from S(1) have "take (length w) (cl1'@CTRL p#cl2) = map STACK w" by auto
      hence "map STACK w = take (length w) cl1'" 
        by (cases "length w - length cl1'") auto
      hence "cl1'=map STACK w @ drop (length w) cl1'" by auto
      thus ?thesis using that by blast
    qed
    with S have 
      P: "c2cl c2=cl1'2@CTRL p#cl2" and 
      LEN: "length cl1'2 \<le> length cl1'" by auto
    from Cons.hyps[OF LEN P] obtain c1x wx c2x where 
      IHAPP: "c2 = c1x@(p,wx)#c2x" 
             "cl2=map STACK wx @ c2cl c2x" and 
      [simp]: "cl1'2 = c2cl c1x" 
      by metis
    hence 1: "c=((p',w)#c1x)@(p,wx)#c2x" by auto
    show ?thesis by (rule Cons.prems(2)[OF 1 IHAPP(2)]) auto
  qed
qed

text {* Then we show that our representation mapping is injective and surjective on 
  valid configurations. *}

lemma c2cl_inj[simp, intro!]: "inj c2cl"
  apply (rule injI)
proof -
  case (goal1 c c')
  thus ?case proof (induct c arbitrary: c')
    case Nil thus ?case by auto
  next
    case (Cons \<pi> c)
    thus ?case
      apply (cases c')
      apply simp
      apply simp
      apply (cases \<pi>)
      apply (case_tac a)
      apply auto
      done
  qed
qed
   
lemmas c2cl_inj_simps[simp] = inj_eq[OF c2cl_inj]
lemmas c2cl_img_Int[simp] = image_Int[OF c2cl_inj]

lemma c2cl_valid[simp,intro!]: "c2cl c \<in> clvalid" 
  by (cases c) (auto simp add: clvalid_def)

lemma c2cl_surj: "\<lbrakk>cl\<in>clvalid; !!c. cl=c2cl c \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
  apply (unfold clvalid_def)
  apply auto
proof -
  case goal1 thus ?case proof (induct c arbitrary: p)
    case Nil from Nil[of "[(p,[])]"] show ?case by auto
  next
    case (Cons s c) show ?case
      apply (cases s)
      apply (rule_tac p=Q in Cons.hyps)
      apply (rule_tac c="(p,[])#c" in Cons.prems)
      apply simp
      apply (rule_tac p=p in Cons.hyps)
      apply (case_tac c)
      apply simp
      apply (case_tac a)
      apply simp
      apply (rule_tac c="(p,\<Gamma>#b)#list" in Cons.prems)
      apply simp
      done
  qed
qed

subsubsection {* Step Relation on List-Configurations *} 
text_raw {* \label{ListSemantics:stepr} *}

lemma cltr_pres_valid: "(cl,l,cl')\<in>cltr \<Delta> \<Longrightarrow> cl\<in>clvalid \<longleftrightarrow> cl'\<in>clvalid"
  apply (erule cltr.cases)
  apply (auto simp add: clvalid_def)
  apply (case_tac c1)
  apply auto
  apply (case_tac c1)
  apply auto
  apply (case_tac c1)
  apply auto
  apply (case_tac c1)
  apply auto
  done

lemma dpntr_is_cltr: "\<lbrakk>(c,l,c')\<in>dpntr \<Delta>\<rbrakk> \<Longrightarrow> (c2cl c,l,c2cl c')\<in>cltr \<Delta>"
  apply (erule dpntr.cases)
  apply (unfold c2cl_def)
  apply (auto)
  apply (drule_tac ?c2.0="map STACK r@c2cl_abbrv c2" in cltr_no_spawn)
  apply simp
  apply (drule_tac ?c2.0="map STACK r@c2cl_abbrv c2" in cltr_spawn)
  apply simp
done

lemma cltr_is_dpntr: "\<lbrakk> (c2cl c,l,c2cl c')\<in>cltr \<Delta> \<rbrakk> \<Longrightarrow> (c,l,c')\<in>dpntr \<Delta>"
  apply (erule cltr.cases)
  apply auto
  apply (erule c2cl_find_process1)
  apply (erule c2cl_find_process1)
  apply auto
  apply (erule c2cl_s_cl_eqE)
  apply (auto simp del: map_append append_assoc 
              simp add: map_append[symmetric] append_assoc[symmetric] 
              intro: dpntr_no_spawn)
  apply (erule c2cl_find_process1)
  apply (erule c2cl_find_process1)
  apply auto
  apply (erule c2cl_s_cl_eqE)
  apply auto
  apply (case_tac c2b)
  apply simp
  apply (case_tac a)
  apply (auto simp del: map_append append_assoc 
              simp add: map_append[symmetric] append_assoc[symmetric] 
              intro: dpntr_spawn)
done

text {* The following theorem formulates the equivalence of the original semantics 
  and the list-based semantics. 
*}
theorem cltr_eq_dpntr: "(c2cl c,l,c2cl c')\<in>cltr \<Delta>  \<longleftrightarrow> (c,l,c')\<in>dpntr \<Delta>"
  by (metis cltr_is_dpntr dpntr_is_cltr)


text {*
  The next two lemmas ease the derivation of executions of the original semantics 
  from executions of the list-based semantics.
*}
lemma cltr2dpntr_fwd: 
  "\<lbrakk> (c2cl c,l,cl')\<in>cltr \<Delta>; 
     !!c'. \<lbrakk>cl'=c2cl c'; (c,l,c')\<in>dpntr \<Delta>\<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
proof -
  assume 
    A: "(c2cl c,l,cl')\<in>cltr \<Delta>" and 
    C: "!!c'. \<lbrakk>cl'=c2cl c'; (c,l,c')\<in>dpntr \<Delta>\<rbrakk> \<Longrightarrow> P"
  from cltr_pres_valid[OF A] have V: "cl'\<in>clvalid" by auto
  from c2cl_surj[OF V] obtain c' where [simp]: "cl'=c2cl c'" .
  from A show ?thesis by (auto intro: C simp add: cltr_is_dpntr) 
qed

lemma cltr2dpntr_bwd: 
  "\<lbrakk> (cl,l,c2cl c')\<in>cltr \<Delta>; 
     !!c. \<lbrakk>cl=c2cl c; (c,l,c')\<in>dpntr \<Delta>\<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
proof -
  assume 
    A: "(cl,l,c2cl c')\<in>cltr \<Delta>" and 
    C: "!!c. \<lbrakk>cl=c2cl c; (c,l,c')\<in>dpntr \<Delta>\<rbrakk> \<Longrightarrow> P"
  from cltr_pres_valid[OF A] have V: "cl\<in>clvalid" by auto
  from c2cl_surj[OF V] obtain c where [simp]: "cl=c2cl c" .
  from A show ?thesis by (auto intro: C simp add: cltr_is_dpntr) 
qed


text {* Finally, we give some lemmas to directly reason about the transitive closure 
  of the step relation: *}
lemma cltr_is_dpntrc: 
  "(c2cl c,l,c2cl c')\<in>trcl (cltr \<Delta>)  \<Longrightarrow> (c,l,c')\<in>dpntrc \<Delta>"
  by (induct l arbitrary: c) (auto elim!: trcl_unconsE cltr2dpntr_fwd)

lemma dpntrc_is_cltr: 
  "(c,l,c')\<in>dpntrc \<Delta> \<Longrightarrow> (c2cl c,l,c2cl c')\<in>trcl (cltr \<Delta>)"
  by (induct rule: trcl.induct) (auto dest: dpntr_is_cltr)

theorem cltr_eq_dpntrc: 
  "(c2cl c,l,c2cl c')\<in>trcl (cltr \<Delta>)  \<longleftrightarrow> (c,l,c')\<in>dpntrc \<Delta>"
  apply safe
  apply (induct l arbitrary: c)
  apply (auto elim!: trcl_unconsE cltr2dpntr_fwd)
  apply (induct rule: trcl.induct)
  apply (auto dest: dpntr_is_cltr)
  done

lemma cltrc_pres_valid: 
  "(cl,w,cl')\<in>trcl (cltr \<Delta>) \<Longrightarrow> cl\<in>clvalid \<longleftrightarrow> cl'\<in>clvalid"
  by (induct rule: trcl.induct) (auto simp add: cltr_pres_valid) 


lemma cltr2dpntrc_fwd: 
  "\<lbrakk> (c2cl c,l,cl')\<in>trcl (cltr \<Delta>); 
     !!c'. \<lbrakk>cl'=c2cl c'; (c,l,c')\<in>dpntrc \<Delta>\<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
proof -
  assume 
    A: "(c2cl c,l,cl')\<in>trcl (cltr \<Delta>)" and 
    C: "!!c'. \<lbrakk>cl'=c2cl c'; (c,l,c')\<in>dpntrc \<Delta>\<rbrakk> \<Longrightarrow> P"
  from cltrc_pres_valid[OF A] have V: "cl'\<in>clvalid" by auto
  from c2cl_surj[OF V] obtain c' where [simp]: "cl'=c2cl c'" .
  from A show ?thesis by (auto intro: C simp add: cltr_is_dpntrc) 
qed

lemma cltr2dpntrc_bwd: 
  "\<lbrakk> (cl,l,c2cl c')\<in>trcl (cltr \<Delta>); 
     !!c. \<lbrakk>cl=c2cl c; (c,l,c')\<in>dpntrc \<Delta>\<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
proof -
  assume 
    A: "(cl,l,c2cl c')\<in>trcl (cltr \<Delta>)" and 
    C: "!!c. \<lbrakk>cl=c2cl c; (c,l,c')\<in>dpntrc \<Delta>\<rbrakk> \<Longrightarrow> P"
  from cltrc_pres_valid[OF A] have V: "cl\<in>clvalid" by auto
  from c2cl_surj[OF V] obtain c where [simp]: "cl=c2cl c" .
  from A show ?thesis by (auto intro: C simp add: cltr_is_dpntrc) 
qed


subsection {* Predecessor Sets on List-Semantics*}

text {* We also define predecessor sets for the list-semantics: *}
definition precl ("pre\<^bsub>cl\<^esub>") where 
  "pre\<^bsub>cl\<^esub> \<Delta> C' == { c . \<exists>l c'. c'\<in>C' \<and> (c,l,c') \<in> cltr \<Delta> }"
definition precl_star ("pre\<^sup>*\<^bsub>cl\<^esub>") where 
  "pre\<^sup>*\<^bsub>cl\<^esub> \<Delta> C' == { c . \<exists>ll c'. c'\<in>C' \<and> (c,ll,c') \<in> trcl (cltr \<Delta>) }"


text {* And show that they are equivalent to their counterparts defined over the 
  original semantics: *}
lemma precl_is_pre: "pre\<^bsub>cl\<^esub> \<Delta> (c2cl`C) = c2cl`(pre \<Delta> C)"
  apply (unfold precl_def pre_def)
  apply (auto elim!: cltr2dpntr_bwd intro: dpntr_is_cltr)
  done

lemma precl_star_is_pre_star: "pre\<^sup>*\<^bsub>cl\<^esub> \<Delta> (c2cl`C) = c2cl`(pre\<^sup>* \<Delta> C)"
  apply (unfold precl_star_def pre_star_def)
  apply (auto elim!: cltr2dpntrc_bwd intro: dpntrc_is_cltr)
  done


end
