header "DPNs With Locks"
theory LockSem
imports DPN Semantics
begin
  text {*
    In this theory, we define an extension of DPNs, where synchronization of the 
    processes via a finite set of locks is allowed. 

    For this purpose, we assume that the rules are labeled with lock operations.
    *}

  subsection "Model"
  -- {* If a label has either no effect on locks, we allow it to be 
    labeled by some other generic type @{typ 'L}. Otherwise, the label indicates 
    either the acquisition or the release of a lock: *}
  datatype ('L,'X) lockstep = LNone 'L | LAcq 'X | LRel 'X

  -- "Abbreviation for the datatype of a DPN with locks:"
  types ('P,'\<Gamma>,'L,'X) ldpn = "('P,'\<Gamma>,('L,'X) lockstep) dpn"

  text {*
    We encode DPNs with locks in a locale.

    To save some case distinctions in proofs, we assume that only non-spawning rules 
    are labeled with lock operations.
    *}
  locale LDPN = DPN +
    constrains 
      \<Delta> :: "('P,'\<Gamma>,'L,'X::finite) ldpn"
    assumes 
      spawn_no_locks: "\<lbrakk>(p,\<gamma> \<hookrightarrow>\<^sub>a ps,ws \<sharp> p',w) \<in> \<Delta>; !!l. a=LNone l \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
  begin
    lemma snl_simps[simp, intro!]:
      "(p,\<gamma> \<hookrightarrow>\<^bsub>LAcq x\<^esub> ps,ws \<sharp> p',w) \<notin> \<Delta>"
      "(p,\<gamma> \<hookrightarrow>\<^bsub>LRel x\<^esub> ps,ws \<sharp> p',w) \<notin> \<Delta>"
      by (auto elim: spawn_no_locks)

    lemma X_finite: "finite (UNIV::'X set)" by simp
  end

  subsection "Interleaving Semantics"

  text {* The following predicate models the step-relation on the set of allocated 
    locks: *}
  inductive lock_valid :: "'X set \<Rightarrow> ('L,'X) lockstep \<Rightarrow> 'X set \<Rightarrow> bool" where
    -- {* A @{const LNone}-step does not change the set of allocated locks: *}
    lv_none: "lock_valid X (LNone l) X" |
    -- {* A @{const LAcq}-step adds the acquired lock to the set of locks. It is
      only executable if the lock was not allocated before: *}
    lv_acquire: "lock_valid (X-{x}) (LAcq x) (insert x X)" |
    -- {* A @{const LRel}-step removes the released lock from the set of locks. It is
      only executable if the lock was allocated before: *}
    lv_release: "lock_valid (insert x X) (LRel x) (X-{x})"

  lemma lock_valid_simps[simp]:
    "lock_valid X (LNone l) X' \<longleftrightarrow> X=X'"
    "lock_valid X (LAcq x) X' \<longleftrightarrow> X'=insert x X \<and> x\<notin>X"
    "lock_valid X (LRel x) X' \<longleftrightarrow> X=insert x X' \<and> x\<notin>X'"
    apply (auto elim: lock_valid.cases intro: lock_valid.intros)
    apply (subst set_minus_singleton_eq[symmetric], assumption)
    apply (rule lock_valid.intros)
    apply (subst (3) set_minus_singleton_eq[symmetric], assumption)
    apply (rule lock_valid.intros)
    done


  text {* Configurations of the lock-sensitive step-relation consists of the 
    list of processes and the set of currently acquired locks. Note that, 
    at this point in the formalization, we do not make any assumptions on which
    process may release a lock, or on well-nestedness of locks.

    That is, we allow a process releasing a lock that it has not acquired before, or
    locks being used in non-well-nestedness fashion.

    However, in Section~\ref{thy:WellNested}, we formalize such assumptions.

    The lock-sensitive step-relation is the intersection of the original 
    step-relation and the step-relation on allocated locks. *}
  definition ldpntr 
    :: "('P,'\<Gamma>,'L,'X) ldpn \<Rightarrow> (('P,'\<Gamma>) conf \<times> 'X set, ('L,'X) lockstep) LTS" 
    where
    "ldpntr \<Delta> = { ((c,X),l,(c',X')) . (c,l,c') \<in> dpntr \<Delta> \<and> lock_valid X l X'}"

  abbreviation "ldpntrc \<Delta> == trcl (ldpntr \<Delta>)"

  lemma ldpntr_subset: "((c,X),w,(c',X'))\<in>ldpntr \<Delta> \<Longrightarrow> (c,w,c')\<in>dpntr \<Delta>"
    by (auto simp add: ldpntr_def)
  lemma ldpntrc_subset: "((c,X),w,(c',X'))\<in>ldpntrc \<Delta> \<Longrightarrow> (c,w,c')\<in>dpntrc \<Delta>"
    by (induct rule: trcl_pair_induct) (auto dest: ldpntr_subset)

  subsection "Tree Semantics"
  text {*
    For the tree semantics, we only need to redefine the scheduler, such that it keeps
    track of the allocated locks.
    *}

  -- "Abbreviation for type of execution trees and hedges with locks: "
  types ('Q,'\<Gamma>,'L,'X) lex_tree = "('Q,'\<Gamma>,('L,'X) lockstep) ex_tree"
  types ('Q,'\<Gamma>,'L,'X) lex_hedge = "('Q,'\<Gamma>,('L,'X) lockstep) ex_hedge"

  -- {*
    The definition of the lock-sensitive scheduler is straightforward:
    *}
  inductive lsched 
    :: "('Q,'\<Gamma>,'L,'X) lex_hedge \<Rightarrow> 'X set \<Rightarrow> ('L,'X) lockstep list \<Rightarrow> bool" 
    where
    lsched_final: "final h \<Longrightarrow> lsched h X []" |
    lsched_nospawn: 
      "\<lbrakk>lsched (h1@t#h2) X' w; lock_valid X l X'\<rbrakk> \<Longrightarrow> 
         lsched (h1@(NNOSPAWN l t)#h2) X (l#w)" |
    lsched_spawn: 
      "\<lbrakk>lsched (h1@ts#t#h2) X' w; lock_valid X l X'\<rbrakk> \<Longrightarrow> 
         lsched (h1@(NSPAWN l ts t)#h2) X (l#w)"

  -- "Obviously, a lock-sensitive schedule is also a schedule of the original 
    scheduler:"
  lemma lsched_is_sched: "lsched h X ll \<Longrightarrow> sched h ll"
    by (induct rule: lsched.induct) (auto intro: sched.intros)

  subsection "Equivalence of Interleaving and Tree Semantics"

  -- {* Straightforward adoption of proof of @{thm [source] sched_correct1} *}
  lemma lsched_correct1: 
    "((c,X),ll,(c',X'))\<in>ldpntrc \<Delta> \<Longrightarrow> \<exists>h. hsem \<Delta> c h c' \<and> lsched h X ll"
  proof (induct rule: trcl_pair_induct)
    case (empty c X) 
    thus ?case 
      by (induct c) 
         (fastsimp intro!: hsem_cons_single lsched_final elim: lsched.cases)+
  next
    case (cons c X l ch Xh ll c' X')
    from cons.hyps(3) obtain h where 
      IHAPP: "hsem \<Delta> ch h c'" "lsched h Xh ll" 
      by blast
    from cons.hyps(1) have 
      "(c,l,ch)\<in>dpntr \<Delta>" and 
      LV: "lock_valid X l Xh" 
      by (unfold ldpntr_def) auto
    thus ?case proof (cases)
      case (dpntr_no_spawn p \<gamma> la p' w c1 r c2) 
      hence 
        C_simp[simp]: "c = c1 @ (p, \<gamma> # r) # c2" 
                      "ch = c1 @ (p', w @ r) # c2" and 
        C: "(p,\<gamma> \<hookrightarrow>\<^sub>l p',w) \<in> \<Delta>" 
        by auto
      from hsem_lel[OF IHAPP(1)[simplified]] obtain h1 t h2 c1' ct' c2' where 
        [simp]: "h = h1 @ t # h2" "c' = c1' @ ct' @ c2'" and 
        HSPLIT: "hsem \<Delta> c1 h1 c1'" "tsem \<Delta> (p', w @ r) t ct'" 
                "hsem \<Delta> c2 h2 c2'" 
        .
      from tsem_nospawn[OF C HSPLIT(2)] have 
        ST: "tsem \<Delta> (p,\<gamma>#r) (NNOSPAWN l t) ct'" .
      from hsem_conc_lel[OF HSPLIT(1) ST HSPLIT(3)] have 
        "hsem \<Delta> c (h1 @ NNOSPAWN l t # h2) c'" 
        by simp
      moreover from lsched_nospawn[OF IHAPP(2)[simplified] LV] have 
        "lsched (h1 @ NNOSPAWN l t # h2) X (l#ll)" .
      ultimately show ?thesis by blast
    next
      case (dpntr_spawn p \<gamma> la ps ws p' w c1 r c2) 
      hence 
        [simp]: "c = c1 @ (p, \<gamma> # r) # c2" 
                "ch = c1 @ (ps, ws) # (p', w @ r) # c2" and 
        C: "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w) \<in> \<Delta>" 
        by auto
      from IHAPP(1)[simplified] obtain h1 ts t h2 c1' cs' ct' c2' where 
        [simp]: "h = h1 @ ts # t # h2" "c' = c1' @ cs' @ ct' @ c2'" and
        HSPLIT: "hsem \<Delta> c1 h1 c1'" "tsem \<Delta> (ps,ws) ts cs'" 
                "tsem \<Delta> (p', w @ r) t ct'" "hsem \<Delta> c2 h2 c2'" 
        by (fastsimp elim: hsem_split hsem_split_single)
      from tsem_spawn[OF C HSPLIT(2,3)] have 
        ST: "tsem \<Delta> (p,\<gamma>#r) (NSPAWN l ts t) (cs'@ct')" .
      from hsem_conc_lel[OF HSPLIT(1) ST HSPLIT(4)] have 
        "hsem \<Delta> c (h1 @ NSPAWN l ts t # h2) c'" 
        by simp
      moreover from lsched_spawn[OF IHAPP(2)[simplified] LV] have 
        "lsched (h1 @ NSPAWN l ts t # h2) X (l#ll)" .
      ultimately show ?thesis by blast
    qed
  qed
  
  -- {*Straightforward adoption of proof of @{thm [source] sched_correct2} *}
  lemma lsched_correct2: 
    "\<lbrakk> lsched h X ll; hsem \<Delta> c h c' \<rbrakk> \<Longrightarrow> \<exists>X'. ((c,X),ll,(c',X'))\<in>ldpntrc \<Delta>"
  proof (induct h X ll arbitrary: c c' rule: lsched.induct)
    case (lsched_final h X c c') thus ?case by (auto dest: final_hsem_nostep)
  next
    case (lsched_nospawn h1 t h2 Xh ll X l c c') 
    from hsem_lel_h[OF lsched_nospawn.prems] obtain c1 p\<gamma>r c2 c1' ct' c2' where
      [simp]: "c = c1 @ p\<gamma>r # c2" "c' = c1' @ ct' @ c2'" and 
        SPLIT: "hsem \<Delta> c1 h1 c1'" "tsem \<Delta> p\<gamma>r (NNOSPAWN l t) ct'" 
               "hsem \<Delta> c2 h2 c2'"
      .
    from SPLIT(2) obtain p \<gamma> r p' w where 
      [simp]: "p\<gamma>r=(p,\<gamma>#r)" and 
        ST: "(p,\<gamma> \<hookrightarrow>\<^sub>l p',w)\<in>\<Delta>" "tsem \<Delta> (p',w@r) t ct'" 
      by (erule_tac tsem.cases) fastsimp+
    from dpntr_no_spawn[OF ST(1)] have "(c,l,c1 @ (p', w @ r) # c2)\<in>dpntr \<Delta>" 
      by auto
    with lsched_nospawn.hyps(3) have 
      "((c,X),l,(c1 @ (p', w @ r) # c2,Xh))\<in>ldpntr \<Delta>" 
      by (unfold ldpntr_def) auto
    also 
    from lsched_nospawn.hyps(2)[OF hsem_conc_lel[OF SPLIT(1) ST(2) SPLIT(3)]] 
    obtain X' where 
      SST: "((c1 @ (p', w @ r) # c2,Xh), ll, (c1' @ ct' @ c2',X')) \<in> ldpntrc \<Delta>" 
      by blast
    finally show ?case by auto
  next
    case (lsched_spawn h1 ts t h2 Xh ll X l c c')
    from hsem_lel_h[OF lsched_spawn.prems] obtain c1 p\<gamma>r c2 c1' ct' c2' where 
      [simp]: "c = c1 @ p\<gamma>r # c2" "c' = c1' @ ct' @ c2'" and 
        SPLIT: "hsem \<Delta> c1 h1 c1'" "tsem \<Delta> p\<gamma>r (NSPAWN l ts t) ct'" 
               "hsem \<Delta> c2 h2 c2'"
      .
    from SPLIT(2) obtain p \<gamma> r ps ws p' w cts' ctt' where 
      [simp]: "p\<gamma>r=(p,\<gamma>#r)" "ct'=cts'@ctt'" and 
        ST: "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w)\<in>\<Delta>" "tsem \<Delta> (ps,ws) ts cts'" 
            "tsem \<Delta> (p',w@r) t ctt'" 
      by (erule_tac tsem.cases) fastsimp+
    from dpntr_spawn[OF ST(1)] have 
      "(c,l,c1 @ (ps,ws) # (p', w @ r) # c2)\<in>dpntr \<Delta>" 
      by auto
    with lsched_spawn.hyps(3) have 
      "((c,X),l,(c1 @ (ps,ws)#(p', w @ r) # c2,Xh))\<in>ldpntr \<Delta>" 
      by (unfold ldpntr_def) auto
    also from 
      lsched_spawn.hyps(2)[OF hsem_conc_leel[OF SPLIT(1) ST(2,3) SPLIT(3)]] 
    obtain X' where 
      SST: "((c1 @ (ps,ws) # (p', w @ r) # c2,Xh), ll, (c',X')) \<in> ldpntrc \<Delta>" 
      by fastsimp
    finally show ?case by auto
  qed

  theorem lsched_correct: 
    "(\<exists>X'. ((c,X),ll,(c',X'))\<in>ldpntrc \<Delta>) \<longleftrightarrow> (\<exists>h. hsem \<Delta> c h c' \<and> lsched h X ll)"
    by (auto intro: lsched_correct1 lsched_correct2)


end
