header "Monitors (aka Block-Structured Locks)"
theory Monitors
imports LockSem WellNested As_hc
begin
text_raw {*\label{thy:Monitors}*}


text {*
  We model monitors by binding locks to stack symbols, and making some 
  restrictions on rules:
  
  \begin{itemize}
    \item A rule labeled by @{const LNone} must not change the allocated locks, 
       nor must it push or pop stack symbols associated with locks.

    \item An acquisition rule must be a rule that pushes a stack-symbol with the 
       acquired lock, and does not change the locks of the stacl-symbol at the 
       bottom.

    \item A release rule must be a rule that pops a stack-symbol with the 
          released lock.
  \end{itemize}
  *}

text {*
  One purpose of this theory is, that it gives strong evidence that our model is 
  not too restrictive. This is done by defining an introduction rule for encoded 
  DPNs with initial configurations that only depends on local properties of the 
  rules and the initial configuration.
*}

  -- "Lock-stack encoded into stack"
definition lstackm_s :: "('\<Gamma> \<rightharpoonup> 'X) \<Rightarrow> '\<Gamma> \<Rightarrow> 'X list" where
  "lstackm_s mon \<gamma> = (case mon \<gamma> of None \<Rightarrow> [] | Some x \<Rightarrow> [x])"

lemma lstackm_s_simps[simp]: 
  "mon \<gamma> = None \<Longrightarrow> lstackm_s mon \<gamma> = []"
  "mon \<gamma> = Some x \<Longrightarrow> lstackm_s mon \<gamma> = [x]"
  by (auto simp add: lstackm_s_def)


fun lstackm :: "('\<Gamma> \<rightharpoonup> 'X) \<Rightarrow> '\<Gamma> list \<Rightarrow> 'X list" where
  "lstackm mon [] = []" |
  "lstackm mon (\<gamma>#s) = lstackm_s mon \<gamma> @ lstackm mon s"

lemma lstackm_conc[simp]: 
  "lstackm mon (s@s') = lstackm mon s @ lstackm mon s'"
  by (induct s) auto

lemma lstack_spawn_empty[simp]: 
  "\<lbrakk> (\<forall>\<gamma>s\<in>set w. mon \<gamma>s=None) \<rbrakk> \<Longrightarrow> lstackm mon w = []"
  by (induct w) (auto)


locale MDPN = EncodedLDPN +
  constrains 
    \<Delta> :: "('P\<times>'X set,'\<Gamma>,'L,'X::finite) ldpn"
  fixes mon :: "'\<Gamma> \<Rightarrow> 'X option" -- "Maps stack symbols to associated monitors"

  assumes
    locks_lnone_pop_nospawn: 
      "(p,\<gamma> \<hookrightarrow>\<^bsub>LNone a\<^esub> p',[])\<in>\<Delta> \<Longrightarrow> mon \<gamma> = None" and
    locks_lnone_pop_spawn: 
      "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',[])\<in>\<Delta> \<Longrightarrow> mon \<gamma> = None" and
    locks_lnone_nospawn: 
      "(p,\<gamma> \<hookrightarrow>\<^bsub>LNone a\<^esub> p',w@[\<gamma>'])\<in>\<Delta> \<Longrightarrow> mon \<gamma>' = mon \<gamma> \<and> 
                                       (\<forall>\<gamma>s\<in>set w. mon \<gamma>s=None)" and
    locks_lnone_spawn: 
      "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w@[\<gamma>'])\<in>\<Delta> \<Longrightarrow> mon \<gamma>' = mon \<gamma> \<and> 
                                       (\<forall>\<gamma>s\<in>set w. mon \<gamma>s=None)" and
    locks_spawn: 
      "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w)\<in>\<Delta> \<Longrightarrow> (\<forall>\<gamma>s\<in>set ws. mon \<gamma>s=None)" and
    locks_acquire: 
      "\<lbrakk> (p,\<gamma> \<hookrightarrow>\<^bsub>LAcq x\<^esub> p',w)\<in>\<Delta>; 
         !!w' \<gamma>2 \<gamma>1. \<lbrakk> w=w'@[\<gamma>1,\<gamma>2]; mon \<gamma>2 = mon \<gamma>; mon \<gamma>1 = Some x; 
                      (\<forall>\<gamma>s\<in>set w'. mon \<gamma>s=None) 
                     \<rbrakk> \<Longrightarrow> P 
       \<rbrakk> \<Longrightarrow> P " and
    locks_release: 
      "(p,\<gamma> \<hookrightarrow>\<^bsub>LRel x\<^esub> p',w)\<in>\<Delta> \<Longrightarrow> w=[] \<and> mon \<gamma> = Some x"

begin

  abbreviation "lstack_s == lstackm_s mon"
  abbreviation "lstack == lstackm mon"

  lemma lstack_lnone_nospawn: 
    "\<lbrakk>(p,\<gamma> \<hookrightarrow>\<^bsub>LNone a\<^esub> p',w)\<in>\<Delta>\<rbrakk> \<Longrightarrow> lstack (\<gamma>#r) = lstack (w@r)"
    apply (cases w rule: rev_cases)
    apply simp
    apply (drule locks_lnone_pop_nospawn) 
    apply (simp)
    apply (simp)
    apply (drule locks_lnone_nospawn)
    apply (cases "mon \<gamma>")
    apply (simp_all)
    done

  lemma lstack_lnone_spawn: 
    "\<lbrakk>(p,\<gamma> \<hookrightarrow>\<^bsub>a\<^esub> ps,ws \<sharp> p',w)\<in>\<Delta>\<rbrakk> \<Longrightarrow> lstack (\<gamma>#r) = lstack (w@r)"
    apply (cases w rule: rev_cases)
    apply simp
    apply (drule locks_lnone_pop_spawn) 
    apply (simp)
    apply (simp)
    apply (drule locks_lnone_spawn) 
    apply (cases "mon \<gamma>")
    apply (simp_all)
    done


  lemma well_nested_t:
    assumes CONS: "distinct (lstack (snd \<pi>))"
    assumes H: "tsem \<Delta> \<pi> t c'"
    assumes COINC: "snd (fst \<pi>) = set (lstack (snd \<pi>))"
    shows "wn_t' t (lstack (snd \<pi>))"
    using H CONS COINC
  proof (induct rule: tsem.induct)
    case tsem_leaf thus ?case by (auto intro: wn_t.intros)
  next
    case (tsem_spawn p \<gamma> l ps ws p' w ts cs r t c')
    from spawn_no_locks[OF tsem_spawn.hyps(1)] obtain la where 
      [simp]: "l=LNone la" 
      by auto
    from locks_spawn[OF tsem_spawn.hyps(1)] have 
      [simp]: "lstack ws = []" 
      by (simp add: lstack_spawn_empty)
    from encoding_correct_spawn2'[OF tsem_spawn.hyps(1)] have 
      [simp]: "snd ps = {}" .
    from tsem_spawn.hyps(3) have 
      IHAPP1: "wn_t' ts (lstack (snd (ps, ws)))" 
      by simp
    moreover
    from lstack_lnone_spawn[OF tsem_spawn.hyps(1)] have 
      LSF[simplified, simp]: "lstack (\<gamma> # r) = lstack (w @ r)" .
    moreover from encoding_correct_spawn'[OF tsem_spawn.hyps(1)] have 
      [simp]: "snd p = snd p'" 
      by simp
    from tsem_spawn.prems tsem_spawn.hyps(5) LSF have 
      IHAPP2: "wn_t' t (lstack (w@r))" 
      by simp
    ultimately show ?case by simp
  next
    case (tsem_nospawn p \<gamma> l p' w r t c')
    show ?case 
    proof (cases l)
      case (LNone la)[simp]
      from lstack_lnone_nospawn tsem_nospawn.hyps(1) have 
        [simplified, simp]: "lstack (\<gamma>#r) = lstack (w@r)" 
        by simp
    moreover from encoding_correct_nospawn'[OF tsem_nospawn.hyps(1)] have 
      [simp]: "snd p = snd p'" 
      by simp
      from tsem_nospawn.prems tsem_nospawn.hyps(3) have 
        IHAPP: "wn_t' t (lstack (w@r))" 
        by simp
      thus ?thesis by simp
    next
      case (LAcq x)[simp]
      from tsem_nospawn.hyps(1)[simplified] show ?thesis 
      proof (cases rule: locks_acquire[consumes 1, case_names C])
        case (C w' \<gamma>2 \<gamma>1)
        note [simp] = C(1)
        from C(4) have [simp]: "lstack w' = []" by simp
        from C(3) have [simp]: "lstack_s \<gamma>1 = [x]" by simp
        from C(2) have [simp]: "lstack_s \<gamma>2 = lstack_s \<gamma>" 
          by (cases "mon \<gamma>") simp_all
        
        from encoding_correct_nospawn'[OF tsem_nospawn.hyps(1)] have 
          XNSP: "x\<notin>snd p" and 
          SP'F[simp]: "snd p' = insert x (snd p)" 
          by auto
        from tsem_nospawn.prems(2) XNSP have 
          XNIS: "x\<notin>set (lstack (\<gamma>#r))" 
          by simp
        from XNIS[simplified] tsem_nospawn.prems(1)[simplified] have 
          P1: "distinct (lstack (w@r))" 
          by (simp)
        from tsem_nospawn.prems(2)[simplified] tsem_nospawn.hyps P1 have 
          IHAPP: "wn_t' t (lstack (w@r))" 
          by simp
        thus ?thesis using XNIS by simp
      qed
    next
      case (LRel x)[simp]
      from tsem_nospawn.hyps(1)[simplified] locks_release have 
        [simp]: "w=[]" "mon \<gamma> = Some x" 
        by auto
      from encoding_correct_nospawn'[OF tsem_nospawn.hyps(1)] have 
        XNSP: "x\<notin>snd p'" and SPF[simp]: "snd p' = snd p - {x}" 
        by auto
      from tsem_nospawn.prems(1)[simplified] have 
        P1: "distinct (lstack (w@r))" 
        by (simp)
      from tsem_nospawn.prems have P2: "snd p' = set (lstack (w @ r))" by simp
      from tsem_nospawn.hyps P1 P2 have IHAPP: "wn_t' t (lstack (w@r))" by simp
      thus ?thesis using tsem_nospawn.prems(1) by simp
    qed
  qed

  lemma well_nested_h:
    assumes CONS: "cons_\<mu> (map (lstack \<circ> snd) c)"
    assumes H: "hsem \<Delta> c h c'"
    assumes COINC: "map (snd\<circ>fst) c = map (set\<circ>lstack\<circ>snd) c"
    shows "wn_h h (map (lstack \<circ> snd) c)"
    using H CONS COINC
    by (induct rule: hsem.induct) (auto intro: well_nested_t)


  theorem well_nested:
    assumes CONS: "cons_\<mu> (map (lstack \<circ> snd) c)"
    assumes COINC: "map (snd\<circ>fst) c = map (set\<circ>lstack\<circ>snd) c"
    shows "wn_c \<Delta> c (map (lstack \<circ> snd) c)"
    apply (simp add: wnc_eq_wnch)
    apply (unfold wn_c_h_def)
    apply (blast intro: well_nested_h[OF CONS _ COINC])
    done


  text {*
    This theorem can be used to show that an MDPN along with a consistent start 
    configuration is a DPN with well-nested lock usage, as described by the 
    locale @{const EncodedLDPN_c0}.
    *}
  theorem EncodedLDPN_c0_intro[intro?]:
    assumes start_config_cons: "cons_\<mu> \<mu>0"
    assumes start_config_coinc: "map (snd\<circ>fst) c0 = map set \<mu>0"
    assumes start_config_match: "map (lstack \<circ> snd) c0 = \<mu>0"
    shows "EncodedLDPN_c0 \<Delta> c0 \<mu>0"
  proof
    from start_config_coinc start_config_match[symmetric] have 
      "map (snd\<circ>fst) c0 = map set (map (lstack \<circ> snd) c0)" 
      by simp
    also have "\<dots> = map (set \<circ> lstack \<circ> snd) c0" by (simp add: map_compose)
    finally show "wn_c \<Delta> c0 \<mu>0" 
      using start_config_cons start_config_match by (blast intro: well_nested)
  qed (rule start_config_coinc)

end

theorem EncodedLDPN_c0_intro_external: 
  assumes MDPN: "MDPN \<Delta> mon"
  assumes start_config_cons: "cons_\<mu> \<mu>0"
  assumes start_config_coinc: "map (snd\<circ>fst) c0 = map set \<mu>0"
  assumes start_config_match: "map (lstackm mon \<circ> snd) c0 = \<mu>0"
  shows "EncodedLDPN_c0 \<Delta> c0 \<mu>0"
proof -
  interpret MDPN[\<Delta> mon] using MDPN .
  from EncodedLDPN_c0_intro[OF start_config_cons start_config_coinc 
                               start_config_match] 
  show ?thesis .
qed

subsection {* Non-Trivial Instance of a Well-Nested DPN *}

text {*
  In this section, we define a non-trivial Well-nested DPN by hand.
  This gives strong evidence that our model assumptions are not too restrictive.
*}

text {*
  We start by introducing some finite set of locks that we can use in our programs:
*}
typedef t_my_locks = "{1..6::nat}" by auto

instance t_my_locks::finite 
proof (intro_classes)
  have "Rep_t_my_locks ` UNIV \<subseteq> t_my_locks" using Rep_t_my_locks by auto
  moreover have "finite t_my_locks" by (unfold t_my_locks_def) auto
  ultimately show "finite (UNIV::t_my_locks set)" 
    apply (rule_tac f=Rep_t_my_locks in finite_imageD)
    apply (drule finite_subset)
    apply assumption+
    apply (rule injI)
    apply (simp add: Rep_t_my_locks_inject)
    done
qed

definition l1 :: t_my_locks where "l1 = Abs_t_my_locks (1::nat)"
definition l2 :: t_my_locks where "l2 = Abs_t_my_locks (2::nat)"

lemma [simp, intro!]: "l1\<noteq>l2" "l2\<noteq>l1" 
  apply (unfold l1_def l2_def)
  apply (auto simp add: Abs_t_my_locks_inject t_my_locks_def)
  done

text {*
  The following rules correspond to a by-hand 
  translation of the (nonsense) program:
  \begin{verbatim}
    procedure p1:
      sync l1 {
        sync l2 {
          spawn p1
          spawn p2
        }
      }

    procedure p2:
      if ? then
        spawn p2
        call p2
      else
        sync l2 {
          sync l1 {
            spawn p1
          }
        }

  \end{verbatim}
*}

definition my\<Delta> :: "(nat \<times> t_my_locks set,nat,unit,t_my_locks) ldpn" where
  "my\<Delta> = {
    ((0,{}),1 \<hookrightarrow>\<^bsub>LAcq l1\<^esub> (0,{l1}),[2,3]),
    ((0,{l1}),2 \<hookrightarrow>\<^bsub>LAcq l2\<^esub> (0,{l1,l2}),[4,5]),
    ((0,{l1,l2}),4 \<hookrightarrow>\<^bsub>LNone ()\<^esub> (0,{}),[1]\<sharp>(0,{l1,l2}),[6]),
    ((0,{l1,l2}),6 \<hookrightarrow>\<^bsub>LNone ()\<^esub> (0,{}),[11]\<sharp>(0,{l1,l2}),[7]),
    ((0,{l1,l2}),7 \<hookrightarrow>\<^bsub>LRel l2\<^esub> (0,{l1}),[]),
    ((0,{l1}),5 \<hookrightarrow>\<^bsub>LRel l1\<^esub> (0,{}),[]),
    ((0,{}),3 \<hookrightarrow>\<^bsub>LNone ()\<^esub> (0,{}),[]),

    ((0,{}),11 \<hookrightarrow>\<^bsub>LNone ()\<^esub> (0,{}),[11]\<sharp>(0,{}),[12]),
    ((0,{}),12 \<hookrightarrow>\<^bsub>LNone ()\<^esub> (0,{}),[11,13]),
    ((0,{}),11 \<hookrightarrow>\<^bsub>LAcq l2\<^esub> (0,{l2}),[14,13]),
    ((0,{l2}),14 \<hookrightarrow>\<^bsub>LAcq l1\<^esub> (0,{l1,l2}),[16,17]),
    ((0,{l1,l2}),16 \<hookrightarrow>\<^bsub>LNone ()\<^esub> (0,{}),[1]\<sharp>(0,{l1,l2}),[18]),
    ((0,{l1,l2}),18 \<hookrightarrow>\<^bsub>LRel l1\<^esub> (0,{l2}),[]),
    ((0,{l2}),17 \<hookrightarrow>\<^bsub>LRel l2\<^esub> (0,{}),[]),
    ((0,{}),13 \<hookrightarrow>\<^bsub>LNone ()\<^esub> (0,{}),[])
  }"

definition my_mon :: "nat \<Rightarrow> t_my_locks option" where
  "my_mon s = (
    if      s=1 then None
    else if s=2 then Some l1
    else if s=3 then None
    else if s=4 then Some l2
    else if s=5 then Some l1
    else if s=6 then Some l2
    else if s=7 then Some l2
    else if s=11 then None
    else if s=12 then None
    else if s=13 then None
    else if s=14 then Some l2
    else if s=15 then None
    else if s=16 then Some l1
    else if s=17 then Some l2
    else if s=18 then Some l1
    else None
  )"

text {* It is straightforward to show that this is an MDPN *}
interpretation MDPN[my\<Delta> my_mon]
  apply (unfold_locales)
  apply (unfold my\<Delta>_def)
  apply auto
  apply (unfold my_mon_def)
  apply simp_all
  apply blast+
  done

text {* And with the stuff proven above, we also get that this program is a 
  well-nested LDPN w.r.t. the start configuration @{term "[((0,{}),[1])]"}, which 
  corresponds to starting with procedure {\tt p1}.*}
interpretation EncodedLDPN_c0[my\<Delta> "[((0,{}),[1])]" "[[]]"]
  apply rule
  apply auto
  apply (unfold lstackm_s_def my_mon_def)
  apply simp
  done

end
