(* Additions to "RegSet.thy" from afp-entry "Functional Automata"
  * Substitutions 
  * Regular sets are a Kleene algebra. 

  TODO: Also want inverse homomorphisms !
*)


header "Regular Sets (Addendum)"

theory RegSet_add
(* FIXME: The import order between Kleene_algebras and RegSet is important here. 
  \<longrightarrow> Use qualified names!*)
imports Main 
        "$ISABELLE_HOME/src/HOL/SizeChange/Kleene_Algebras" 
        "common/Misc" 
        "../Functional-Automata/RegSet"
begin
text_raw {*\label{thy:regSet_add}*}

text {*
  In this theory, we define some additions to the theory ,,RegSet'' in the 
  ,,Functional-Automata'' entry of the archive of formal proofs.
  In particular
  \begin{itemize}
    \item We define a bunch of useful lemmas about concatenation and star, some of 
      them being declared as [simp],
    \item We show that languages with concatenation and star form a Kleene algebra,
    \item and we define substitution of languages, w.r.t. a map from letters to 
      languages.
  \end{itemize}
*}

subsection "Definitions"

subsubsection "Substitution"
text {* We define a substitution by extending a substitution on letters to words and 
  then to languages. *}
types ('a,'b) subst = "('a \<Rightarrow> 'b list set)"

text {* For this extension, we first need to generalize concatenation of languages to 
  lists of languages: *}
definition "conc_list Ls = foldl conc {[]} Ls"

definition subst_word :: "('a,'b) subst \<Rightarrow> 'a list \<Rightarrow> 'b list set" where
  "subst_word f w == conc_list (map f w)"

definition subst :: "('a \<Rightarrow> 'b list set) \<Rightarrow> 'a list set \<Rightarrow> 'b list set" where
  "subst f L == \<Union>subst_word f ` L"

subsubsection {* Projection and Inverse Projection *}
(* TODO: Do general homomorphism here *)
definition lang_proj :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a list set \<Rightarrow> 'b list set" where
  "lang_proj f L == map f ` L"

definition lang_inv_proj :: "('a \<Rightarrow> 'b) \<Rightarrow> 'b list set \<Rightarrow> 'a list set" where
  "lang_inv_proj f L == { w . map f w \<in> L }"


subsection "Theorems"

subsubsection "Concatenation"
  lemma concat_select_last: "\<lbrakk>
    w@[a] = concat (U@[l]); 
      !!l'. \<lbrakk>l=l'@[a]; w=concat (U@[l'])\<rbrakk> \<Longrightarrow> P; 
      l=[] \<Longrightarrow> P\<rbrakk> 
    \<Longrightarrow> P"
    apply auto
    apply (cases l rule: rev_cases)
    apply auto
    done

  lemma conc_assoc: "conc (conc A B) C = conc A (conc B C)"
    apply (unfold conc_def)
    apply auto
    apply force
    apply (rule_tac x="xs@xsa" in exI)
    apply auto
    done

  text {* Concatenation forms a (multiplicative) monoid where the language 
    @{term "{[]}"}, that solely contains the empty word, is the 1-element *}
  interpretation lc_monoid: monoid_mult["{[]}" conc]
    apply (unfold_locales) 
    apply (auto simp add: conc_assoc) [1]
    apply (auto simp add: conc_def)
    done

  declare lc_monoid.mult_assoc[simp add]

  text {* Languages with concatenation (*) and union (+) form a semiring 
    with 1-element @{term "{[]}"} and 0-element @{term "{}"}: *}
  interpretation regset_semiring: semiring_1["{[]}" conc "{}" "op \<union>"]
    by (unfold_locales) (auto simp add: conc_def)

  text {* Now follow some additional simplification, introduction and elimination 
    rules: *}
  lemma conc_single[simp]: 
    "conc {l} L = op @ l ` L"
    "conc L {l} = (\<lambda>x. x@l) ` L"
    by (unfold conc_def) auto

  lemma conc_list_simps[simp]: 
    "conc_list [] = {[]}"
    "conc_list (L#Ls) = conc L (conc_list Ls)" 
    apply (unfold conc_list_def)
    apply (auto simp add: lc_monoid.foldl_absorb1)
    done

  lemma conc_list_unconc[simp]: 
    "conc_list (L1@L2) = conc (conc_list L1) (conc_list L2)"
    by (induct L1) auto

  lemma conc_list_map_id[simp]: "conc_list (map (\<lambda>x. {[x]}) ll) = {ll}"
    by (induct ll) auto

  lemma concI: "\<lbrakk>a\<in>A; b\<in>B\<rbrakk> \<Longrightarrow> a@b\<in>conc A B"
    by (unfold conc_def) auto
  lemma concE: "\<lbrakk>ab\<in>conc A B; !!a b. \<lbrakk> ab=a@b; a\<in>A; b\<in>B \<rbrakk> \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
    by (unfold conc_def) auto



subsubsection "Star"
  lemmas star_consI_single = star.ConsI[of "[a]", simplified, standard]

  lemma star_refl[intro, simp]: "a\<in>A \<Longrightarrow> a\<in>star A"
    by (fact ConsI[OF _ NilI, simplified])

  lemma star_rev_consI[intro, simp]: "\<lbrakk> as \<in> star A; a\<in>A \<rbrakk> \<Longrightarrow> as@a \<in> star A"
    by (induct rule: star.induct) auto

  lemma in_star':
   "w \<in> star U \<longleftrightarrow> (\<exists>us. (\<forall>u \<in> set(us). u \<in> U \<and> u\<noteq>[]) \<and> (w = concat us))"
  apply (rule iffI)
   apply (erule star.induct)
    apply (rule_tac x = "[]" in exI)
    apply simp
   apply clarify
   apply (case_tac a)
     apply (rule_tac x="us" in exI)
     apply simp
   apply (rule_tac x = "a#us" in exI)
   apply simp
  apply clarify
  apply (simp add: concat_in_star)
  done

  lemma in_starE: "\<lbrakk> w\<in>star U; 
    !!us. \<lbrakk> !!u. u\<in>set us \<Longrightarrow> u\<in>U; w=concat us \<rbrakk> \<Longrightarrow> P 
    \<rbrakk> \<Longrightarrow> P"
    by (fastsimp simp add: in_star)

  lemma in_starE': "\<lbrakk> w\<in>star U; 
    !!us. \<lbrakk> !!u. u\<in>set us \<Longrightarrow> u\<in>U; !!u. u\<in>set us \<Longrightarrow> u\<noteq>[]; w=concat us \<rbrakk> \<Longrightarrow> P 
    \<rbrakk> \<Longrightarrow> P"
    by (fastsimp simp add: in_star')

  lemma star_rev_appendE: "\<lbrakk> w@[a]\<in>star U; 
    !!w1 w2. \<lbrakk>w=w1@w2; w1\<in>star U; w2@[a]\<in>U\<rbrakk> \<Longrightarrow> P
    \<rbrakk> \<Longrightarrow> P"
  proof -
    assume A: "w @ [a] \<in> RegSet.star U"
    assume B: "!!w1 w2. \<lbrakk>w=w1@w2; w1\<in>star U; w2@[a]\<in>U\<rbrakk> \<Longrightarrow> P"
    from in_starE'[OF A] obtain us where 
      CNC: "!!u. u\<in>set us \<Longrightarrow> u\<in>U" "!!u. u\<in>set us \<Longrightarrow> u\<noteq>[]" "w@[a]=concat us" 
      by blast
    obtain us' u' where [simp]: "us=us'@[u']" 
      using CNC(3) by (cases us rule: rev_cases) auto
    with CNC(2,3) obtain u'' where [simp]: "u'=u''@[a]" 
      by (cases u' rule: rev_cases) (auto)
    from CNC(3) have [simp]: "w = concat us' @ u''" by simp
    with CNC(1) have "concat us' \<in> star U" by (auto simp add: in_star)
    moreover from CNC(1) have "u''@[a] \<in> U" by auto
    ultimately show ?thesis by (auto intro: B)
  qed  

  lemma star_rev_induct[consumes 1, induct set, case_names empty concat]: 
    "\<lbrakk>w \<in> RegSet.star A; 
      P []; 
      \<And>w1 w2. \<lbrakk>w1 \<in> RegSet.star A; w2 \<in> A; P w1\<rbrakk> \<Longrightarrow> P (w1@w2)
    \<rbrakk> \<Longrightarrow> P w"
    apply (induct w rule: length_induct)
    apply (case_tac xs rule: rev_cases)
    apply (auto elim: star_rev_appendE)
    done

  text {* With the next few lemmas we show that languages with concatenation and star 
    form a Kleene algebra *}
  lemma star_ka_helper1: "\<lbrakk>
    xs \<in> RegSet.star a; 
    {xs @ ys |xs ys. xs \<in> a \<and> ys \<in> x} \<subseteq> x; 
    ys \<in> x\<rbrakk>
    \<Longrightarrow> xs @ ys \<in> x"
    by (induct rule: star.induct) auto

  lemma star_ka_helper2: "\<lbrakk>
    ys \<in> RegSet.star a; 
    {xs @ ys |xs ys. xs \<in> x \<and> ys \<in> a} \<subseteq> x; 
    xs \<in> x\<rbrakk> 
    \<Longrightarrow> xs @ ys \<in> x"
  proof (induct rule: star_rev_induct) 
    case empty thus ?case by auto
  next
    case (concat w1 w2)
    from concat.prems(1) have CR: "!!xs ys. \<lbrakk>xs\<in>x; ys\<in>a\<rbrakk> \<Longrightarrow> xs@ys \<in> x" by blast
    from concat have "xs@w1\<in>x" by blast
    from CR[OF this concat.hyps(2)] show ?case by simp
  qed

  interpretation kleene["op \<subseteq>" "op \<subset>" "op \<union>" "{[]}" conc "{}" star]
    apply (unfold_locales)
    apply auto
    apply (unfold conc_def)
    apply (auto intro: star_ka_helper1 star_ka_helper2)
    done



subsubsection "Substitution"

lemma subst_id[simp]: "subst (\<lambda>x. {[x]}) L = L" 
  by (auto simp add: subst_def subst_word_def)

lemma subst_single[simp]: "subst f {[a]} = f a" 
  by (auto simp add: subst_def subst_word_def)

lemma subst_word_single[simp]: "subst_word (\<lambda>x. {[f x]}) w = {map f w}"
  by (induct w) (auto simp add: subst_word_def)

text {* Projection and inverse projection can be expressed as special substitutions: *}
lemma subst_proj[simp]: "subst (\<lambda>x. {[f x]}) L = lang_proj f L"
  by (unfold lang_proj_def subst_def) auto

lemma subst_inv_proj[simp]: "subst (\<lambda>x. {[y]| y. f y = x}) L = lang_inv_proj f L"
proof safe
  fix w assume "w\<in>subst (\<lambda>x. {[y]| y. f y = x}) L"
  then obtain w' where 
    W'IL: "w'\<in>L" and 
    S: "w\<in>subst_word (\<lambda>x. {[y]| y. f y = x}) w'" 
    by (auto simp add: subst_def)
  from S have "map f w = w'" 
    by (induct w' arbitrary: w) (auto simp add: subst_word_def conc_def)
  with W'IL show "w\<in>lang_inv_proj f L" 
    by (auto simp add: lang_inv_proj_def)
next
  fix w assume "w\<in>lang_inv_proj f L"
  then obtain w' where 
    W'IL: "w'\<in>L" and 
    S: "map f w = w'" 
    by (auto simp add: lang_inv_proj_def)

  have [simp]: 
    "!!a w aa W. a#w \<in> conc {[y] |y. f y = aa} W \<longleftrightarrow> f a = aa \<and> w\<in>W"
    apply (auto simp add: conc_def)
    apply (rule_tac x="[a]" in exI)
    by auto

  from S have "w\<in>subst_word (\<lambda>x. {[y]| y. f y = x}) w'" 
    apply (induct w arbitrary: w') 
    apply (simp add: subst_word_def)
    apply (case_tac w')
    apply (simp_all add: subst_word_def)
    done
  with W'IL show "w\<in>subst (\<lambda>x. {[y] |y. f y = x}) L" 
    by (unfold subst_def) auto
qed

text {* Substitution distributes over concatenation: *}
lemma subst_distr_conc[simp]: 
  "subst f (conc L1 L2) = conc (subst f L1) (subst f L2)"
  apply (unfold subst_def conc_def)
  apply auto
  apply (unfold subst_word_def)
  apply auto
  apply (fastsimp simp add: conc_def)
  apply (rule_tac x="?x" in exI)
  apply safe
  apply (rule_tac x="xa" in exI)
  apply (rule_tac x="xb" in exI)
  apply auto
  apply (unfold conc_def)
  apply auto
  done

text {* Substitution distributes over union: *}
lemma subst_distr_Un[simp]: 
  "subst f (L1 \<union> L2) = (subst f L1) \<union> (subst f L2)"
  by (unfold subst_def) auto


end
