header "Semantics"
theory Semantics
imports DPN RegSet_add
begin
text_raw {*\label{thy:Semantics}*}

text {*
  In this theory, we define an interleaving and a tree-based semantics of DPNs.
  We show the equivalence of the two semantics.
*}

subsection "Interleaving Semantics"
text {*
  The interleaving semantics models the execution of a DPN on a single processor, that 
  makes one step at a time, and may switch the currently executed process after each 
  step. This is the original semantics of DPNs \cite{BMT05}.
*}

text {*
  The interleaving semantics is formalized by means of a labeled transition system. 
  A single process is modeled as a pair of its control state and its stack.

  A configuration of the DPN is modeled as a list of processes. Note that we use 
  lists of processes here, rather than multisets, to enable representation of 
  configurations as regular sets, as required by the algorithms of \cite{BMT05}.
*}
types 
  ('Q,'\<Gamma>) pconf = "'Q \<times> '\<Gamma> list"
  ('Q,'\<Gamma>) conf = "('Q,'\<Gamma>) pconf list"

text {*
  The (single-) step relation @{text dpntr} of the interleavings semantics is 
  defined as the least solution of the following constraints:
*}
inductive_set dpntr :: "('Q,'\<Gamma>,'L) dpn \<Rightarrow> (('Q,'\<Gamma>) conf \<times> 'L \<times> ('Q,'\<Gamma>) conf) set" 
  for \<Delta> where
  -- "A non-spawning step modifies a single pushdown process 
      according to a non-spawning rule in the DPN:"
  dpntr_no_spawn: 
    "(p,\<gamma> \<hookrightarrow>\<^sub>l p',w)\<in>\<Delta> \<Longrightarrow> 
       (c1@(p,\<gamma>#r)#c2,l,c1@(p',w@r)#c2) \<in> dpntr \<Delta>" |
  -- "A spawning step modifies a pushdown process 
      according to a spawning rule in the DPN and adds the 
      spawned process immediately before the spawning process:"
  dpntr_spawn: 
    "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w)\<in>\<Delta> \<Longrightarrow> 
       (c1@(p,\<gamma>#r)#c2,l,c1@(ps,ws)#(p',w@r)#c2) \<in> dpntr \<Delta>"

text {*
  We denote the reflexive, transitive closure of the single-step relation 
  by @{text dpntrc}: 
*}
abbreviation "dpntrc M == trcl (dpntr M)"

subsection "Tree Semantics"

text {*
  Now we regard a true concurrency semantics, where an execution does not contain 
  the interleaving between independent steps. When starting at a single process, 
  we model such an execution as a tree, where each node corresponds to an applied 
  step. 
  A node corresponding to a non-spawning step has one successor, a node corresponding 
  to a spawning step has two successors. We annotate the leafs of the tree by the 
  configuration of the reached process.

  When starting at a configuration consisting of (a list of) multiple processes, we 
  model the execution as a list of multiple execution trees, one for each process.
*}

datatype ('Q,'\<Gamma>,'L) ex_tree = 
  NLEAF "('Q,'\<Gamma>) pconf" | 
  NNOSPAWN 'L "('Q,'\<Gamma>,'L) ex_tree" | 
  NSPAWN 'L "('Q,'\<Gamma>,'L) ex_tree" "('Q,'\<Gamma>,'L) ex_tree" 

types ('Q,'\<Gamma>,'L) ex_hedge = "('Q,'\<Gamma>,'L) ex_tree list"

inductive tsem 
  :: "('Q,'\<Gamma>,'L) dpn \<Rightarrow> ('Q,'\<Gamma>) pconf \<Rightarrow> ('Q,'\<Gamma>,'L) ex_tree \<Rightarrow> ('Q,'\<Gamma>) conf \<Rightarrow> bool" 
  for \<Delta> where
  tsem_leaf[simp, intro!]: 
    "tsem \<Delta> pw (NLEAF pw) [pw]" |
  tsem_nospawn: 
    "\<lbrakk> (p,\<gamma> \<hookrightarrow>\<^sub>l p',w)\<in>\<Delta>; tsem \<Delta> (p',w@r) t c' \<rbrakk> \<Longrightarrow> 
       tsem \<Delta> (p,\<gamma>#r) (NNOSPAWN l t) c'" |
  tsem_spawn: 
    "\<lbrakk> (p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w)\<in>\<Delta>; tsem \<Delta> (ps,ws) ts cs; tsem \<Delta> (p',w@r) t c' \<rbrakk> \<Longrightarrow> 
       tsem \<Delta> (p,\<gamma>#r) (NSPAWN l ts t) (cs@c')"

inductive hsem 
  :: "('Q,'\<Gamma>,'L) dpn \<Rightarrow> ('Q,'\<Gamma>) conf \<Rightarrow> ('Q,'\<Gamma>,'L) ex_hedge \<Rightarrow> ('Q,'\<Gamma>) conf \<Rightarrow> bool" 
  for \<Delta> where
  hsem_empty[simp, intro!]: "hsem \<Delta> [] [] []" |
  hsem_cons: "\<lbrakk>tsem \<Delta> \<pi> t cf'; hsem \<Delta> c h c'\<rbrakk> \<Longrightarrow> hsem \<Delta> (\<pi>#c) (t#h) (cf'@c')"

text {*
  In the following we show some basic facts about the 
  @{text tsem}- and @{text hsem}-relations.
*}

lemma hsem_empty_h[simp]: 
  "hsem \<Delta> c [] c' \<longleftrightarrow> c=[] \<and> c'=[]" 
  by (auto elim: hsem.cases intro: hsem.intros)

lemma hsem_length: "hsem \<Delta> c h c' \<Longrightarrow> length c = length h"
  by (induct rule: hsem.induct) auto

text {*
  The hedges and configurations of the hedge semantics can be concatenated.
*}
lemmas hsem_cons_single = hsem_cons[where cf'="[\<pi>']", simplified, standard]

lemma hsem_conc: "\<lbrakk>hsem \<Delta> c1 h1 c1'; hsem \<Delta> c2 h2 c2'\<rbrakk> \<Longrightarrow> 
  hsem \<Delta> (c1@c2) (h1@h2) (c1'@c2')"
  by (induct c1 h1 c1' rule: hsem.induct) (auto intro: hsem_cons)

lemmas hsem_conc_lel = hsem_conc[OF _ hsem_cons]

lemmas hsem_conc_leel = hsem_conc[OF _ hsem_cons[OF _ hsem_cons]]

lemma tsem_not_empty[simp]: "\<not> tsem \<Delta> \<pi> t []"
  by (induct t arbitrary: \<pi>) (auto elim: tsem.cases)
lemma hsem_empty_simps1[simp]:
  "hsem \<Delta> [] h c' \<longleftrightarrow> (h=[] \<and> c'=[])" 
  "hsem \<Delta> c h [] \<longleftrightarrow> (c=[] \<and> h=[])"
  by (auto elim: hsem.cases)

lemma hsem_id[simp, intro!]: "hsem \<Delta> c (map NLEAF c) c" 
  by (induct c) (auto intro: hsem_cons_single)
lemmas hsem_id'[simp, intro!] = hsem_id[of _ "\<pi>#c", simplified, standard]


text {*
  Given a partition of the starting configuration, we can construct a corresponding
  partition of the hedge and the final configuration.
*}
lemma hsem_split': 
  "\<lbrakk>hsem \<Delta> (c1@c2) h c'\<rbrakk> \<Longrightarrow> \<exists>h1 h2 c1' c2'. 
                                h=h1@h2 \<and> c'=c1'@c2' \<and> 
                                hsem \<Delta> c1 h1 c1' \<and> hsem \<Delta> c2 h2 c2'"
proof (induct c1 arbitrary: c2 h c')
  case Nil hence "h=[]@h" "c'=[]@c'" "hsem \<Delta>  [] [] []" "hsem \<Delta> c2 h c'" 
    by (auto intro: hsem.intros)
  with Nil show ?case by blast
next
  case (Cons p c1)
  from Cons.prems[simplified] show ?case 
  proof (cases rule: hsem.cases)
    case hsem_empty hence False by simp thus ?thesis ..
  next
    case (hsem_cons px t ct' c hx cx') 
    hence CC: "h=t#hx" "tsem \<Delta> p t ct'" "hsem \<Delta> (c1@c2) hx cx'" "c'=ct'@cx'" 
      by simp_all
    from Cons.hyps[OF CC(3)] obtain h1 h2 c1' c2' where 
      IHAPP: "hx=h1@h2" "cx'=c1'@c2'" "hsem \<Delta> c1 h1 c1'" "hsem \<Delta> c2 h2 c2'" 
      by blast
    have "h=(t#h1)@h2" "c'=(ct'@c1')@c2'" using CC IHAPP 
      by simp_all
    with hsem.intros(2)[OF CC(2) IHAPP(3)] IHAPP(4) show ?thesis by blast
  qed
qed

lemma hsem_split[consumes 1]: "\<lbrakk>hsem \<Delta> (c1@c2) h c'; 
  !!h1 h2 c1' c2'. 
    \<lbrakk>h=h1@h2; c'=c1'@c2'; hsem \<Delta> c1 h1 c1'; hsem \<Delta> c2 h2 c2'\<rbrakk> \<Longrightarrow> P 
  \<rbrakk> \<Longrightarrow> P" 
  by (blast dest: hsem_split')

lemma hsem_single: 
  "\<lbrakk>hsem \<Delta> [\<pi>] h c'; !!t. \<lbrakk> h=[t]; tsem \<Delta> \<pi> t c' \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
  by (auto intro: hsem.intros elim!: hsem.cases)

lemma hsem_split_single[consumes 1]: "\<lbrakk>hsem \<Delta> (\<pi>#c2) h c'; 
  !!t1 h2 c1' c2'. 
    \<lbrakk>h=t1#h2; c'=c1'@c2'; tsem \<Delta> \<pi> t1 c1'; hsem \<Delta> c2 h2 c2'\<rbrakk> \<Longrightarrow> P 
  \<rbrakk> \<Longrightarrow> P"
  by (fastsimp elim: hsem_split[where ?c1.0="[\<pi>]", simplified] hsem_single)

lemma hsem_lel: "\<lbrakk> hsem \<Delta> (c1@\<pi>#c2) h c'; 
    !!h1 t h2 c1' ct' c2'. \<lbrakk>
      h=h1@t#h2; c'=c1'@ct'@c2'; 
      hsem \<Delta> c1 h1 c1'; tsem \<Delta> \<pi> t ct'; hsem \<Delta> c2 h2 c2' 
    \<rbrakk> \<Longrightarrow> P 
  \<rbrakk> \<Longrightarrow> P"
  by (fastsimp elim: hsem_split hsem_split_single)

  
text {*
  Given a partition of the hedge, we can construct a corresponding partition of 
  the initial and final configuration.
*}
lemma hsem_split_h': "\<lbrakk>hsem \<Delta> c (h1@h2) c'\<rbrakk> \<Longrightarrow> 
  \<exists>c1 c2 c1' c2'. c=c1@c2 \<and> c'=c1'@c2' \<and> 
                  hsem \<Delta> c1 h1 c1' \<and> hsem \<Delta> c2 h2 c2'"
proof (induct h1 arbitrary: h2 c c')
  case Nil hence "c=[]@c" "c'=[]@c'" "hsem \<Delta> [] [] []" "hsem \<Delta> c h2 c'" 
    by (auto intro: hsem.intros)
  with Nil show ?case by blast
next
  case (Cons t h1)
  from Cons.prems[simplified] show ?case proof (cases rule: hsem.cases)
    case hsem_empty hence False by simp thus ?thesis ..
  next
    case (hsem_cons p tx ct' cx hx cx') 
    hence CC: "c=p#cx" "tsem \<Delta> p t ct'" "hsem \<Delta> cx (h1@h2) cx'" "c'=ct'@cx'" 
      by simp_all
    from Cons.hyps[OF CC(3)] obtain c1 c2 c1' c2' where 
      IHAPP: "cx=c1@c2" "cx'=c1'@c2'" "hsem \<Delta>  c1 h1 c1'" "hsem \<Delta> c2 h2 c2'" 
      by blast
    have "c=(p#c1)@c2" "c'=(ct'@c1')@c2'" using CC IHAPP by simp_all
    with hsem.intros(2)[OF CC(2), OF IHAPP(3)] IHAPP(4) show ?thesis 
      by blast
  qed
qed

lemma hsem_split_h: 
  "\<lbrakk> hsem \<Delta> c (h1@h2) c'; 
     !!c1 c2 c1' c2'. 
       \<lbrakk>c=c1@c2; c'=c1'@c2'; hsem \<Delta> c1 h1 c1'; hsem \<Delta> c2 h2 c2'\<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
  by (blast dest: hsem_split_h')

lemma hsem_single_h: 
  "\<lbrakk>hsem \<Delta> c [t] c'; !!p. \<lbrakk> c=[p]; tsem \<Delta> p t c' \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
  by (force intro: hsem.intros elim!: hsem.cases)
  
lemmas hsem_split_h_single = hsem_split_h[where ?h1.0="[t]", simplified, standard]

lemma hsem_lel_h: "\<lbrakk> hsem \<Delta> c (h1@t#h2) c'; 
    !!c1 p c2 c1' ct' c2'. \<lbrakk>
      c=c1@p#c2; c'=c1'@ct'@c2'; 
      hsem \<Delta> c1 h1 c1'; tsem \<Delta> p t ct'; hsem \<Delta> c2 h2 c2' 
    \<rbrakk> \<Longrightarrow> P 
  \<rbrakk> \<Longrightarrow> P"
  by (fastsimp elim!: hsem_split_h hsem_split_h_single hsem_single_h)


subsubsection "Scheduler"
text {*
  The scheduler maps execution hedges to compatible label sequences. This is done by 
  eating up the given hedge from the roots to the leafs, until all non-leaf nodes 
  have been consumed. From an ordering point of view, the hedge represents a partial 
  ordering on the steps, and the scheduler maps this ordering to the set of all its 
  topological sorts.
*}

text {*
  An execution hedge is called {\em final} if it solely consists of leaf nodes. 
*}
inductive final_t where
  [simp, intro!]: "final_t (NLEAF pw)" 

lemma [simp, intro!]:
  "\<not>final_t (NNOSPAWN l t)"
  "\<not>final_t (NSPAWN l ts t)"
  by (auto elim: final_t.cases)

abbreviation "final == list_all final_t"

text {*
  Final execution hedges contain no steps, hence they do not change the configuration.
*}
lemma final_tsem_nostep: "\<lbrakk>final_t t; tsem \<Delta> pw t c'\<rbrakk> \<Longrightarrow> c'=[pw]"
  by (cases t) (auto elim: tsem.cases)

lemma final_hsem_nostep: "\<lbrakk>final h; hsem \<Delta> c h c'\<rbrakk> \<Longrightarrow> c'=c"
  apply (rotate_tac)
  apply (induct rule: hsem.induct)
  apply (auto intro: final_tsem_nostep)
  done

text {*
  As described above, the scheduler eats up the execution hedge from the roots to 
  the leafs, until there are no inner nodes remaining, i.e. the hedge is final.
*}
inductive sched :: "('Q,'\<Gamma>,'L) ex_hedge \<Rightarrow> 'L list \<Rightarrow> bool" where
  sched_final: "final h \<Longrightarrow> sched h []" |
  sched_nospawn: 
    "sched (h1@t#h2) w \<Longrightarrow> sched (h1@(NNOSPAWN l t)#h2) (l#w)" |
  sched_spawn: 
    "sched (h1@ts#t#h2) w \<Longrightarrow> sched (h1@(NSPAWN l ts t)#h2) (l#w)"

inductive_set sched_rel :: "(('Q,'\<Gamma>,'L) ex_hedge, 'L) LTS" where
  sched_rel_nospawn: "((h1@(NNOSPAWN l t)#h2),l,h1@t#h2)\<in>sched_rel" |
  sched_rel_spawn: "((h1@(NSPAWN l ts t)#h2),l,(h1@ts#t#h2))\<in>sched_rel"

definition "sched' h ll == (\<exists>h'. (h,ll,h')\<in>trcl sched_rel \<and> final h')"

lemma sched_alt1: "sched h ll \<Longrightarrow> sched' h ll"
  by (unfold sched'_def, induct rule: sched.induct) 
     (auto intro: trcl.intros sched_rel.intros)

lemma sched_rel_alt2: "\<lbrakk>(h,ll,h')\<in>trcl sched_rel; final h'\<rbrakk> \<Longrightarrow> sched h ll"
  by (induct rule: trcl.induct) (auto intro: sched.intros elim: sched_rel.cases)

lemma sched_alt: "sched' h ll \<longleftrightarrow> sched h ll"
  by (unfold sched'_def, auto intro: sched_alt1[unfolded sched'_def] sched_rel_alt2)


text {* We now show some basic facts about the scheduler. *}
lemma sched_empty_seq[simp]: "sched h [] \<longleftrightarrow> final h"
  by (auto intro: sched_final elim: sched.cases)

lemma sched_empty_hedge[simp]: "sched [] ll \<longleftrightarrow> ll=[]"
  by (auto intro: sched_final elim: sched.cases)

lemma sched_empty_empty[simp, intro!]: "sched [] []" by (auto intro: sched_final)

lemma sched_final_simp[simp]: "final h \<Longrightarrow> sched h c \<longleftrightarrow> c=[]"
  by (auto elim: sched.cases)


text {*
  In the following few lemmas we derive an induction scheme that reasons about hedges 
  in the way they are consumed by the scheduler
*}

fun sched_ind_size where
  "sched_ind_size (NLEAF \<pi>) = 0" |
  "sched_ind_size (NNOSPAWN l t) = Suc (sched_ind_size t)" |
  "sched_ind_size (NSPAWN l ts t) = Suc (sched_ind_size ts + sched_ind_size t)"
  
abbreviation "sched_ind_sizeh h == listsum (map sched_ind_size h)"

lemma sched_ind_h_cases[consumes 1, case_names NOSPAWN SPAWN]: 
  "\<lbrakk> sched_ind_sizeh h > 0; 
     !!h1 l t h2. h=h1@(NNOSPAWN l t)#h2 \<Longrightarrow> P; 
     !!h1 ts t h2 l. h = h1@(NSPAWN l ts t)#h2 \<Longrightarrow> P 
  \<rbrakk> \<Longrightarrow> P"
proof (induct h)
  case Nil thus ?case by auto
next
  case (Cons t h)
  show ?case proof (cases t)
    case (NLEAF \<pi>) 
    with Cons.prems(1) have I: "0 < sched_ind_sizeh h" by simp
    show ?thesis proof (rule Cons.hyps[OF I])
      fix h1 l tt h2 
      assume "h=h1 @ NNOSPAWN l tt # h2" 
      hence "t#h = (t#h1) @ NNOSPAWN l tt # h2" by simp 
      with Cons.prems(2) show ?thesis by blast
    next
      fix h1 ts tt h2 l 
      assume "h = h1 @ NSPAWN l ts tt # h2" 
      hence "t#h = (t#h1) @ NSPAWN l ts tt # h2" by simp 
      with Cons.prems(3) show ?thesis by blast
    qed
  next
    case (NNOSPAWN L tt) 
    with Cons.prems(2)[of "[]", simplified] show ?thesis by auto
  next
    case (NSPAWN L ts tt) 
    with Cons.prems(3)[of "[]", simplified] show ?thesis by auto
  qed
qed
      
lemma sched_ind_helper: 
  "\<lbrakk> !!h. final h \<Longrightarrow> P h; 
     !!h1 t h2 l. P (h1@t#h2) \<Longrightarrow> P (h1@(NNOSPAWN l t)#h2); 
     !!h1 ts t h2 l. P (h1@ts#t#h2) \<Longrightarrow> P (h1@(NSPAWN l ts t)#h2); 
     sched_ind_sizeh h = k 
   \<rbrakk> \<Longrightarrow> P h"
proof (induct k arbitrary: h)
  case 0 note C=this from C(4) have "final h" 
    apply (induct h)
    apply simp
    apply (case_tac a)
    apply auto
    done
  with C(1) show ?case by blast
next
  case (Suc k) hence S: "sched_ind_sizeh h > 0" by simp
  thus ?case proof (cases rule: sched_ind_h_cases)
    case (NOSPAWN h1 l t h2) 
    with Suc.prems(4) have I: "sched_ind_sizeh (h1@t#h2) = k" by simp
    with Suc.prems(1,2,3) NOSPAWN show ?thesis 
      by (drule_tac Suc.hyps) blast+
  next
    case (SPAWN h1 ts t h2 l) 
    with Suc.prems(4) have I: "sched_ind_sizeh (h1@ts#t#h2) = k" by simp
    with Suc.prems(1,2,3) SPAWN show ?thesis 
      by (drule_tac Suc.hyps) blast+
  qed
qed

lemma sched_ind[case_names FINAL NOSPAWN SPAWN]: 
  "\<lbrakk> !!h. final h \<Longrightarrow> P h; 
     !!h1 t h2 l. P (h1@t#h2) \<Longrightarrow> P (h1@(NNOSPAWN l t)#h2); 
     !!h1 ts t h2 l. P (h1@ts#t#h2) \<Longrightarrow> P (h1@(NSPAWN l ts t)#h2) 
  \<rbrakk> \<Longrightarrow> P h"
  using sched_ind_helper by blast

text {*
  Every tree/hedge has at least one schedule. From an ordering point of view, 
  this is because hedge-structures are acyclic, and thus have always at least one 
  topological sort. However, using the inductive definition of the scheduler, the 
  proof of this lemma is by straightforward induction.
*}
lemma exists_schedule: "\<lbrakk>!!ll. sched h ll \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
  by (induct h rule: sched_ind) (auto intro: sched.intros)
  

text {*
  Next, we want to show that the true concurrency semantics corresponds to the 
  interleaving semantics. For this purpose, we show that we have an execution with 
  labeling sequence @{text ll} in the interleaving semantics if and only if 
  there is an execution @{text h} in the true concurrency semantics that has 
  @{text ll} in its set of schedules.

  The next two lemmas show the two directions of this claim.
*}
lemma sched_correct1: "(c,ll,c')\<in>dpntrc \<Delta> \<Longrightarrow> \<exists>h. hsem \<Delta> c h c' \<and> sched h ll"
proof (induct rule: trcl.induct)
  case (empty c) thus ?case by (induct c) (auto intro: hsem_cons_single)
next
  case (cons c l ch ll c')
  from cons.hyps(3) obtain h where IHAPP: "hsem \<Delta> ch h c'" "sched h ll" by blast
  from cons.hyps(1) show ?case 
  proof (cases)
    case (dpntr_no_spawn p \<gamma> la p' w c1 r c2) 
    hence 
      C_simp[simp]: "c = c1 @ (p, \<gamma> # r) # c2" "ch = c1 @ (p', w @ r) # c2" and 
      C: "(p,\<gamma> \<hookrightarrow>\<^sub>l p',w) \<in> \<Delta>" 
      by auto
    from hsem_lel[OF IHAPP(1)[simplified]] obtain h1 t h2 c1' ct' c2' where 
      [simp]: "h = h1 @ t # h2" "c' = c1' @ ct' @ c2'" and 
      HSPLIT: "hsem \<Delta> c1 h1 c1'" "tsem \<Delta> (p', w @ r) t ct'" "hsem \<Delta> c2 h2 c2'" 
      .
    from tsem_nospawn[OF C HSPLIT(2)] have 
      ST: "tsem \<Delta> (p,\<gamma>#r) (NNOSPAWN l t) ct'" .
    from hsem_conc_lel[OF HSPLIT(1) ST HSPLIT(3)] have 
      "hsem \<Delta> c (h1 @ NNOSPAWN l t # h2) c'" 
      by simp
    moreover from sched_nospawn[OF IHAPP(2)[simplified]] have 
      "sched (h1 @ NNOSPAWN l t # h2) (l#ll)" .
    ultimately show ?thesis by blast
  next
    case (dpntr_spawn p \<gamma> la ps ws p' w c1 r c2) 
    hence 
      [simp]: "c = c1 @ (p, \<gamma> # r) # c2" 
              "ch = c1 @ (ps, ws) # (p', w @ r) # c2" and 
        C: "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w) \<in> \<Delta>" 
      by auto
    from IHAPP(1)[simplified] obtain h1 ts t h2 c1' cs' ct' c2' where 
      [simp]: "h = h1 @ ts # t # h2" "c' = c1' @ cs' @ ct' @ c2'" and
      HSPLIT: "hsem \<Delta> c1 h1 c1'" "tsem \<Delta> (ps,ws) ts cs'" 
              "tsem \<Delta> (p', w @ r) t ct'" "hsem \<Delta> c2 h2 c2'" 
      by (fastsimp elim: hsem_split hsem_split_single)
    from tsem_spawn[OF C HSPLIT(2,3)] have 
      ST: "tsem \<Delta> (p,\<gamma>#r) (NSPAWN l ts t) (cs'@ct')" .
    from hsem_conc_lel[OF HSPLIT(1) ST HSPLIT(4)] have 
      "hsem \<Delta> c (h1 @ NSPAWN l ts t # h2) c'" by simp
    moreover from sched_spawn[OF IHAPP(2)[simplified]] have 
      "sched (h1 @ NSPAWN l ts t # h2) (l#ll)" .
    ultimately show ?thesis by blast
  qed
qed
    

lemma sched_correct2: "\<lbrakk> sched h ll; hsem \<Delta> c h c' \<rbrakk> \<Longrightarrow> (c,ll,c')\<in>dpntrc \<Delta>"
proof (induct h ll arbitrary: c c' rule: sched.induct)
  case (sched_final h c c') thus ?case by (auto dest: final_hsem_nostep)
next
  case (sched_nospawn h1 t h2 ll l c c') 
  from hsem_lel_h[OF sched_nospawn.prems] obtain c1 p\<gamma>r c2 c1' ct' c2' where 
    [simp]: "c = c1 @ p\<gamma>r # c2" "c' = c1' @ ct' @ c2'" and 
      SPLIT: "hsem \<Delta> c1 h1 c1'" 
             "tsem \<Delta> p\<gamma>r (NNOSPAWN l t) ct'" 
             "hsem \<Delta> c2 h2 c2'" 
    .
  from SPLIT(2) obtain p \<gamma> r p' w where 
    [simp]: "p\<gamma>r=(p,\<gamma>#r)" and 
      ST: "(p,\<gamma> \<hookrightarrow>\<^sub>l p',w)\<in>\<Delta>" "tsem \<Delta> (p',w@r) t ct'" 
    by (erule_tac tsem.cases) fastsimp+
  from dpntr_no_spawn[OF ST(1)] have "(c,l,c1 @ (p', w @ r) # c2)\<in>dpntr \<Delta>" by auto
  also from sched_nospawn.hyps(2)[OF hsem_conc_lel[OF SPLIT(1) ST(2) SPLIT(3)]] have 
    SST: "(c1 @ (p', w @ r) # c2, ll, c1' @ ct' @ c2') \<in> dpntrc \<Delta>" .
  finally show ?case by auto
next
  case (sched_spawn h1 ts t h2 ll l c c')
  from hsem_lel_h[OF sched_spawn.prems] obtain c1 p\<gamma>r c2 c1' ct' c2' where 
    [simp]: "c = c1 @ p\<gamma>r # c2" "c' = c1' @ ct' @ c2'" and 
      SPLIT: "hsem \<Delta> c1 h1 c1'" 
             "tsem \<Delta> p\<gamma>r (NSPAWN l ts t) ct'" 
             "hsem \<Delta> c2 h2 c2'" 
    .
  from SPLIT(2) obtain p \<gamma> r ps ws p' w cts' ctt' where 
    [simp]: "p\<gamma>r=(p,\<gamma>#r)" "ct'=cts'@ctt'" and 
      ST: "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w)\<in>\<Delta>" "tsem \<Delta> (ps,ws) ts cts'" 
          "tsem \<Delta> (p',w@r) t ctt'" 
    by (erule_tac tsem.cases) fastsimp+
  from dpntr_spawn[OF ST(1)] have 
    "(c,l,c1 @ (ps,ws) # (p', w @ r) # c2)\<in>dpntr \<Delta>" 
    by auto
  also from sched_spawn.hyps(2)[OF hsem_conc_leel[OF SPLIT(1) ST(2,3) SPLIT(3)]] have 
    SST: "(c1 @ (ps,ws) # (p', w @ r) # c2, ll, c') \<in> dpntrc \<Delta>" 
    by simp
  finally show ?case by auto
qed

text {* Finally, we formulate the correspondance between the interleaving and the 
  true concurrency semantics as a single equivalence: *}
theorem sched_correct: "(c,ll,c')\<in>dpntrc \<Delta> \<longleftrightarrow> (\<exists>h. hsem \<Delta> c h c' \<and> sched h ll)"
  by (auto intro: sched_correct1 sched_correct2)

text {* As any hedge has at least one schedule, we always get an interleaving 
  execution from a hedge execution: *}
lemma obtain_schedule: "
  \<lbrakk> hsem \<Delta> c h c'; 
    !!ll. \<lbrakk>(c,ll,c')\<in>dpntrc \<Delta>; sched h ll \<rbrakk> \<Longrightarrow> P 
  \<rbrakk> \<Longrightarrow> P"
  apply (rule_tac h=h in exists_schedule)
  apply (metis sched_correct)
  done


section "Predecessor Sets"
text {*
  Following \cite{BMT05}, we define the set of immediate 
  predecessors @{text "pre \<Delta> C"} and predecessors @{text "pre\<^sup>* \<Delta> C"} of a set of 
  configurations @{text C}. The set of immediate predecessors contains those 
  configurations from that we can reach (a configuration in) @{text C} with exactly 
  one step. The set of predecessors contains those configurations from that we can 
  reach @{text C} with an arbitrary number of steps, including no steps at all 
  (i.e. @{text "pre\<^sup>*"} is reflexive).

  Computing predecessor sets is the key to model checking and analysis of DPNs, 
  see \cite{BMT05} for details.
*}

definition "pre \<Delta> C' == { c . \<exists>l c'. c'\<in>C' \<and> (c,l,c') \<in> dpntr \<Delta> }"
definition pre_star ("pre\<^sup>*") where 
  "pre\<^sup>* \<Delta> C' == { c . \<exists>ll c'. c'\<in>C' \<and> (c,ll,c') \<in> dpntrc \<Delta> }"

subsection "Hedge-Constrained Predecessor Sets"
text {*
  For a set of configurations @{text "C'"} and a set of execution hedges @{text H}, 
  we define the {\em hedge-constrained predecessor set} of @{text "C'"} 
  w.r.t. @{text H} as the set of those configurations from that we can reach
  @{text C'} with an execution hedge in @{text H}.
*}
definition "prehc \<Delta> H C' == { c . \<exists>h c'. h\<in>H \<and> c'\<in>C' \<and> hsem \<Delta> c h c' }"

lemma prehcI: "\<lbrakk>h\<in>H; c'\<in>C'; hsem \<Delta> c h c'\<rbrakk> \<Longrightarrow> c\<in>prehc \<Delta> H C'"
  by (unfold prehc_def) auto

lemma prehcE: 
  "\<lbrakk>c\<in>prehc \<Delta> H C'; !!h c'. \<lbrakk> h\<in>H; c'\<in>C'; hsem \<Delta> c h c' \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
  by (unfold prehc_def) auto

text {*
  The hedge-constrained predecessor set is monotonic in the constraint
*}
lemma prehc_mono: "H\<subseteq>H' \<Longrightarrow> prehc \<Delta> H C' \<subseteq> prehc \<Delta> H' C'"
  by (auto simp add: prehc_def)

text {*
  The hedge-constrained predecessor set without constraints is the same as the 
  original predecessor set.
*}
lemma prehc_triv_is_pre_star: "prehc \<Delta> UNIV C' = pre\<^sup>* \<Delta> C'"
  apply (unfold prehc_def pre_star_def)
  apply auto
  apply (rule_tac h=h in exists_schedule)
  apply (metis sched_correct)
  apply (metis sched_correct)
  done

text {*
  The hedge-constrained predecessor set is always a subset of the 
  unconstrained predecessor set.
*}
lemma prehc_subset_pre_star: "prehc \<Delta> H C' \<subseteq> pre\<^sup>* \<Delta> C'"
  apply (unfold prehc_def pre_star_def)
  apply auto
  apply (rule_tac h=h in exists_schedule)
  apply (metis sched_correct)
  done


text {* We can use a hedge-constraint to express immediate predecessor sets. *}
definition Hpre :: "('P,'\<Gamma>,'L) ex_hedge set" where 
  "Hpre == { hl1@t#hl2 | hl1 t hl2 lab ts t'. 
                 final hl1 \<and> final hl2 \<and> final_t ts \<and> final_t t' \<and> 
                 (t=NNOSPAWN lab t' \<or> t=NSPAWN lab ts t') }"

lemma HpreI_nospawn:
  "\<lbrakk>final h1; final h2; final_t t'\<rbrakk> \<Longrightarrow> h1@NNOSPAWN lab t'#h2 \<in> Hpre"
  by (unfold Hpre_def) blast

lemma HpreI_spawn:
  "\<lbrakk>final h1; final h2; final_t ts; final_t t'\<rbrakk> \<Longrightarrow> h1@NSPAWN lab ts t'#h2 \<in> Hpre"
  by (unfold Hpre_def) blast

lemmas HpreI = HpreI_nospawn HpreI_spawn

lemma HpreE[cases set, consumes 1, case_names nospawn spawn]: 
  "\<lbrakk> h\<in>Hpre; 
     !!h1 lab t' h2. \<lbrakk>
       h=h1@NNOSPAWN lab t'#h2; final h1; final h2; final_t t'
     \<rbrakk> \<Longrightarrow> P; 
     !!h1 lab ts t' h2. \<lbrakk> 
       h=h1@NSPAWN lab ts t'#h2; 
       final h1; final h2; final_t ts; final_t t'
     \<rbrakk> \<Longrightarrow> P
   \<rbrakk> \<Longrightarrow> P"
  by (unfold Hpre_def) blast

text {*
  In order to show that @{const Hpre} is correct, we first show that it exactly 
  admits the schedules of length one.
*}
lemma Hpre_length1: "\<lbrakk>h\<in>Hpre; sched h ll\<rbrakk> \<Longrightarrow> length ll = 1" 
proof (erule HpreE)
  case (goal1 h1 lab t' h2) note C=this -- "nospawn"
  note [simp] = C(2-)
  from C(1) obtain l ll' where "ll=l#ll'" "sched (h1@t'#h2) ll'" 
    by (erule_tac sched.cases) (auto dest!: prop_matchD[where P=final_t])
  moreover have "final (h1@t'#h2)" by auto
  ultimately show ?case by auto
next
  case (goal2 h1 lab ts t' h2) note C=this -- "spawn"
  note [simp] = C(2-)
  from C(1) obtain l ll' where "ll=l#ll'" "sched (h1@ts#t'#h2) ll'" 
    by (erule_tac sched.cases) (auto dest!: prop_matchD[where P=final_t])
  moreover have "final (h1@ts#t'#h2)" by auto
  ultimately show ?case by auto
qed

lemma Hpre_length2: "\<lbrakk>sched h ll; length ll = 1\<rbrakk> \<Longrightarrow> h\<in>Hpre"
  by (erule sched.cases) (auto intro: HpreI)

theorem Hpre_length: "sched h ll \<Longrightarrow> h\<in>Hpre \<longleftrightarrow> length ll = 1" 
  using Hpre_length1 Hpre_length2 by blast

text {*
  It is then straightforward to show that @{term "prehc \<Delta> Hpre = pre \<Delta>"}
*}  
lemma Hpre_correct1: "c\<in>prehc \<Delta> Hpre C' \<Longrightarrow> c\<in>pre \<Delta> C'"
  apply (unfold prehc_def)
  apply auto
  apply (rule_tac h=h in exists_schedule)
  apply (simp only: Hpre_length)
  apply (drule (1) sched_correct2)
  apply (case_tac ll)
  apply simp
  apply simp
  apply (auto simp add: pre_def)
  done

lemma Hpre_correct2: "c\<in>pre \<Delta> C' \<Longrightarrow> c\<in>prehc \<Delta> Hpre C'"
  apply (unfold pre_def)
  apply auto
  apply (drule iffD2[OF trcl_single])
  apply (drule sched_correct1)
  apply auto
  apply (drule Hpre_length2)
  apply (auto simp add: prehc_def)
  done

theorem Hpre_correct: "prehc \<Delta> Hpre = pre \<Delta>" 
  using Hpre_correct1 Hpre_correct2 by (blast intro: ext)

end
