header "Special Lemmas"
theory SpecialLemmas
imports Main "common/Misc"
begin
text_raw {*\label{thy:SpecialLemmas}*}
  text {*
    This theory contains some special lemmas on standard structures, that are useful 
    in particular with DPNs. Some of these lemmas might be moved into more generic 
    libraries, i.e. common/Misc at some time.
    *}


lemma append_take_nth_drop_id: 
  "i < length c \<Longrightarrow> take i c @ (c ! i) # drop (Suc i) c = c"
  by (metis append_take_drop_id nth_drop')

text {*
  This lemma makes a case distinction of how a particular element of one list is 
  related to the structure of another list. It is particularly tailored to analyze 
  the relation of two consecutive DPN-steps: The case-names are the following:
  \begin{description}
    \item[C1] The second step was performed on some process before that of the 
              first step.
    \item[SPAWNED] The second step was performed on a process that was spawned 
                   by the first step.
    \item[SAME] The second step was performed on the same process as the first step.
    \item[C2] - The second step was performed on some process after that of 
                the first step.
  \end{description}
*}
lemma step_cons_cases[consumes 1, case_names C1 SPAWNED SAME C2]:
  assumes 
    A: "c1@cs@qs#c2 = c1'@qs'#c2'" and
    C1: "!!c21'. \<lbrakk> c1=c1'@qs'#c21'; c2'=c21'@cs@qs#c2 \<rbrakk> \<Longrightarrow> P" and
    C2: "!!cs1 c21'. \<lbrakk> c1'=c1@cs1; cs=cs1@qs'#c21'; c2'=c21'@qs#c2 \<rbrakk> \<Longrightarrow> P" and
    C3: "\<lbrakk> c1'=c1@cs; qs'=qs; c2'=c2 \<rbrakk> \<Longrightarrow> P" and
    C4: "!!c21. \<lbrakk>c1'=c1@cs@qs#c21; c2=c21@qs'#c2' \<rbrakk> \<Longrightarrow> P"
  shows P
proof -
  let ?c="c1@cs@qs#c2" and ?c'="c1'@qs'#c2'"
  from A have 
    B: "c1'=take (length c1') ?c" 
       "qs'=?c ! (length c1')" 
       "c2'=drop (Suc (length c1')) ?c" 
    by auto
  {
    assume C: "length c1' < length c1"
    with B have 
      C1'F: "c1' = take (length c1') c1" and 
      QS'F: "qs'=c1 ! (length c1')" and 
      C2'F: "c2'=drop (Suc (length c1')) c1 @ cs@qs#c2" 
      by auto (metis A C add_0_right nth_Cons' nth_append nth_append_length_plus)
    hence C1F: "c1=c1'@qs'#drop (Suc (length c1')) c1" 
      by (metis C append_take_drop_id nth_drop')
    from C1[OF C1F C2'F] have P .
  } moreover {
    assume C: "length c1' \<ge> length c1" "length c1' < length c1 + length cs"
    with B have 
      C1'F: "c1' = c1 @ take (length c1' - length c1) cs" and 
      QS'F: "qs'=cs ! (length c1' - length c1)" and 
      C2'F: "c2'=drop (Suc (length c1') - length c1) (cs) @ qs#c2" 
      by (auto simp add: nth_take[OF C(2), of ?c, simplified] nth_append)
    with C have 
      CSF: "cs = take (length c1' - length c1) cs @ 
                 qs' # 
                 drop (Suc (length c1') - length c1) (cs)"
      using append_take_drop_id[of "(length c1' - length c1)" cs] 
        nth_drop'[of "(length c1' - length c1)" cs] 
        Suc_diff_le[OF C(1)]
      by auto
    from C2[OF C1'F CSF C2'F] have P .
  } moreover {
    assume C: "length c1' = length c1 + length cs"
    with B have "c1'=c1@cs" "qs'=qs" "c2'=c2" by auto
    hence P by (rule C3)
  } moreover {
    assume C: "length c1' > length c1 + length cs"
    from A have C': "length c1' < length c1 + length cs + 1 + length c2"
    proof -
      from A have "length ?c = length ?c'" by simp
      thus ?thesis by auto
    qed
    with C have LL: "length c1' - length c1 - length cs - 1 < length c2" by auto
    from C C' have 
      LL': "Suc (length c1' - Suc (length c1 + length cs)) = 
              length c1' - (length c1 + length cs)" 
      by auto
    from B C C' have 
      C1'F: "c1' = c1@cs@qs#take (length c1' - length c1 - length cs - 1) c2" 
      and 
      QS'F: "qs' = c2 ! (length c1' - length c1 - length cs - 1)" and 
      C2'F: "c2' = drop (length c1' - length c1 - length cs) c2"
      by (auto simp add: take_Cons' nth_append nth_Cons' drop_Cons')
    hence 
      C2F: "c2 = take (length c1' - length c1 - length cs - 1) c2 @ qs' # c2'"
      using append_take_nth_drop_id[OF LL] LL'
      by auto
    from C4[OF C1'F C2F] have P .
  }
  moreover have 
    "(length c1' < length c1) \<or> 
     (length c1' \<ge> length c1 \<and> length c1' < length c1 + length cs) \<or> 
     (length c1' = length c1 + length cs) \<or> 
     (length c1' > length c1 + length cs)"
    by auto
  ultimately show P by blast
qed



text {*
  This lemma makes a case distinction whether an embedded step was made on the first or second part of a context.
  It is used to prove, e.g., split-lemmas for executions.
*}
lemma step_split_cases[consumes 1, case_names CA CB]:
  assumes A: "c1@qs#c2 = ca@cb"
  assumes CA: "!!c21. \<lbrakk> c2=c21@cb; ca=c1@qs#c21 \<rbrakk> \<Longrightarrow> P"
  assumes CB: "!!c12. \<lbrakk> c1=ca@c12; cb=c12@qs#c2 \<rbrakk> \<Longrightarrow> P"
  shows P
proof -
  from A have EQ: "ca=take (length ca) (c1@qs#c2)" 
                  "cb=drop (length ca) (c1@qs#c2)" 
    by auto
  {
    assume "length c1 < length ca"
    with EQ have 
      2: "ca=c1@qs#(take (length ca - length c1 - 1) c2)" and 
         "cb=drop (length ca - length c1 - 1) c2" 
      by (auto simp add: take_Cons' drop_Cons')
    hence 1: "c2=(take (length ca - length c1 - 1) c2) @ cb" by simp
    from CA[OF 1 2] have P .
  } moreover {
    assume "length c1 \<ge> length ca"
    with EQ have 
      CAS: "ca=take (length ca) c1" and 
      2:"cb=(drop (length ca) c1)@qs#c2" by auto
    have 1: "c1=ca@(drop (length ca) c1)" by (subst CAS) auto
    from CB[OF 1 2] have P .
  } ultimately show P by arith
qed


lemma comp_into_rtrancl: "R O R' \<subseteq> (R\<union>R')\<^sup>*"
  by (metis Un_subset_iff Un_upper1 rel_comp_mono rtrancl_idemp_self_comp 
            trancl_reflcl trancl_unfold)

lemmas comp_into_rtranclI = set_rev_mp[OF _ comp_into_rtrancl]


lemma un_rtrancl_as_comp1: 
  "(c,c')\<in>(A\<union>B)\<^sup>* \<Longrightarrow> (c,c')\<in>A\<^sup>* O ( B O A\<^sup>* )\<^sup>*" (is "_\<Longrightarrow>_\<in>?R")
proof (induct rule: rtrancl_induct, auto)
  fix ch1 ch2 ch3
  assume "(c,ch1)\<in>(B O A\<^sup>*)\<^sup>*"
  also assume "(ch1,ch2)\<in>A\<^sup>*" "(ch2,ch3)\<in>B" hence "(ch1,ch3)\<in>(B O A\<^sup>*)\<^sup>*" 
    by auto
  finally show "(c, ch3) \<in> (A\<^sup>* O (B O A\<^sup>*)\<^sup>*)" by auto
qed

lemma un_rtrancl_as_comp2: "(c,c')\<in>(A\<union>B)\<^sup>* \<Longrightarrow> (c,c')\<in>(A\<^sup>* O B)\<^sup>* O A\<^sup>*" (is "_\<Longrightarrow>_\<in>?R")
proof (induct rule: rtrancl_induct)
  case base thus ?case by auto
  case (step ch2 c') 
  then obtain ch where S: "(c,ch)\<in>A\<^sup>*" "(ch,ch2)\<in>(A\<^sup>* O B)\<^sup>*" by auto
  note `(ch2,c')\<in>A\<union>B` moreover
  {
    assume CASE: "(ch2,c')\<in>A"
    from S(2) have ?case 
    proof (cases rule: rtrancl.cases)
      case rtrancl_refl with S(1) have "(c,ch2)\<in>A\<^sup>*" by simp
      with CASE have "(c,c')\<in>A\<^sup>*" by auto
      thus ?thesis by auto
    next
      case (rtrancl_into_rtrancl _ ch1 _) 
      hence CC: "(ch,ch1)\<in>(A\<^sup>* O B)\<^sup>*" "(ch1,ch2)\<in>A\<^sup>* O B" by auto
      with CASE have "(ch1,c')\<in>A\<^sup>* O B" by auto
      hence "(ch1,c')\<in>(A\<^sup>* O B)\<^sup>*" by auto
      with CC(1) have "(ch,c')\<in>(A\<^sup>* O B)\<^sup>*" by auto
      with S(1) show ?thesis by auto
    qed
  } moreover {
    assume CASE: "(ch2,c')\<in>B"
    hence "(ch2,c')\<in>(A\<^sup>* O B)\<^sup>*" by auto
    with S have ?case by auto
  } ultimately show ?case by blast
qed

lemma un_rtrancl_as_comp3: "(c,c')\<in>(B\<union>A)\<^sup>* \<Longrightarrow> (c,c')\<in>(A\<^sup>* O B)\<^sup>* O A\<^sup>*" 
  using un_rtrancl_as_comp2 by (auto simp add: Un_ac)


lemma in_conv: "x\<in>A \<Longrightarrow> A\<noteq>{}"
  by (metis ex_in_conv)


lemma le_cases[consumes 1, case_names lt eq]: 
  "\<lbrakk>a\<le>(b::'x::order); a<b \<Longrightarrow> P; a=b \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P" 
  by (blast dest: order_neq_le_trans)



lemma singl_eq_comp[simp]: "[a]=c@b#d \<longleftrightarrow> a=b \<and> c=[] \<and> d=[]"
  by (metis append_eq_conv_conj append_is_Nil_conv butlast.simps(2) 
            butlast_append list.simps(1) list_tail_coinc self_append_conv2 take_eq_Nil)

lemma comp_eq_singl[simp]: "c@b#d=[a] \<longleftrightarrow> a=b \<and> c=[] \<and> d=[]"
  by (metis append_eq_conv_conj append_is_Nil_conv butlast.simps(2) 
            butlast_append list.simps(1) list_tail_coinc self_append_conv2 
            take_eq_Nil)


subsection "k-Step Closure"

fun krtcl :: "('a\<times>'a) set \<Rightarrow> nat \<Rightarrow> ('a\<times>'a) set" ("_\<^bsup>_\<^esup>" [1000] 1000) where
  "S\<^bsup>0\<^esup> = { (c,c) | c. True }" |
  "S\<^bsup>Suc k\<^esup> = S\<^bsup>k\<^esup> O S"
  
lemma rtrancl_to_krtcl: "\<lbrakk>(c,c')\<in>S\<^sup>*; !!k. (c,c')\<in>S\<^bsup>k\<^esup> \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P "
proof (induct arbitrary: P rule: converse_rtrancl_induct[case_names refl step])
  case refl have "(c',c')\<in>S\<^bsup>0\<^esup>" by simp thus ?case using refl by blast
next
  case (step c ch)
  from step.hyps(3) obtain k where "(ch,c')\<in>S\<^bsup>k\<^esup>" .
  with step.hyps(1) have "(c,c')\<in>S\<^bsup>Suc k\<^esup>" by auto
  with step.prems(1) show P .
qed
    
lemma krtcl_to_rtrancl: "S\<^bsup>k\<^esup> \<subseteq> S\<^sup>*"
  by (induct k) auto

lemmas krtcl_to_rtranclD = set_rev_mp[OF _ krtcl_to_rtrancl]

text {* Lists of unique elements *}
lemma distinct_list_length_bound: 
  "\<lbrakk> card (UNIV::'a::finite set) = k; distinct (l::'a list) \<rbrakk> \<Longrightarrow> length l \<le> k"
  by (metis card_mono distinct_card finite subset_UNIV)

text {* 
  These lemmas reason about graphs where all edges point to one single node.
  Such graphs are the graphs of singleton acquisition histories.
  *}
lemma trancl_m21_edges[simp]: "(X\<times>{m})\<^sup>+ = X\<times>{m}"
proof -
  { fix a b
    have "(a,b)\<in>(X\<times>{m})\<^sup>+ \<Longrightarrow> (a,b)\<in>X\<times>{m}" by (induct rule: trancl_induct) auto
  } thus ?thesis
    apply (intro equalityI subsetI)
    apply (case_tac x)
    apply (simp only:)
    by (rule r_into_trancl')
qed

lemma acyclic_m21_edges[simp]: "acyclic (X\<times>{m}) \<longleftrightarrow> m\<notin>X" 
  by (auto simp add: acyclic_def)

text {*
  This lemma reasons about the graph of an updated acquisition history. Any path in 
  this graph is either already contained in the original graph, or passes via an 
  inserted edge. Because all the inserted edges point to the same target node, in the
  second case, the path can be short-circuited to use exactly one inserted edge.
    *}
lemma trancl_multi_insert[cases set, case_names orig via]: 
  "\<lbrakk> (a,b)\<in>(r\<union>X\<times>{m})\<^sup>+; 
     (a,b)\<in>r\<^sup>+ \<Longrightarrow> P; 
     !!x. \<lbrakk> x\<in>X; (a,x)\<in>r\<^sup>*; (m,b)\<in>r\<^sup>* \<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
proof (induct arbitrary: P rule: trancl_induct)
  case (base b) thus ?case by auto
next
  case (step b c) show ?case proof (rule step.hyps(3))
    assume A: "(a,b)\<in>r\<^sup>+" 
    note step.hyps(2) 
    moreover {
      assume "(b,c)\<in>r" 
      with A have "(a,c)\<in>r\<^sup>+" by auto 
      with step.prems have P by blast
    } moreover {
      assume "b\<in>X" "c=m"
      with A have P by (rule_tac step.prems(2)) simp+
    } ultimately show P by auto
  next
    fix x
    assume A: "x \<in> X" "(a, x) \<in> r\<^sup>*" "(m, b) \<in> r\<^sup>*"
    note step.hyps(2) 
    moreover {
      assume "(b,c)\<in>r" 
      with A(3) have "(m,c)\<in>r\<^sup>*" by auto 
      with step.prems(2)[OF A(1,2)] have P by blast
    } moreover {
      assume "b\<in>X" "c=m"
      with A have P by (rule_tac step.prems(2)) simp+
    } ultimately show P by auto
  qed
qed


lemma trancl_multi_insert2[cases set, case_names orig via]: 
  "\<lbrakk>(a,b)\<in>(r\<union>{m}\<times>X)\<^sup>+; (a,b)\<in>r\<^sup>+ \<Longrightarrow> P; !!x. \<lbrakk> x\<in>X; (a,m)\<in>r\<^sup>*; (x,b)\<in>r\<^sup>* \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
proof -
  case goal1 from goal1(1) have "(b,a)\<in>((r\<union>{m}\<times>X)\<^sup>+)\<inverse>" by simp
  also have "((r\<union>{m}\<times>X)\<^sup>+)\<inverse> = (r\<inverse>\<union>X\<times>{m})\<^sup>+" by (simp add: converse_add_simps)
  finally have "(b, a) \<in> (r\<inverse> \<union> X \<times> {m})\<^sup>+" .
  thus ?case 
    by (auto elim!: trancl_multi_insert 
             intro: goal1(2,3) 
             simp add: trancl_converse rtrancl_converse
       )
qed


text {*
  Some proposition about ranges of restricted maps and addition of maps.
  Used in the acquisition history theory for a proposition that the union 
  of ranges of the acquisition history restricted to the root nodes of the 
  hedge is already the union of ranges of the unrestricted acquisition history.
  I.e. a proposition of the form @{term "\<Union>ran (e|`roots) = \<Union>ran e"}.

  In an induction step, we have to derive this property for a hedge @{term "t#h"} from
  the property for @{term t} and @{term h}. This is the point where this lemma is 
  needed.
*}
lemma union_ran_add_aux:
  assumes 
  A: "\<Union>ran (e1 |` r1) = u1" "\<Union>ran (e2 |` r2) = u2" 
     "r1 \<inter> dom e2 = {}" "dom e1 \<inter> r2 = {}"
  shows "\<Union>ran ((e1++e2) |` (r1 \<union> r2)) = u1 \<union> u2" 
  using A
  apply (auto simp add: restrict_map_def ran_def map_add_def 
              split: split_if_asm option.split)
  apply fastsimp+
  done

end
