header "Property Specifications"
theory Specification
imports DPN_c0 Semantics LockSem "common/SublistOrder"
begin
text_raw {*\label{thy:Specification}*}
  text {*
    We develop a formalism that allows a concise and readable notation for a class of 
    properties that are checkable via cascaded predecessor computations.

    A specification consists of a list of atoms, where each atom either restricts the 
    current configuration or describes some step.
    *}

  subsection "Specification Formulas"

  text {* The base element of a property is an atom, that describes a step or 
    restricts the current configuration *}
  datatype ('Q,'\<Gamma>,'L,'X) spec_atom = 
      -- "Restrict current configuration to be in a specified set"
      SPEC_RESTRICT "('Q,'\<Gamma>) conf set" |
      -- "Go forward one step, using a rule with labels from a specified set"
      SPEC_STEP "('L,'X) lockstep set" |
      -- "Go forward any number of steps, using rules with labels from a specified set"
      SPEC_STEPS "('L,'X) lockstep set"    

  text {* A property is a list of atoms *}
  types ('Q,'\<Gamma>,'L,'X) spec = "('Q,'\<Gamma>,'L,'X) spec_atom list"

  subsection "Semantics"
  text {* 
    The semantics of a property specification @{text "\<Phi>"} w.r.t. the current DPN is 
    modelled by a transition relation @{term "spec_tr \<Phi>"}, that contains all pairs 
    @{text "(c,c')"} of configurations, such that there is a path between @{text c} 
    and @{text c'} satisfying the property. 
    *}
  
  context LDPN
  begin
    fun spec_tr where
      "spec_tr [] = Id" |
      "spec_tr (SPEC_RESTRICT C # \<Phi>) = {(c,c') . (c,c')\<in>spec_tr \<Phi> \<and> fst c\<in>C}" |
      "spec_tr (SPEC_STEP L # \<Phi>) = 
        {(c,c') . \<exists>l\<in>L. \<exists>ch. (c,l,ch)\<in>ldpntr \<Delta> \<and> (ch,c')\<in>spec_tr \<Phi>}" |
      "spec_tr (SPEC_STEPS L # \<Phi>) = 
        {(c,c') . \<exists>ll\<in>lists L. \<exists>ch. (c,ll,ch)\<in>ldpntrc \<Delta> \<and> (ch,c')\<in>spec_tr \<Phi>}"
  end

  context LDPN_c0
  begin
    text {* In most cases, it suffices to check whether there is a path matching the 
      specification from the initial configuration. *}
    definition "model_check_ref \<Phi> == (c0,locks_\<mu> \<mu>0)\<in>Domain (spec_tr \<Phi>)"
  end

  subsection "Examples"
  text {*
    In this section, we present two short examples to justify the usefulness of our 
    property specifications.
    *}

  subsubsection "Conflict analysis"
  text {*
    Given two stack symbols @{text "u,v\<in>\<Gamma>"}, conflict analysis asks whether a 
    configuration @{text c} is reachable that has a conflict between @{text u} and 
    @{text v}.

    A configuration has a conflict between @{text u} and @{text v}, iff it contains a 
    process with top stack symbol @{text u} and another (different) process with top 
    stack symbol @{text v}.
    *}
  context LDPN_c0
  begin
    text {* @{text "atUV u v"} is the set of configurations that have a conflict 
      between @{text u} and @{text v}. *}
    definition "atUV_ordered u v == { c. \<exists>q r q' r'. [(q,u#r),(q',v#r')] \<le> c }"
    definition "atUV u v == (atUV_ordered u v) \<union> (atUV_ordered v u)"

    text {* The following property specification describes all executions reaching a 
      conflict: *}
    definition "conflict_spec u v == 
      [SPEC_STEPS UNIV, SPEC_RESTRICT (atUV u v)]"

    text {* The following definition is a direct definition of a conflict between 
      @{text u} and @{text v} being reachable from an initial configuration 
      @{text "[(qmain,[\<gamma>main])]"}: *}
    definition "has_conflict_ref u v == \<exists>(c,X)\<in>reachablels. c \<in> atUV u v"

    text {* The next lemma shows that the direct definition of a conflict matches the 
      property specification: *}
    lemma "has_conflict_ref u v \<longleftrightarrow> model_check_ref (conflict_spec u v)"
      by (unfold model_check_ref_def conflict_spec_def has_conflict_ref_def 
                 Domain_def reachablels_def) 
         auto

  end

  subsubsection "Bitvector analysis"
  text {*
    Given a set of generator labels @{text "G::'L set"}, a set of killer labels 
    @{text "K::'L set"} and a stack symbol @{text "u::'\<Gamma>"}, bitvector analysis asks 
    whether there is a path to a configuration that has process being at @{text u}, 
    such that the path executes a generator rule, and after that no killer rule
    is executed.
    *}
  context LDPN_c0
  begin
    text {* For a stack symbol, @{text "u\<in>\<Gamma>"}, the set @{text "atU u"} is the set of 
      all configurations that have a process with @{text u} at the top of the stack. *}
    definition "atU u == { c . \<exists>q r. (q,u#r)\<in>set c }"

    text {* The following property specification describes all paths that lead to 
      @{text u} and have the bit set: *}
    definition "bitvector_fwd_spec G K u == 
      [ SPEC_STEPS UNIV, 
        SPEC_STEP G, 
        SPEC_STEPS (UNIV-K), 
        SPEC_RESTRICT (atU u)
      ]"

    text {* The following is the direct definition of bitvector analysis: *}
    definition "bitvector_fwd_ref G K u == 
      \<exists>c1 X1 lg c2 X2 ll c3 X3 q r. 
        (c1,X1)\<in>reachablels \<and> 
        ((c1,X1),lg,(c2,X2))\<in>ldpntr \<Delta> \<and> 
        lg\<in>G \<and> 
        ((c2,X2),ll,(c3,X3))\<in>ldpntrc \<Delta> \<and> 
        ll\<in>lists (UNIV-K) \<and> 
        (q,u#r)\<in>set c3"

    text {* This lemma shows that the direct definition matches the property 
      specification: *}
    lemma "bitvector_fwd_ref G K u \<longleftrightarrow> 
             model_check_ref (bitvector_fwd_spec G K u)"
      by (unfold model_check_ref_def bitvector_fwd_spec_def 
                 bitvector_fwd_ref_def Domain_def atU_def reachablels_def) 
        fastsimp

  end
end
