header "Well-Nestedness of Locks"
theory WellNested
imports DPN Semantics LockSem
begin
text_raw {*\label{thy:WellNested}*}

text {*
  Well-nestedness of locks is the property that no locks are re-acquired by the 
  same process and a released locks is always the last one that was acquired 
  and not yet released by the releasing process. 
  Usually, these two properties are called non-reentrance and well-nestedness. 

  In this theory, we formulate a sufficient condition for well-nestedness, that 
  regards every possible lock-insensitive run of the DPN from some initial 
  configuration. We then define an equivalent condition on execution hedges.

  Note that our condition may rule out DPNs where some non-well-nested runs are 
  blocked by deadlocks or other lock-induced effects. However, important classes of 
  programs, in particular programs that use locks in a block-structured way (like 
  synchronized-blocks in Java), always satisfy our condition.

  Further work required at this point is to formalize a program analysis or some 
  sufficient conditions (like block-structured lock-acquisition [monitors]) for 
  well-nestedness. We would then be able to prove some non-trivial DPNs to have
  well-nested configurations, thus giving a stronger indication that the 
  well-nestedness assumption is correct. In the current state, we have no formal 
  proof that the well-nestedness assumption is correct, i.e. an uncorrect 
  well-nestedness assumption, e.g. a too strict one, would affect the scope of 
  all our proofs that use this assumption. In the worst case, there would be no
  well-nested DPNs at all (or only trivial ones).
*}


subsection "Well-Nestedness Condition on Paths"

text {*
  We first define the set of all paths that may occur from a process. We collect 
  local paths and environment paths.
    \begin{description}
      \item[@{text "ppairs (q,w) False l"}] means that there is a local 
        path @{text l} from process @{text "(q,w)"}.
      \item[@{text "ppairs (q,w) True l"}] means that we can reach a spawn step 
        from process @{text "(q,w)"} that spawns a process having path "@{text l}".
    \end{description}
    *}
inductive ppairs 
  :: "('P,'\<Gamma>,'L,'X) ldpn \<Rightarrow> ('P,'\<Gamma>) pconf \<Rightarrow> bool \<Rightarrow> ('L,'X) lockstep list \<Rightarrow> bool" 
  for \<Delta> where
  ppairs_empty: "ppairs \<Delta> (q,w) False []" |
  ppairs_prepend1: 
  "\<lbrakk>(q,\<gamma> \<hookrightarrow>\<^bsub>a\<^esub> q',w) \<in> \<Delta>; ppairs \<Delta> (q',w@r) False l \<rbrakk> \<Longrightarrow> 
    ppairs \<Delta> (q,\<gamma>#r) False (a#l)" |
  ppairs_mvenv1: 
  "\<lbrakk>(q,\<gamma> \<hookrightarrow>\<^bsub>a\<^esub> q',w) \<in> \<Delta>; ppairs \<Delta> (q',w@r) True l \<rbrakk> \<Longrightarrow> 
    ppairs \<Delta> (q,\<gamma>#r) True l" |
  ppairs_prepend2: 
  "\<lbrakk>(q,\<gamma> \<hookrightarrow>\<^bsub>a\<^esub> qs,ws \<sharp> q',w) \<in> \<Delta>; ppairs \<Delta> (q',w@r) False l \<rbrakk> \<Longrightarrow> 
    ppairs \<Delta> (q,\<gamma>#r) False (a#l)" |
  ppairs_mvenv2: "\<lbrakk>(q,\<gamma> \<hookrightarrow>\<^bsub>a\<^esub> qs,ws \<sharp> q',w) \<in> \<Delta>; ppairs \<Delta> (q',w@r) True l \<rbrakk> \<Longrightarrow> 
    ppairs \<Delta> (q,\<gamma>#r) True l" |
  ppairs_genenv: "\<lbrakk>(q,\<gamma> \<hookrightarrow>\<^sub>a qs,ws \<sharp> q',w) \<in> \<Delta>; ppairs \<Delta> (qs,ws) x l\<rbrakk> \<Longrightarrow> 
    ppairs \<Delta> (q,\<gamma>#r) True l"

text {*
  This function checks whether a path is well-nested by using a lock stack.
  *}
fun wn_p :: "('L,'X) lockstep list \<Rightarrow> 'X list \<Rightarrow> bool" where
  "wn_p [] \<mu> = distinct \<mu>" |
  "wn_p (LAcq x#l) \<mu> \<longleftrightarrow> wn_p l (x#\<mu>)" |
  "wn_p (LRel x#l) \<mu> \<longleftrightarrow> (\<exists>\<mu>'. \<mu>=x#\<mu>' \<and> x\<notin>set \<mu>' \<and> wn_p l \<mu>')" |
  "wn_p (_#l) \<mu> \<longleftrightarrow> wn_p l \<mu>"

text {*
  A process @{text \<pi>} is defined to be well-nested w.r.t. some initial lock 
  stack @{text \<mu>} if all reachable path -- local paths and environment paths -- are 
  well-nested.
  *}
definition "wn_\<pi> \<Delta> \<pi> \<mu> == 
  case \<pi> of (p,w) \<Rightarrow> 
    \<forall>l. (ppairs \<Delta> (p,w) False l \<longrightarrow> wn_p l \<mu>) \<and> 
        (ppairs \<Delta> (p,w) True l \<longrightarrow> wn_p l [])"

text {* Introduction and elimination rules for @{const wn_\<pi>} *}
lemma wn_\<pi>I:
  "\<lbrakk> 
    !!l. ppairs \<Delta> (q,w) False l \<Longrightarrow> wn_p l \<mu>; 
    !!l. ppairs \<Delta> (q,w) True l \<Longrightarrow> wn_p l [] 
  \<rbrakk> \<Longrightarrow> wn_\<pi> \<Delta> (q,w) \<mu>"
  by (unfold wn_\<pi>_def) auto

lemma wn_\<pi>E: 
  "\<lbrakk>wn_\<pi> \<Delta> (q,w) \<mu>; 
     \<lbrakk>
       !!l. ppairs \<Delta> (q,w) False l \<Longrightarrow> wn_p l \<mu>; 
       !!l. ppairs \<Delta> (q,w) True l \<Longrightarrow> wn_p l [] 
     \<rbrakk> \<Longrightarrow> P
  \<rbrakk> \<Longrightarrow> P"
  by (unfold wn_\<pi>_def) auto
  

text {*
  We have set up the definitions such that well-nestedness w.r.t a lock stack implies 
  distinctness of this lock stack.
  *}
lemma wn_p_distinct: "wn_p l \<mu> \<Longrightarrow> distinct \<mu>" 
  by (induct rule: wn_p.induct) auto

lemma wn_\<pi>_distinct: "wn_\<pi> \<Delta> \<pi> \<mu> \<Longrightarrow> distinct \<mu>" 
  using ppairs.intros(1)
  apply (unfold wn_\<pi>_def)
  apply (simp split: prod.split_asm)
  apply (rule wn_p_distinct)
  apply (fast)
  done


text {* Well-nestedness is preserved by steps: *}
lemma wn_\<pi>_none: 
  "\<lbrakk> (q,\<gamma> \<hookrightarrow>\<^bsub>(LNone l)\<^esub> q',w)\<in>\<Delta>; wn_\<pi> \<Delta> (q,\<gamma>#r) \<mu> \<rbrakk> \<Longrightarrow> wn_\<pi> \<Delta> (q',w@r) \<mu>"
  by (unfold wn_\<pi>_def) (auto intro: ppairs.intros)
lemma (in LDPN) wn_\<pi>_spawn1: 
  "\<lbrakk> (q,\<gamma> \<hookrightarrow>\<^sub>a qs,ws \<sharp> q',w)\<in>\<Delta>; wn_\<pi> \<Delta> (q,\<gamma>#r) \<mu> \<rbrakk> \<Longrightarrow> wn_\<pi> \<Delta> (q',w@r) \<mu>"
  by (cases a, unfold wn_\<pi>_def) (auto intro: ppairs.intros)
lemma wn_\<pi>_spawn2: 
  "\<lbrakk> (q,\<gamma> \<hookrightarrow>\<^sub>a qs,ws \<sharp> q',w)\<in>\<Delta>; wn_\<pi> \<Delta> (q,\<gamma>#r) \<mu> \<rbrakk> \<Longrightarrow> wn_\<pi> \<Delta> (qs,ws) []"
  by (cases a, unfold wn_\<pi>_def) (auto intro: ppairs.intros)
lemma wn_\<pi>_acq: 
  "\<lbrakk> (q,\<gamma> \<hookrightarrow>\<^bsub>LAcq x\<^esub> q',w)\<in>\<Delta>; wn_\<pi> \<Delta> (q,\<gamma>#r) \<mu> \<rbrakk> \<Longrightarrow> wn_\<pi> \<Delta> (q',w@r) (x#\<mu>)"
  by (unfold wn_\<pi>_def) (auto intro: ppairs.intros)
lemma wn_\<pi>_rel: 
  assumes A: "(q,\<gamma> \<hookrightarrow>\<^bsub>LRel x\<^esub> q',w)\<in>\<Delta>" "wn_\<pi> \<Delta> (q,\<gamma>#r) \<mu>" and 
          C: "!!\<mu>'. \<lbrakk>\<mu>=x#\<mu>'; x\<notin>set \<mu>'; wn_\<pi> \<Delta> (q',w@r) \<mu>' \<rbrakk> \<Longrightarrow> P" 
  shows P
proof -
  from wn_\<pi>E[OF A(2)] have X: "!!l. ppairs \<Delta> (q, \<gamma> # r) False l \<Longrightarrow> wn_p l \<mu>" 
    by blast
  from X[OF ppairs_prepend1[OF A(1) ppairs_empty],simplified] obtain \<mu>' where
    [simp]: "\<mu>=x#\<mu>'" "x\<notin>set \<mu>'" 
    by blast
  moreover from A have "wn_\<pi> \<Delta> (q',w@r) \<mu>'" 
    by (unfold wn_\<pi>_def) (auto intro: ppairs.intros)
  ultimately show P by (rule C)
qed

lemma (in LDPN) wn_\<pi>_preserve:
  "\<lbrakk> (q,\<gamma> \<hookrightarrow>\<^sub>l q',w)\<in>\<Delta>; wn_\<pi> \<Delta> (q,\<gamma>#r) xs; 
     !!xs'. wn_\<pi> \<Delta> (q',w@r) xs' \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"

  "\<lbrakk> (q,\<gamma> \<hookrightarrow>\<^sub>l qs,ws \<sharp> q',w)\<in>\<Delta>; wn_\<pi> \<Delta> (q,\<gamma>#r) xs; 
     !!xs'. \<lbrakk> wn_\<pi> \<Delta> (q',w@r) xs'; wn_\<pi> \<Delta> (qs,ws) [] \<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
  apply (cases l)
  apply (auto dest!: wn_\<pi>_none wn_\<pi>_acq elim!: wn_\<pi>_rel) [3]
  apply (frule (1) wn_\<pi>_spawn1)
  apply (auto dest!: wn_\<pi>_spawn2)
done

subsection "Well-Nestedness of Configurations"

text {* The locks of a list of lock stacks *}
abbreviation locks_\<mu> :: "'X list list \<Rightarrow> 'X set" where  
  "locks_\<mu> \<mu> == list_collect_set set \<mu>"


text {*
  A configuration @{text "c=\<pi>\<^sub>1\<dots>\<pi>\<^sub>n"} is well-nested w.r.t. a list 
  @{text "\<mu>=s\<^sub>1\<dots>s\<^sub>n"} of lock stacks (@{text "wn_h h \<mu>"}), iff  
  all @{text "\<pi>\<^sub>i"} are well-nested w.r.t. stack @{text "s\<^sub>i"} and @{text \<mu>} 
  is consistent, i.e. contains no duplicate locks.
*}
fun wn_c where
  "wn_c \<Delta> [] [] \<longleftrightarrow> True" |
  "wn_c \<Delta> (\<pi>#c) (xs#\<mu>) \<longleftrightarrow> 
     wn_c \<Delta> c \<mu> \<and> set xs \<inter> locks_\<mu> \<mu> = {} \<and> wn_\<pi> \<Delta> \<pi> xs" |
  "wn_c \<Delta> _ _ \<longleftrightarrow> False"

subsubsection {* Auxilliary Lemmas about @{const wn_c} *}

lemma wn_c_simps[simp]: 
  "wn_c \<Delta> c [] \<longleftrightarrow> c=[]"
  "wn_c \<Delta> [] \<mu> \<longleftrightarrow> \<mu>=[]"
  apply (induct c)
  apply auto
  apply (induct \<mu>)
  apply auto
  done

lemma wn_c_length: "wn_c \<Delta> c \<mu> \<Longrightarrow> length c = length \<mu>"
  by (induct \<Delta> c \<mu> rule: wn_c.induct) auto

lemma wn_c_prepend_c: 
  "\<lbrakk> wn_c \<Delta> (\<pi>#c) \<mu>; 
     !!xs \<mu>'. \<lbrakk> \<mu>=xs#\<mu>'; wn_c \<Delta> c \<mu>'; 
                set xs \<inter> locks_\<mu> \<mu>' = {}; wn_\<pi> \<Delta> \<pi> xs 
              \<rbrakk> \<Longrightarrow> P
   \<rbrakk> \<Longrightarrow> P"
  by (induct \<mu> arbitrary: \<pi> c) fastsimp+

lemma wn_c_prepend_\<mu>: 
  "\<lbrakk> wn_c \<Delta> c (xs#\<mu>); 
     !!\<pi> c'. \<lbrakk> c=\<pi>#c'; wn_c \<Delta> c' \<mu>; 
               set xs \<inter> locks_\<mu> \<mu> = {}; wn_\<pi> \<Delta> \<pi> xs 
             \<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
  by (induct c arbitrary: \<mu>) auto

lemma wn_c_append_c_helper: 
  assumes 
    A: "wn_c \<Delta> c \<mu>" "c1@c2=c" and 
    C: "!!\<mu>1 \<mu>2. \<lbrakk> \<mu>=\<mu>1@\<mu>2 \<and> wn_c \<Delta> c1 \<mu>1 \<and> wn_c \<Delta> c2 \<mu>2 \<and> 
                   locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}
                 \<rbrakk> \<Longrightarrow> P" 
  shows P
  using A C
  apply (induct \<Delta> c \<mu> arbitrary: c1 c2 P rule: wn_c.induct)
  apply auto
  apply fastsimp
  apply (case_tac c1)
  apply fastsimp
  apply auto
proof -
  case goal1
  show P 
    apply (rule goal1(1))
    apply simp
    apply (rule_tac ?\<mu>1.0 = "xs#\<mu>1" and ?\<mu>2.0 = "\<mu>2" in goal1(2))
    apply (insert goal1(3-))
    apply auto
    done
qed

lemma wn_c_append_c: 
  "\<lbrakk>wn_c \<Delta> (c1@c2) \<mu>; 
     !!\<mu>1 \<mu>2. \<lbrakk> \<mu>=\<mu>1@\<mu>2 \<and> wn_c \<Delta> c1 \<mu>1 \<and> wn_c \<Delta> c2 \<mu>2 \<and> 
                locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}\<rbrakk> \<Longrightarrow> P
   \<rbrakk> \<Longrightarrow> P" 
  using wn_c_append_c_helper
  by blast

lemma wn_c_append_\<mu>_helper: 
  assumes 
    A: "wn_c \<Delta> c \<mu>" "\<mu>1@\<mu>2=\<mu>" and 
    C: "!!c1 c2. \<lbrakk> c=c1@c2 \<and> wn_c \<Delta> c1 \<mu>1 \<and> wn_c \<Delta> c2 \<mu>2 \<and> 
                   locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}\<rbrakk> \<Longrightarrow> P" 
  shows P
  using A C
  apply (induct \<Delta> c \<mu> arbitrary: \<mu>1 \<mu>2 P rule: wn_c.induct)
  apply auto
  apply (case_tac \<mu>1)
  apply fastsimp
  apply auto
proof -
  case goal1
  show P 
    apply (rule goal1(1))
    apply simp
    apply (rule_tac ?c1.0 = "(a,b)#c1" and ?c2.0 = "c2" in goal1(2))
    apply (insert goal1(3-))
    apply auto
    done
qed

lemma wn_c_append_\<mu>: 
  "\<lbrakk>wn_c \<Delta> c (\<mu>1@\<mu>2); 
    !!c1 c2. \<lbrakk> c=c1@c2 \<and> wn_c \<Delta> c1 \<mu>1 \<and> wn_c \<Delta> c2 \<mu>2 \<and> 
               locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}\<rbrakk> \<Longrightarrow> P
  \<rbrakk> \<Longrightarrow> P" 
  using wn_c_append_\<mu>_helper 
  by blast

lemma wn_c_appendI: 
  "\<lbrakk>wn_c \<Delta> c1 \<mu>1; wn_c \<Delta> c2 \<mu>2; locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}\<rbrakk> \<Longrightarrow> 
    wn_c \<Delta> (c1@c2) (\<mu>1@\<mu>2)"
  by (induct \<Delta> c1 \<mu>1 arbitrary: c2 \<mu>2 rule: wn_c.induct) auto

lemma wn_c_prependI: 
  "\<lbrakk>wn_\<pi> \<Delta> \<pi> xs; wn_c \<Delta> c \<mu>; set xs \<inter> locks_\<mu> \<mu> = {}\<rbrakk> \<Longrightarrow> wn_c \<Delta> (\<pi>#c) (xs#\<mu>)"
  by auto

lemma wn_c_singlecE: "\<lbrakk>wn_c \<Delta> [\<pi>] \<mu>; !!xs. \<lbrakk>\<mu>=[xs]; wn_\<pi> \<Delta> \<pi> xs\<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
  by (cases \<mu>) auto

lemma wn_c_split_aux: 
  assumes 
    WN: "wn_c \<Delta> c \<mu>" and 
    HFMT[simp]: "c=c1@\<pi>#c2" and
    C: "!!\<mu>1 xs \<mu>2. \<lbrakk> \<mu>=\<mu>1@xs#\<mu>2; wn_\<pi> \<Delta> \<pi> xs; wn_c \<Delta> c1 \<mu>1; wn_c \<Delta> c2 \<mu>2; 
                      locks_\<mu> \<mu>1 \<inter> set xs = {}; locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}; 
                      set xs \<inter> locks_\<mu> \<mu>2 = {} 
                    \<rbrakk> \<Longrightarrow> P"
  shows "P" 
  using WN[simplified]
  apply (elim wn_c_append_c wn_c_prepend_c conjE)
  apply (rule C)
  apply (auto)
  done

text {*
  Well-nestedness of configurations is preserved by lock-sensitive steps.
*}
lemma (in LDPN) wnc_preserve_singlestep: 
  assumes 
    A: "((c,locks_\<mu> \<mu>),l,(c',X'))\<in>ldpntr \<Delta>" "wn_c \<Delta> c \<mu>" and 
    C: "!!\<mu>'. \<lbrakk>X'=locks_\<mu> \<mu>'; wn_c \<Delta> c' \<mu>'\<rbrakk> \<Longrightarrow> P" 
  shows "P"
proof -
  from A have TR: "(c,l,c')\<in>dpntr \<Delta>" and LV: "lock_valid (locks_\<mu> \<mu>) l X'" 
    by (auto simp add: ldpntr_def)
  from TR show ?thesis proof (cases rule: dpntr.cases)
    case (dpntr_no_spawn p \<gamma> _ p' w c1 r c2) 
    hence 
      FMT[simp]: "c = c1 @ (p, \<gamma> # r) # c2" "c' = c1 @ (p', w @ r) # c2" and 
      R: "(p,\<gamma> \<hookrightarrow>\<^sub>l p',w) \<in> \<Delta>" 
      by auto
    from wn_c_split_aux[OF A(2) FMT(1)] obtain \<mu>1 xs \<mu>2 where 
      [simp]: "\<mu> = \<mu>1 @ xs # \<mu>2" and 
      WNS: "wn_\<pi> \<Delta> (p, \<gamma> # r) xs" "wn_c \<Delta> c1 \<mu>1" "wn_c \<Delta> c2 \<mu>2" and
      DISJ: "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" 
            "set xs \<inter> locks_\<mu> \<mu>2 = {}" 
      .
    obtain xs' where 
      "wn_\<pi> \<Delta> (p',w@r) xs'" "X'=(locks_\<mu> (\<mu>1@xs'#\<mu>2))" 
      "locks_\<mu> \<mu>1 \<inter> set xs' = {}" "set xs' \<inter> locks_\<mu> \<mu>2 = {}" 
    proof (cases l)
      case LNone[simp] 
      from that[OF wn_\<pi>_none[OF R[simplified] WNS(1)]] DISJ LV show ?thesis
        by simp
    next
      case (LAcq x)[simp] 
      from that[OF wn_\<pi>_acq[OF R[simplified] WNS(1)]] LV DISJ show ?thesis
        by simp
    next
      case (LRel x)[simp] 
      from wn_\<pi>_rel[OF R[simplified] WNS(1)] obtain xs' where 
        [simp]: "xs=x#xs'" and 
          1: "x\<notin>set xs'" and 
          2: "wn_\<pi> \<Delta> (p',w@r) xs'" 
        .
      from 1 LV DISJ show ?thesis by (rule_tac that[OF 2]) auto
    qed
    with WNS(2,3) DISJ(2) show P 
      by (rule_tac \<mu>'="\<mu>1@xs'#\<mu>2" in C) (auto intro!: wn_c_appendI wn_c_prependI)
  next
    case (dpntr_spawn p \<gamma> _ ps ws p' w c1 r c2) 
    hence 
      FMT[simp]: "c = c1 @ (p, \<gamma> # r) # c2" 
                 "c' = c1 @ (ps, ws) # (p', w @ r) # c2" and 
      R: "(p,\<gamma> \<hookrightarrow>\<^sub>l ps,ws \<sharp> p',w) \<in> \<Delta>" 
      by auto
    from R obtain ll where [simp]: "l=LNone ll" by (cases l) auto
    from wn_c_split_aux[OF A(2) FMT(1)] obtain \<mu>1 xs \<mu>2 where 
      [simp]: "\<mu> = \<mu>1 @ xs # \<mu>2" and 
      WNS: "wn_\<pi> \<Delta> (p, \<gamma> # r) xs" "wn_c \<Delta> c1 \<mu>1" "wn_c \<Delta> c2 \<mu>2" and
      DISJ: "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" 
            "set xs \<inter> locks_\<mu> \<mu>2 = {}" 
      .
    from wn_\<pi>_spawn1[OF R WNS(1)] wn_\<pi>_spawn2[OF R WNS(1)] 
         WNS(2,3) DISJ 
    have "wn_c \<Delta> c' (\<mu>1@[]#xs#\<mu>2)" 
      by (auto intro!: wn_c_appendI wn_c_prependI)
    thus ?thesis using LV by (rule_tac \<mu>'="\<mu>1@[]#xs#\<mu>2" in C) auto
  qed
qed
    
lemma (in LDPN) wnc_preserve: 
  assumes A: "((c,locks_\<mu> \<mu>),ll,(c',X'))\<in>ldpntrc \<Delta>" "wn_c \<Delta> c \<mu>" and 
          C: "!!\<mu>'. \<lbrakk>X'=locks_\<mu> \<mu>'; wn_c \<Delta> c' \<mu>'\<rbrakk> \<Longrightarrow> P" 
  shows "P"
proof -
  {
    fix c X \<mu> ll c' X' P
    assume A: "((c,X),ll,(c',X'))\<in>ldpntrc \<Delta>" "wn_c \<Delta> c \<mu>" "X=locks_\<mu> \<mu>" and 
           C: "!!\<mu>'. \<lbrakk>X'=locks_\<mu> \<mu>'; wn_c \<Delta> c' \<mu>'\<rbrakk> \<Longrightarrow> P" 
    hence P 
    proof (induct arbitrary: \<mu> P rule: trcl_pair_induct)
      case empty thus ?case by auto
    next
      case (cons c x l ch Xh ll c' X' \<mu> P) note [simp]=`x=locks_\<mu> \<mu>`
      from wnc_preserve_singlestep[OF cons.hyps(1)[simplified] cons.prems(1)] 
      obtain \<mu>' where P: "wn_c \<Delta> ch \<mu>'" "Xh=locks_\<mu> \<mu>'" .
      from cons.hyps(3)[OF P] cons.prems(3) show ?case by blast
    qed
  } with A C show ?thesis by blast
qed


subsection {* Well-Nestedness Condition on Trees*}

text {*
  Now we define well-nestedness on scheduling trees. Note that scheduling trees that 
  contain spawn steps with locks interaction are not well-nested.

  We define two equivalent formulations of well-nestedness of a tree:
  *}

fun wn_t' :: "('P,'\<Gamma>,'L,'X) lex_tree \<Rightarrow> 'X list \<Rightarrow> bool" where
  "wn_t' (NLEAF \<pi>) \<mu> \<longleftrightarrow> distinct \<mu>" |
  "wn_t' (NNOSPAWN (LNone l) t) \<mu> \<longleftrightarrow> wn_t' t  \<mu>" |
  "wn_t' (NSPAWN (LNone l) ts t) \<mu> \<longleftrightarrow> wn_t' t \<mu> \<and> wn_t' ts []" |
  "wn_t' (NNOSPAWN (LAcq x) t) \<mu> \<longleftrightarrow> wn_t' t (x#\<mu>) \<and> x\<notin>set \<mu>" |
  "wn_t' (NNOSPAWN (LRel x) t) \<mu> \<longleftrightarrow> 
           (\<exists>\<mu>'. \<mu>=x#\<mu>' \<and> wn_t' t \<mu>' \<and> x\<notin>set \<mu>')" |
  "wn_t' _ _ \<longleftrightarrow> False"

inductive wn_t :: "('P,'\<Gamma>,'L,'X) lex_tree \<Rightarrow> 'X list \<Rightarrow> bool" where
  "distinct \<mu> \<Longrightarrow> wn_t (NLEAF \<pi>) \<mu>" |
  "wn_t t \<mu> \<Longrightarrow> wn_t (NNOSPAWN (LNone l) t) \<mu>" |
  "\<lbrakk>wn_t t \<mu>; wn_t ts [] \<rbrakk> \<Longrightarrow> wn_t (NSPAWN (LNone l) ts t) \<mu>" |
  "\<lbrakk>wn_t t (x#\<mu>); x\<notin>set \<mu>\<rbrakk> \<Longrightarrow> wn_t (NNOSPAWN (LAcq x) t) \<mu>" |
  "\<lbrakk>wn_t t \<mu>; x\<notin>set \<mu>\<rbrakk> \<Longrightarrow> wn_t (NNOSPAWN (LRel x) t) (x#\<mu>)"

inductive lock_valid_xs where
  "distinct xs \<Longrightarrow> lock_valid_xs (LNone l) xs xs" |
  "\<lbrakk>distinct xs; x\<notin>set xs\<rbrakk> \<Longrightarrow> lock_valid_xs (LRel x) (x#xs) xs" |
  "\<lbrakk>distinct xs; x\<notin>set xs\<rbrakk> \<Longrightarrow> lock_valid_xs (LAcq x) xs (x#xs)"


text {* The two formulations of well-nestedness of trees are, indeed, equivalent: *}
lemma wnt_eq_wnt': "wn_t t \<mu> = wn_t' t \<mu>"
  apply safe
  apply (induct rule: wn_t.induct)
  apply auto
  apply (induct rule: wn_t'.induct)
  apply (auto intro: wn_t.intros)
  done

text {* Well-nestedness of trees also implies distinctness of the lock stacks *}
lemma wnt_distinct: "wn_t t \<mu> \<Longrightarrow> distinct \<mu>"
  by (induct rule: wn_t.induct) auto
lemma wnt_distinct': "wn_t' t ms \<Longrightarrow> distinct ms" 
  using wnt_distinct wnt_eq_wnt' by auto

lemma all_t_wnt_distinct: "\<forall>t c'. tsem \<Delta> (q,w) t c' \<longrightarrow> wn_t t \<mu> \<Longrightarrow> distinct \<mu>"
  by (auto intro: wn_t.intros wnt_distinct)

subsection "Well-Nestedness of Hedges"
text {*
  The well-nestedness property of a hedge expresses that each tree is well-nested, 
  and the allocated locks of the trees are consistent.
  *}
  
text {* Consistency of a list of lock stacks. @{text "\<mu>=s\<^sub>1\<dots>s\<^sub>n"} is consistent, 
  iff all @{text "s\<^sub>i"} are distinct and 
  @{text "\<forall>i j. i\<noteq>j \<longrightarrow> set s\<^sub>i \<inter> set s\<^sub>j = {}"}. 
  *}
fun cons_\<mu> :: "'X list list \<Rightarrow> bool" where
  "cons_\<mu> [] \<longleftrightarrow> True" |
  "cons_\<mu> (xs#\<mu>) \<longleftrightarrow> cons_\<mu> \<mu> \<and> distinct xs \<and> set xs \<inter> locks_\<mu> \<mu> = {}"

text {*
  A hedge @{text "h=t\<^sub>1\<dots>t\<^sub>n"} is well-nested w.r.t. a list @{text "\<mu>=s\<^sub>1\<dots>s\<^sub>n"} of lock 
  stacks (@{text "wn_h h \<mu>"}), iff all @{text "t\<^sub>i"} are well-nested w.r.t. stack 
  @{text "s\<^sub>i"} and @{text \<mu>} is consistent.
*}
fun wn_h where
  "wn_h [] [] \<longleftrightarrow> True" |
  "wn_h (t#h) (xs#\<mu>) \<longleftrightarrow> wn_h h \<mu> \<and> set xs \<inter> locks_\<mu> \<mu> = {} \<and> wn_t' t xs" |
  "wn_h _ _ \<longleftrightarrow> False"

lemma cons_\<mu>_append[simp]: 
  "cons_\<mu> (\<mu>1@\<mu>2) \<longleftrightarrow> cons_\<mu> \<mu>1 \<and> cons_\<mu> \<mu>2 \<and> locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}"
  by (induct \<mu>1 arbitrary: \<mu>2) auto

subsubsection {*Auxilliary Lemmas about @{const wn_h}*}

lemma wn_h_simps[simp]: 
  "wn_h h [] \<longleftrightarrow> h=[]"
  "wn_h [] \<mu> \<longleftrightarrow> \<mu>=[]"
  apply (induct h)
  apply auto
  apply (induct \<mu>)
  apply auto
  done

lemma wn_h_length: "wn_h h \<mu> \<Longrightarrow> length h = length \<mu>"
  by (induct h \<mu> rule: wn_h.induct) auto

lemma wn_h_prepend_h: 
  "\<lbrakk> wn_h (t#h) \<mu>; 
    !!xs \<mu>'. \<lbrakk> \<mu>=xs#\<mu>'; wn_h h \<mu>'; set xs \<inter> locks_\<mu> \<mu>' = {}; wn_t' t xs \<rbrakk> \<Longrightarrow> P 
  \<rbrakk> \<Longrightarrow> P"
  by (induct \<mu> arbitrary: t h) auto

lemma wn_h_prepend_\<mu>: 
  "\<lbrakk> wn_h h (xs#\<mu>); 
    !!t h'. \<lbrakk> h=t#h'; wn_h h' \<mu>; set xs \<inter> locks_\<mu> \<mu> = {}; wn_t' t xs \<rbrakk> \<Longrightarrow> P 
  \<rbrakk> \<Longrightarrow> P"
  by (induct h arbitrary: s \<mu>) auto

lemma wn_h_append_h_helper: 
  assumes 
    A: "wn_h h \<mu>" "h1@h2=h" and 
    C: "!!\<mu>1 \<mu>2. \<lbrakk> \<mu>=\<mu>1@\<mu>2 \<and> wn_h h1 \<mu>1 \<and> wn_h h2 \<mu>2 \<and> 
                   locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}\<rbrakk> \<Longrightarrow> P" 
  shows P
  using A C
  apply (induct h \<mu> arbitrary: h1 h2 P rule: wn_h.induct)
  apply auto
  apply fastsimp
  apply (case_tac h1)
  apply fastsimp
  apply auto
proof -
  case goal1
  show P 
    apply (rule goal1(1))
    apply simp
    apply (rule_tac ?\<mu>1.0 = "xs#\<mu>1" and ?\<mu>2.0 = "\<mu>2" in goal1(2))
    apply (insert goal1(3-))
    apply auto
    done
qed

lemma wn_h_append_h: 
  "\<lbrakk>wn_h (h1@h2) \<mu>; 
    !!\<mu>1 \<mu>2. \<lbrakk> \<mu>=\<mu>1@\<mu>2 \<and> wn_h h1 \<mu>1 \<and> wn_h h2 \<mu>2 \<and> 
               locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}\<rbrakk> \<Longrightarrow> P
  \<rbrakk> \<Longrightarrow> P" 
  using wn_h_append_h_helper
  by blast

lemma wn_h_append_\<mu>_helper: 
  assumes 
  A: "wn_h h \<mu>" "\<mu>1@\<mu>2=\<mu>" and 
  C: "!!h1 h2. \<lbrakk> h=h1@h2 \<and> wn_h h1 \<mu>1 \<and> wn_h h2 \<mu>2 \<and> 
                 locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}\<rbrakk> \<Longrightarrow> P" 
  shows P
  using A C
  apply (induct h \<mu> arbitrary: \<mu>1 \<mu>2 P rule: wn_h.induct)
  apply auto
  apply (case_tac \<mu>1)
  apply fastsimp
  apply auto
proof -
  case goal1
  show P 
    apply (rule goal1(1))
    apply simp
    apply (rule_tac ?h1.0 = "t#h1" and ?h2.0 = "h2" in goal1(2))
    apply (insert goal1(3-))
    apply auto
    done
qed

lemma wn_h_append_\<mu>: 
  "\<lbrakk>wn_h h (\<mu>1@\<mu>2); 
    !!h1 h2. \<lbrakk> h=h1@h2 \<and> wn_h h1 \<mu>1 \<and> wn_h h2 \<mu>2 \<and> 
               locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}
             \<rbrakk> \<Longrightarrow> P
   \<rbrakk> \<Longrightarrow> P" 
  using wn_h_append_\<mu>_helper by blast

lemma wn_h_appendI: 
  "\<lbrakk>wn_h h1 \<mu>1; wn_h h2 \<mu>2; locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}\<rbrakk> \<Longrightarrow> 
     wn_h (h1@h2) (\<mu>1@\<mu>2)"
  by (induct h1 \<mu>1 arbitrary: h2 \<mu>2 rule: wn_h.induct) auto

lemma wn_h_prependI: 
  "\<lbrakk>wn_t' t xs; wn_h h \<mu>; set xs \<inter> locks_\<mu> \<mu> = {}\<rbrakk> \<Longrightarrow> wn_h (t#h) (xs#\<mu>)"
  by auto

lemma wn_h_singlehE: "\<lbrakk>wn_h [t] \<mu>; !!xs. \<lbrakk>\<mu>=[xs]; wn_t' t xs\<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
  by (cases \<mu>) auto


text {*
  Auxilliary lemma to split the list of lock-stacks w.r.t. to that a hedge is well-nested by some tree in that hedge.
*}
lemma wn_h_split_aux: 
  assumes 
  WN: "wn_h h \<mu>" and 
  HFMT[simp]: "h=h1@t#h2" and
  C: "!!\<mu>1 xs \<mu>2. \<lbrakk> 
        \<mu>=\<mu>1@xs#\<mu>2; 
        wn_t' t xs; wn_h h1 \<mu>1; wn_h h2 \<mu>2; 
        locks_\<mu> \<mu>1 \<inter> set xs = {}; locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}; 
        set xs \<inter> locks_\<mu> \<mu>2 = {} 
      \<rbrakk> \<Longrightarrow> P"
  shows "P" 
  using WN[simplified]
  apply (elim wn_h_append_h wn_h_prepend_h conjE)
  apply (rule C)
  apply (auto)
  done



subsubsection {* Relation to Path Condition*}
text {* We show that the notion of well-nestedness on paths and trees are equivalent, 
  i.e. a configuration is well-nested w.r.t. a lock stack @{text \<mu>} if and only if 
  all trees from that configuration are well-nested w.r.t. @{text \<mu>}. *}


text {*
  A process @{term \<pi>} is well-nested w.r.t. some stack of locks @{term \<mu>}, if all 
  its execution trees are well-nested w.r.t. @{term \<mu>}: 
*}
definition "wn_\<pi>_t \<Delta> \<pi> xs == (\<forall>t c'. tsem \<Delta> \<pi> t c' \<longrightarrow> wn_t t xs)"

definition "wn_c_h \<Delta> c \<mu> == (\<forall>h c'. hsem \<Delta> c h c' \<longrightarrow> wn_h h \<mu>)"

lemma wn_\<pi>_tI[intro?]: "\<lbrakk>!!t c'. tsem \<Delta> \<pi> t c' \<Longrightarrow> wn_t t xs \<rbrakk> \<Longrightarrow> wn_\<pi>_t \<Delta> \<pi> xs" 
  by (auto simp add: wn_\<pi>_t_def)
lemma wn_c_hI[intro?]: "\<lbrakk>!!h c'. hsem \<Delta> c h c' \<Longrightarrow> wn_h h \<mu> \<rbrakk> \<Longrightarrow> wn_c_h \<Delta> c \<mu>" 
  by (auto simp add: wn_c_h_def)

lemma wn_\<pi>_t_distinct: "wn_\<pi>_t \<Delta> \<pi> \<mu> \<Longrightarrow> distinct \<mu>"
  apply (cases \<pi>)
  apply (unfold wn_\<pi>_t_def)
  by (auto intro: wn_t.intros wnt_distinct)


lemma wn_c_h_prepend1: assumes A: "wn_c_h \<Delta> (\<pi>#c) (xs#\<mu>)" 
  shows "wn_\<pi>_t \<Delta> \<pi> xs" "wn_c_h \<Delta> c \<mu>" "set xs \<inter> locks_\<mu> \<mu> = {}" 
proof -
  from A have A': "!!h c'. hsem \<Delta> (\<pi>#c) h c' \<Longrightarrow> wn_h h (xs#\<mu>)" 
    by (auto simp add: wn_c_h_def)
  from A'[of "map NLEAF (\<pi>#c)" "\<pi>#c", simplified] 
  show "set xs \<inter> locks_\<mu> \<mu> = {}" 
    by auto
  show "wn_\<pi>_t \<Delta> \<pi> xs" proof
    fix t c' assume A: "tsem \<Delta> \<pi> t c'"
    from A'[OF hsem_cons[OF A hsem_id]] show "wn_t t xs" 
      by (auto simp add: wnt_eq_wnt')
  qed
  
  show "wn_c_h \<Delta> c \<mu>" proof
    fix h c' assume A: "hsem \<Delta> c h c'"
    from A'[OF hsem_cons[OF tsem_leaf A]] show "wn_h h \<mu>" by auto
  qed
qed

lemma wn_c_h_prepend2: 
  "\<lbrakk>wn_\<pi>_t \<Delta> \<pi> xs; wn_c_h \<Delta> c \<mu>; set xs \<inter> locks_\<mu> \<mu> = {}\<rbrakk> \<Longrightarrow> 
     wn_c_h \<Delta> (\<pi>#c) (xs#\<mu>)"
  apply (auto simp add: wn_c_h_def wn_\<pi>_t_def)
  apply (erule hsem_split_single)
  apply (auto simp add: wnt_eq_wnt')
  done

lemma wn_c_h_prepend[simp]: 
  "wn_c_h \<Delta> (\<pi>#c) (xs#\<mu>) \<longleftrightarrow> 
     wn_\<pi>_t \<Delta> \<pi> xs \<and> wn_c_h \<Delta> c \<mu> \<and> set xs \<inter> locks_\<mu> \<mu> = {}" 
  using wn_c_h_prepend1 wn_c_h_prepend2 by fast

lemma wn_c_h_empty[simp]: "wn_c_h \<Delta> c [] \<longleftrightarrow> (c=[])" by (auto simp add: wn_c_h_def)

lemma wn_c_h_prepend_c: 
  "\<lbrakk>wn_c_h \<Delta> (\<pi>#c) \<mu>; 
     !!xs \<mu>'. \<lbrakk>\<mu>=xs#\<mu>'; wn_\<pi>_t \<Delta> \<pi> xs; wn_c_h \<Delta> c \<mu>'; 
               set xs \<inter> locks_\<mu> \<mu>' = {} \<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
  by (cases \<mu>) (auto)

lemma wn_c_h_simps[simp]: "wn_c_h \<Delta> [] \<mu> \<longleftrightarrow> (\<mu>=[])" 
  by (unfold wn_c_h_def) (auto)
  
lemma (in LDPN) wn\<pi>2wnt: "\<lbrakk>tsem \<Delta> (q,w) t c'; wn_\<pi> \<Delta> (q,w) \<mu>\<rbrakk> \<Longrightarrow> wn_t t \<mu>"
proof (induct arbitrary: \<mu> rule: tsem.induct)
  case tsem_leaf thus ?case by (auto intro: wn_t.intros dest: wn_\<pi>_distinct)
next
  case (tsem_nospawn q \<gamma> l q' w r t ct' \<mu>) note C=this
  show ?case proof (cases l)
    case LNone[simp] 
    from C have "wn_t t \<mu>"
      by (rule_tac C) (auto intro: ppairs.intros C simp add: wn_\<pi>_def)
    thus ?thesis by (auto intro: wn_t.intros)
  next
    case (LAcq x)[simp]
    from C have "wn_t t (x#\<mu>)"
      by (rule_tac C) (auto intro: ppairs.intros C simp add: wn_\<pi>_def)
    moreover hence "x\<notin>set \<mu>" by (auto dest: wnt_distinct)
    ultimately show ?thesis by (auto intro: wn_t.intros)
  next
    case (LRel x)[simp]
    from wn_\<pi>_rel[OF tsem_nospawn.hyps(1)[simplified] tsem_nospawn.prems] 
    obtain \<mu>' where [simp]: "\<mu> = x # \<mu>'" "x \<notin> set \<mu>'" . 
    from C have "wn_t t \<mu>'"
      by (rule_tac C) (auto intro: ppairs.intros C simp add: wn_\<pi>_def)
    thus ?thesis by (auto intro: wn_t.intros)
  qed
next
  case (tsem_spawn q \<gamma> l qs ws q' w ts cs' r t ct' \<mu>) note C=this
  then obtain ll where [simp]: "l=LNone ll" by (cases l) auto
  from C have "wn_t t \<mu>"
    apply simp_all
    apply (rule_tac C)
    apply (auto intro: ppairs.intros C simp add: wn_\<pi>_def)
    done
  moreover from tsem_spawn.hyps(1,3) tsem_spawn.prems[rule_format] 
  have "wn_t ts []" by (auto intro: wn_\<pi>_spawn2)
  ultimately show ?case by (auto intro: wn_t.intros)
qed
      
lemma (in LDPN) wnt2wnp: 
  "\<lbrakk>ppairs \<Delta> (q,w) en l; \<forall>t c'. tsem \<Delta> (q,w) t c' \<longrightarrow> wn_t t \<mu>\<rbrakk> \<Longrightarrow> 
    (\<not>en \<longrightarrow> wn_p l \<mu>) \<and> (en \<longrightarrow> wn_p l [])"
proof (induct arbitrary: \<mu> rule: ppairs.induct)
  case ppairs_empty thus ?case by (auto intro: all_t_wnt_distinct)
next
  case (ppairs_genenv q \<gamma> a qs ws q' w en l r \<mu>)
  have "\<forall>t c'. tsem \<Delta> (qs, ws) t c' \<longrightarrow> wn_t t []" proof (intro allI impI)
    fix t c'
    assume A: "tsem \<Delta> (qs, ws) t c'"
    from ppairs_genenv.prems[rule_format, 
              OF tsem_spawn[OF ppairs_genenv.hyps(1) A tsem_leaf]
         ]
    show "wn_t t []" by (auto elim: wn_t.cases)
  qed
  from ppairs_genenv.hyps(3)[OF this] show ?case by blast
next
  case (ppairs_mvenv1 q \<gamma> a q' w r l \<mu>)[simplified] show ?case 
  proof (simp, cases a)
    case LNone[simp] 
    from ppairs_mvenv1.prems have "\<forall>t c'. tsem \<Delta> (q', w @ r) t c' \<longrightarrow> wn_t t \<mu>"
      by auto (drule tsem_nospawn[OF ppairs_mvenv1.hyps(1)], auto elim: wn_t.cases)
    with ppairs_mvenv1.hyps(3) show "wn_p l []" by auto
  next
    case (LAcq x)
    with tsem_nospawn[OF ppairs_mvenv1.hyps(1)] ppairs_mvenv1.prems 
    show "wn_p l []" 
      by (fastsimp intro: ppairs_mvenv1.hyps(3)[rule_format] elim: wn_t.cases)
  next
    case (LRel x) note [simp]=this
    from tsem_nospawn[OF ppairs_mvenv1.hyps(1)[simplified] tsem_leaf] 
    have T: "Ex (tsem \<Delta> 
                      (q, \<gamma> # r) 
                      (NNOSPAWN (LRel x) (NLEAF (q', w @ r)))
                )" 
      by blast
    obtain \<mu>' where [simp]: "\<mu>=x#\<mu>'" "x\<notin>set \<mu>'"
      apply (rule wn_t.cases[OF ppairs_mvenv1.prems[rule_format, OF T]])
      by simp_all
    from tsem_nospawn[OF ppairs_mvenv1.hyps(1)] ppairs_mvenv1.prems 
    show "wn_p l []" 
      by (fastsimp intro: ppairs_mvenv1.hyps(3)[rule_format] elim: wn_t.cases)

    (* FIXME: Something seems to go incredibly wrong here. This was the original proof, and Isabelle was not able to verify it.
      The only change I'm aware of that was made just before this did not work any more, was adding a label to the LNone-constructor:
    from ppairs_mvenv1.prems[rule_format] tsem_nospawn[OF ppairs_mvenv1.hyps(1)[simplified] tsem_leaf] have "wn_t ((NNOSPAWN (LRel x) (NLEAF (q', w @ r)))) \<mu>" by blast
    then obtain \<mu>' where [simp]: "\<mu>=x#\<mu>'" "x\<notin>set \<mu>'" by (auto elim: wn_t.cases)
    from tsem_nospawn[OF ppairs_mvenv1.hyps(1)] ppairs_mvenv1.prems show "wn_p l []" by (fastsimp intro: ppairs_mvenv1.hyps(3)[rule_format] elim: wn_t.cases)
    *)
  qed
next
  case (ppairs_mvenv2 q \<gamma> a qs ws q' w r l \<mu>)[simplified] 
  show ?case
    using tsem_spawn[OF ppairs_mvenv2.hyps(1)] ppairs_mvenv2.prems 
      ppairs_mvenv2.hyps(1)
    apply (cases a)
    apply (blast intro: ppairs_mvenv2.hyps(3)[rule_format] elim: wn_t.cases)
    apply auto
    done
next
  case (ppairs_prepend1 q \<gamma> a q' w r l \<mu>)[simplified] show ?case
  proof (simp, cases a)
    case LNone 
    with tsem_nospawn[OF ppairs_prepend1.hyps(1)] ppairs_prepend1.prems 
    show "wn_p (a#l) \<mu>" 
      by (fastsimp intro: ppairs_prepend1.hyps(3)[rule_format] elim: wn_t.cases)
  next
    case (LAcq x) 
    with tsem_nospawn[OF ppairs_prepend1.hyps(1)] ppairs_prepend1.prems 
    show "wn_p (a#l) \<mu>" 
      by (fastsimp intro: ppairs_prepend1.hyps(3)[rule_format] elim: wn_t.cases)
  next
    case (LRel x) note [simp]=this
    from tsem_nospawn[OF ppairs_prepend1.hyps(1)[simplified] tsem_leaf] have 
      T: "Ex (tsem \<Delta> (q, \<gamma> # r) (NNOSPAWN (LRel x) (NLEAF (q', w @ r))))" 
      by blast
    obtain \<mu>' where [simp]: "\<mu>=x#\<mu>'" "x\<notin>set \<mu>'"
      apply (rule wn_t.cases[OF ppairs_prepend1.prems[rule_format, OF T]])
      by simp_all
    from tsem_nospawn[OF ppairs_prepend1.hyps(1)] ppairs_prepend1.prems 
    show "wn_p (a#l) \<mu>" 
      by (fastsimp intro: ppairs_prepend1.hyps(3)[rule_format] elim: wn_t.cases)

    (* Original proof that stopped to work. See above.
    from ppairs_prepend1.prems[rule_format] tsem_nospawn[OF ppairs_prepend1.hyps(1)[simplified] tsem_leaf] have "wn_t ((NNOSPAWN (LRel x) (NLEAF (q', w @ r)))) \<mu>" by blast
    then obtain \<mu>' where [simp]: "\<mu>=x#\<mu>'" "x\<notin>set \<mu>'" by (auto elim: wn_t.cases)
    from tsem_nospawn[OF ppairs_prepend1.hyps(1)] ppairs_prepend1.prems show "wn_p (a#l) \<mu>" by (fastsimp intro: ppairs_prepend1.hyps(3)[rule_format] elim: wn_t.cases)
      *)
  qed
next
  case (ppairs_prepend2 q \<gamma> a qs ws q' w r l \<mu>)[simplified] 
  from ppairs_prepend2.prems[rule_format] have 
    H: "!!c t. tsem \<Delta> (q, \<gamma> # r) t c \<Longrightarrow> wn_t t \<mu>" by blast
  show ?case using ppairs_prepend2.hyps(1) 
    by (cases a) 
       (auto intro: ppairs_prepend2.hyps(3)[rule_format] 
             dest: tsem_spawn[OF ppairs_prepend2.hyps(1) tsem_leaf] H 
             elim: wn_t.cases
       )
qed
      
theorem (in LDPN) wn\<pi>_eq_wn\<pi>t: "wn_\<pi> \<Delta> \<pi> \<mu> \<longleftrightarrow> wn_\<pi>_t \<Delta> \<pi> \<mu>" using wnt2wnp
  by (auto intro: wn\<pi>2wnt simp add: wn_\<pi>_def wn_\<pi>_t_def)


theorem (in LDPN) wnc_eq_wnch: "wn_c \<Delta> c \<mu> \<longleftrightarrow> wn_c_h \<Delta> c \<mu>" 
  apply rule
  apply (induct c arbitrary: \<mu>)
  apply simp
  apply (erule wn_c_prepend_c)
  apply (simp add: wn\<pi>_eq_wn\<pi>t)
  apply (induct c arbitrary: \<mu>)
  apply (auto simp add: wn_c_h_def) [1]
  apply (erule wn_c_h_prepend_c)
  apply (simp add: wn\<pi>_eq_wn\<pi>t)
  done


subsection {* Well-Nestedness and Tree Scheduling *}
text {*
  In this section we show that well-nestedness is invariant under the tree scheduling 
  relation. This is important, as it shows that we cannot reach non-well-nested trees
  from well-nested ones.
*}


lemma wnt_preserve_nospawn: 
  "\<lbrakk> lock_valid (set xs) l X'; wn_t' (NNOSPAWN l t) xs\<rbrakk> \<Longrightarrow> 
    \<exists>xs'. X'=set xs' \<and> lock_valid_xs l xs xs' \<and> wn_t' t xs'"
  apply (cases l)
  apply (rule_tac x="xs" in exI)
  apply (force intro: lock_valid_xs.intros dest: wnt_distinct')
  apply (rule_tac x="(X#xs)" in exI)
  apply (force intro: lock_valid_xs.intros dest: wnt_distinct')
  apply (rule_tac x="tl xs" in exI)
  apply (force simp add: insert_ident intro: lock_valid_xs.intros dest: wnt_distinct')
  done

lemma wn_h_preserve_nospawn: 
  "\<lbrakk> lock_valid (locks_\<mu> \<mu>) l X'; wn_h (h1@(NNOSPAWN l t)#h2) \<mu>\<rbrakk> \<Longrightarrow> 
    \<exists>\<mu>'. X'=locks_\<mu> \<mu>' \<and> wn_h (h1@t#h2) \<mu>'"
  apply (cases l)
  apply (auto elim!: wn_h_prepend_h wn_h_append_h)
  apply (rule_tac x="\<mu>1@xs#\<mu>'" in exI)
  apply (force intro!: wn_h_appendI)
  apply (rule_tac x="\<mu>1@(X#xs)#\<mu>'" in exI)
  apply (force intro!: wn_h_appendI)
  apply (rule_tac x="\<mu>1@(\<mu>'a)#\<mu>'" in exI)
  apply (rule conjI)
  apply (rule iffD1[OF insert_ident])
  apply assumption
  apply (auto intro!: wn_h_appendI)
  done

text {*
  All-in-one lemma for reasoning about a non-spawning step on a well-nested hedge.
  In words: If we make a non-speaining step on a well-nested hedge:
  \begin{itemize}
    \item We can split the list of lock stacks according to the tree that made 
          the step,
    \item The lock stack of the tree that made the step changes according to the label 
      (cf. @{const lock_valid_xs}),
    \item And the resulting hedge is well-nested w.r.t. the new locks, too.
  \end{itemize}
*}
lemma wn_h_split_nospawn: 
  assumes 
  A: "lock_valid (locks_\<mu> \<mu>) l Xh" "wn_h (h1@(NNOSPAWN l t)#h2) \<mu>" and 
  C: "!!\<mu>1 xs \<mu>2 xsh. \<lbrakk>
    \<mu>=\<mu>1@xs#\<mu>2; 
    Xh=locks_\<mu> \<mu>1 \<union> set xsh \<union> locks_\<mu> \<mu>2; 
    lock_valid_xs l xs xsh; 
    wn_t' (NNOSPAWN l t) xs; 
    wn_t' t xsh;
    wn_h h1 \<mu>1;
    wn_h h2 \<mu>2;
    wn_h (h1@t#h2) (\<mu>1@xsh#\<mu>2);
    locks_\<mu> \<mu>1 \<inter> set xs = {};
    locks_\<mu> \<mu>1 \<inter> set xsh = {};
    locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {};
    locks_\<mu> \<mu>2 \<inter> set xs = {};
    locks_\<mu> \<mu>2 \<inter> set xsh = {}
  \<rbrakk> \<Longrightarrow> P"
  shows P
proof -
  from A(2) obtain \<mu>1 xs \<mu>2 where 
    SPLIT_simp[simp]: "\<mu>=\<mu>1@xs#\<mu>2" and 
    SPLIT: "wn_h h1 \<mu>1" "wn_t' (NNOSPAWN l t) xs" "wn_h h2 \<mu>2" 
           "locks_\<mu> \<mu>1 \<inter> set xs = {}" "locks_\<mu> \<mu>1 \<inter> locks_\<mu> \<mu>2 = {}" 
           "set xs \<inter> locks_\<mu> \<mu>2 = {}"
    by (fastsimp elim: wn_h_prepend_h wn_h_append_h)
  show ?thesis proof (cases l)
    case LNone[simp] 
    from SPLIT(2) have "wn_t' t xs" "lock_valid_xs l xs xs"
      by (auto intro: lock_valid_xs.intros dest: wnt_distinct')
    moreover with SPLIT have "wn_h (h1@t#h2) (\<mu>1@xs#\<mu>2)" 
      by (auto intro!: wn_h_appendI wn_h_prependI)
    ultimately show ?thesis using A(1)[simplified] SPLIT SPLIT_simp 
      by (blast intro!: C)
  next
    case (LRel x)[simp] 
    from SPLIT(2) obtain xsh where 
      [simp]: "xs=x#xsh" and 
        WN': "wn_t' t xsh" "x\<notin>set xsh" 
      by auto
    moreover with SPLIT have "wn_h (h1@t#h2) (\<mu>1@xsh#\<mu>2)" 
      by (auto intro!: wn_h_appendI wn_h_prependI)
    moreover from wnt_distinct'[OF WN'(1)] WN'(2) have 
      "lock_valid_xs l xs xsh" 
      by (auto intro: lock_valid_xs.intros)
    ultimately show ?thesis 
      using A(1)[simplified] WN' SPLIT SPLIT_simp by (fastsimp intro!: C)
  next
    case (LAcq x)[simp] 
    from SPLIT(2) have "wn_t' t (x#xs)" "lock_valid_xs l xs (x#xs)" 
      by (auto intro: lock_valid_xs.intros dest!: wnt_distinct')
    moreover with SPLIT A(1)[simplified] have "wn_h (h1@t#h2) (\<mu>1@(x#xs)#\<mu>2)"
      by (auto intro!: wn_h_appendI wn_h_prependI)
    ultimately show ?thesis 
      using A(1)[simplified] SPLIT SPLIT_simp
      apply (rule_tac C)
      apply assumption+
      defer
      apply assumption+
      apply auto
      done
  qed
qed
  
lemma wn_h_preserve_spawn: 
  "\<lbrakk> lock_valid (locks_\<mu> \<mu>) l X'; wn_h (h1@(NSPAWN l ts t)#h2) \<mu>\<rbrakk> \<Longrightarrow> 
    \<exists>\<mu>'. X'=locks_\<mu> \<mu>' \<and> wn_h (h1@ts#t#h2) \<mu>'"
  apply (cases l)
  apply (auto elim!: wn_h_prepend_h wn_h_append_h)
  apply (rule_tac x="\<mu>1@[]#xs#\<mu>'" in exI)
  apply (auto intro!: wn_h_appendI)
  done

lemma wn_h_preserve_spawn': 
  "\<lbrakk> lock_valid (locks_\<mu> \<mu>) l X'; wn_h (h1@(NSPAWN l ts t)#h2) \<mu>\<rbrakk> \<Longrightarrow> 
  \<exists>\<mu>1 xs \<mu>2. \<mu>=\<mu>1@xs#\<mu>2 \<and> X'=locks_\<mu> \<mu>1 \<union> set xs \<union> locks_\<mu> \<mu>2 \<and> 
             wn_h (h1@ts#t#h2) (\<mu>1@[]#xs#\<mu>2)"
  apply (cases l)
  apply (auto elim!: wn_h_prepend_h wn_h_append_h)
  apply (rule_tac x="\<mu>1" in exI)
  apply (rule_tac x="xs" in exI)
  apply (rule_tac x="\<mu>'" in exI)
  apply (auto intro!: wn_h_appendI)
  done

lemma wn_h_preserve_rel: 
  "\<lbrakk> (h,l,h')\<in>sched_rel; lock_valid (locks_\<mu> \<mu>) l X'; wn_h h \<mu>; 
     !!\<mu>'. \<lbrakk> X'=locks_\<mu> \<mu>'; wn_h h' \<mu>'\<rbrakk> \<Longrightarrow> P 
   \<rbrakk> \<Longrightarrow> P"
  by (auto elim!: sched_rel.cases dest: wn_h_preserve_spawn wn_h_preserve_nospawn)

lemma wn_h_spawn_simps[simp]: 
  "\<not>wn_h (h @ (NSPAWN (LAcq x) ts t) # h') \<mu>"
  "\<not>wn_h (h @ (NSPAWN (LRel x) ts t) # h') \<mu>"
  by (auto elim!: wn_h_prepend_h wn_h_append_h)

lemmas wn_h_spawn_simps_add[simp] = 
  wn_h_spawn_simps[where h="[]", simplified]
  wn_h_spawn_simps[where h="[tx]", simplified, standard]

lemma wn_h_spawn_imp_LNoneE: 
  "\<lbrakk>wn_h (h @ (NSPAWN l ts t) # h') \<mu>; !!ll. l=LNone ll \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P" 
  by (cases l) auto

(* Not needed, as simplification does this rewriting automatically
lemma wn_h_spawn_imp_LNone[simplified, simp]: 
  "\<not>wn_h (h @ (NSPAWN (LAcq x) ts t) # h') \<mu>" 
  "\<not>wn_h (h @ (NSPAWN (LRel x) ts t) # h') \<mu>" 
  by auto
*)

(* Original lemmas, before change to labeled LNone:
lemma wn_h_spawn_imp_LNone[simp]: "wn_h (h @ (NSPAWN l ts t) # h') \<mu> \<Longrightarrow> l=LNone" by (cases l) auto
lemmas wn_h_spawn_imp_LNone_add[simp] = wn_h_spawn_imp_LNone[where h="[]", simplified] wn_h_spawn_imp_LNone[where h="[tx]", simplified, standard]
*)

end
